#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
PEAS assessment initialization
"""

from collections import OrderedDict
from decimal import Decimal

from persistent import Persistent
from zope.annotation.interfaces import IAnnotations
from zope.cachedescriptors.property import Lazy
from zope.component import adapts
from zope.component import adapter
from zope.container.contained import Contained
from zope.interface import Interface
from zope.interface import implements
from zope.interface import implementer
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.proxy import sameProxiedObjects
from zope.schema import Bool
from zope.schema import Choice
from zope.schema import Decimal as DecimalField
from zope.schema import Int
from zope.schema import TextLine

from schooltool.app.app import StartUpBase
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.gradebook.activity import Activity
from schooltool.gradebook.activity import Worksheet
from schooltool.gradebook.interfaces import IActivities
from schooltool.gradebook.interfaces import ICategoryContainer
from schooltool.requirement.interfaces import IScoreSystemContainer
from schooltool.requirement.scoresystem import CustomScoreSystem
from schooltool.requirement.scoresystem import GlobalDiscreteValuesScoreSystem
from schooltool.requirement.scoresystem import RangedValuesScoreSystem
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.term.interfaces import ITerm
from schooltool.course.interfaces import ISection

from schooltool.peas import PeasMessage as _


ASSESSMENT_PREFERENCES_KEY = 'schooltool.peas.assessment_preferences'
PEAS_SCORESYSTEM_KEY = 'schooltool.peas.scoresystem'
MOCK_SHEET_ID = u'mock'
MOCK_OTHER_SHEET_ID = u'mock-other'
TARGET_SHEET_ID = u'target'
UCE_SHEET_ID = u'uce'
ASSESSMENT_SHEETS = OrderedDict([
    (u'bot',              {'title': u'Beginning of Term', 'editable': True}),
    (u'mot',              {'title': u'Middle of Term',    'editable': True}),
    (u'eot',              {'title': u'End of Term',       'editable': True}),
    (MOCK_SHEET_ID,       {'title': u'MOCK - PEAS',       'editable': False}),
    (MOCK_OTHER_SHEET_ID, {'title': u'MOCK - Other',      'editable': False}),
    (TARGET_SHEET_ID,     {'title': u'Target',            'editable': False}),
    (UCE_SHEET_ID,        {'title': u'UCE',               'editable': False}),
])


class IWeightField(Interface):

    pass


class WeightField(DecimalField):

    implements(IWeightField)


class IAssessmentPreferences(Interface):

    bot = Bool(title=_(u'Beginning of Term'), default=False, required=False)
    bot_weight = WeightField(
        title=_(u'Weight in term summary'),
        description=_(u'Leave blank for equal weighting'),
        min=Decimal(0),
        required=False)
    mot = Bool(title=_(u'Middle of Term'), default=False, required=False)
    mot_weight = WeightField(
        title=_(u'Weight in term summary'),
        description=_(u'Leave blank for equal weighting'),
        min=Decimal(0),
        required=False)
    eot = Bool(title=_(u'End of Term'), default=False, required=False)
    eot_weight = WeightField(
        title=_(u'Weight in term summary'),
        description=_(u'Leave blank for equal weighting'),
        min=Decimal(0),
        required=False)
    mock = Bool(title=_(u'MOCK - PEAS'), default=False, required=False)
    mock_term = Choice(
        title=_(u'Term for deploying the MOCK - PEAS sheet'),
        source='schooltool.peas.schoolyear_terms',
        required=False)
    mock_other = Bool(title=_(u'MOCK - Other'), default=False, required=False)
    mock_other_term = Choice(
        title=_(u'Term for deploying the MOCK - Other sheet'),
        source='schooltool.peas.schoolyear_terms',
        required=False)


class AssessmentPreferences(Persistent, Contained):

    implements(IAssessmentPreferences)

    bot = None
    bot_weight = None
    mot = None
    mot_weight = None
    eot = None
    eot_weight = None
    mock = None
    mock_term = None
    mock_other = None
    mock_other_term = None


@adapter(IAssessmentPreferences)
@implementer(ISchoolYear)
def getAssessmentPreferencesSchoolYear(preferences):
    return preferences.__parent__


def getAssessmentPreferences(schoolyear):
    annotations = IAnnotations(schoolyear)
    name = 'assessment_preferences'
    try:
        return annotations[ASSESSMENT_PREFERENCES_KEY]
    except KeyError:
        annotations[ASSESSMENT_PREFERENCES_KEY] = AssessmentPreferences()
        annotations[ASSESSMENT_PREFERENCES_KEY].__name__ = name
        annotations[ASSESSMENT_PREFERENCES_KEY].__parent__ = schoolyear
        return annotations[ASSESSMENT_PREFERENCES_KEY]


def default_deploy_term(schoolyear):
    terms = sorted(schoolyear.values(), key=lambda term: term.first)
    if terms:
        return terms[0]


def get_peas_scoresystem():
    app = ISchoolToolApplication(None)
    scoresystems = IScoreSystemContainer(app)
    return scoresystems[PEAS_SCORESYSTEM_KEY]


class AssessmentSheetDeployBase(object):

    @Lazy
    def default_deploy_term(self):
        return default_deploy_term(self.schoolyear)

    def get_mock_term(self):
        if self.preferences.mock_term is not None:
            return self.preferences.mock_term
        # deploy mock sheet to second term automatically
        return self.default_deploy_term

    def get_mock_other_term(self):
        if self.preferences.mock_other_term is not None:
            return self.preferences.mock_other_term
        # deploy mock sheet to second term automatically
        return self.default_deploy_term

    def get_target_term(self):
        return self.default_deploy_term

    @Lazy
    def preferences(self):
        return IAssessmentPreferences(self.schoolyear)

    @Lazy
    def peas_scoresystem(self):
        return get_peas_scoresystem()

    @Lazy
    def ranged_scoresystem(self):
        return RangedValuesScoreSystem(
            u'generated', min=0, max=100)

    @Lazy
    def default_category(self):
        app = ISchoolToolApplication(None)
        return ICategoryContainer(app).default_key

    def deploy_assessment_sheets(self, section):
        activities = IActivities(section)
        term = ITerm(section)
        for sheet, sheet_data in ASSESSMENT_SHEETS.items():
            if not sheet_data['editable']:
                continue
            title = sheet_data['title']
            enabled = getattr(self.preferences, sheet, False)
            if enabled:
                if sheet in activities:
                    if activities[sheet].hidden:
                        activities[sheet].hidden = False
                        activities[sheet].deployed = False
                    continue
                activities[sheet] = Worksheet(title)
            elif sheet in activities:
                ws = activities[sheet]
                ws.hidden = True
                ws.deployed = True
        if self.preferences.mock:
            if sameProxiedObjects(term, self.get_mock_term()):
                if MOCK_SHEET_ID not in activities:
                    title = ASSESSMENT_SHEETS[MOCK_SHEET_ID]['title']
                    activities[MOCK_SHEET_ID] = Worksheet(title)
                elif activities[MOCK_SHEET_ID].hidden:
                    activities[MOCK_SHEET_ID].hidden = False
                    activities[MOCK_SHEET_ID].deployed = False
            elif MOCK_SHEET_ID in activities:
                ws = activities[MOCK_SHEET_ID]
                ws.hidden = True
                ws.deployed = True
        elif MOCK_SHEET_ID in activities:
            ws = activities[MOCK_SHEET_ID]
            ws.hidden = True
            ws.deployed = True
        if self.preferences.mock_other:
            if sameProxiedObjects(term, self.get_mock_other_term()):
                if MOCK_OTHER_SHEET_ID not in activities:
                    title = ASSESSMENT_SHEETS[MOCK_OTHER_SHEET_ID]['title']
                    activities[MOCK_OTHER_SHEET_ID] = Worksheet(title)
                elif activities[MOCK_OTHER_SHEET_ID].hidden:
                    activities[MOCK_OTHER_SHEET_ID].hidden = False
                    activities[MOCK_OTHER_SHEET_ID].deployed = False
            elif MOCK_OTHER_SHEET_ID in activities:
                ws = activities[MOCK_OTHER_SHEET_ID]
                ws.hidden = True
                ws.deployed = True
        elif MOCK_OTHER_SHEET_ID in activities:
            ws = activities[MOCK_OTHER_SHEET_ID]
            ws.hidden = True
            ws.deployed = True
        if sameProxiedObjects(term, self.get_target_term()):
            if TARGET_SHEET_ID not in activities:
                title = ASSESSMENT_SHEETS[TARGET_SHEET_ID]['title']
                ws = Worksheet(title)
                ws.deployed = True
                activities[TARGET_SHEET_ID] = ws
                activity_title = title + ' (%)'
                activity_label = u'%'
                activity = Activity(
                    activity_title, self.default_category,
                    self.ranged_scoresystem, None, activity_label)
                ws[TARGET_SHEET_ID+'-ranged'] = activity
                activity_title = title + ' (1-9)'
                activity_label = u'1-9'
                activity = Activity(
                    activity_title, self.default_category,
                    self.peas_scoresystem, None, activity_label)
                ws[TARGET_SHEET_ID] = activity
            elif activities[TARGET_SHEET_ID].hidden:
                activities[TARGET_SHEET_ID].hidden = False
            if UCE_SHEET_ID not in activities:
                title = ASSESSMENT_SHEETS[UCE_SHEET_ID]['title']
                ws = Worksheet(title)
                ws.hidden = True
                ws.deployed = True
                activities[UCE_SHEET_ID] = ws
                activity = Activity(
                    title, self.default_category,
                    self.peas_scoresystem, None, title)
                ws[UCE_SHEET_ID] = activity
        else:
            if TARGET_SHEET_ID in activities:
                ws = activities[TARGET_SHEET_ID]
                ws.hidden = True
                ws.deployed = True


class CreateAssessmentSheetsSubscriber(ObjectEventAdapterSubscriber,
                                       AssessmentSheetDeployBase):

    adapts(IObjectAddedEvent, ISection)

    @Lazy
    def schoolyear(self):
        return ISchoolYear(self.object)

    def __call__(self):
        self.deploy_assessment_sheets(self.object)


PEASScoreSystem = GlobalDiscreteValuesScoreSystem(
    'PEAS Score System',
    _('PEAS Score System'), None,
    [
        ('1', u'D1', Decimal(1), Decimal(80)),
        ('2', u'D2', Decimal(2), Decimal(75)),
        ('3', u'C3', Decimal(3), Decimal(70)),
        ('4', u'C4', Decimal(4), Decimal(65)),
        ('5', u'C5', Decimal(5), Decimal(60)),
        ('6', u'C6', Decimal(6), Decimal(50)),
        ('7', u'P7', Decimal(7), Decimal(45)),
        ('8', u'P8', Decimal(8), Decimal(40)),
        ('9', u'F9', Decimal(9), Decimal(0)),
    ],
    '1', '8', True)


class PeasScoreSystemAppStartup(StartUpBase):

    after = ('schooltool.requirement.scoresystem',)

    def __call__(self):
        container = IScoreSystemContainer(self.app)
        if PEAS_SCORESYSTEM_KEY not in container:
            ss = PEASScoreSystem
            peas_scoresystem = CustomScoreSystem(ss.title, ss.description,
                                                 ss.scores, ss._bestScore,
                                                 ss._minPassingScore,
                                                 ss._isMaxPassingScore)
            container[PEAS_SCORESYSTEM_KEY] = peas_scoresystem


HUMANITIES = ['208', '223', '225', '241', '273']
SCIENCE = ['535', '545', '553']
ENGLISH = ['112']
MATHEMATICS = ['456']
CREDIT_LEVEL_MIN_SCORE = 6


def get_passed_credit(items):
    ss = get_peas_scoresystem()
    passed = {}
    credit = {}
    for code, grade in items:
        if grade is not None and ss.isPassingScore(grade):
            decimal_value = ss.scoresDict[grade]
            passed[code] = decimal_value
            if ss.scoresDict.get(grade) <= CREDIT_LEVEL_MIN_SCORE:
                credit[code] = decimal_value
    return passed, credit


def find_division(items):
    if not items:
        return 0
    passed, credit = get_passed_credit(items)
    if is_division_1(passed, credit):
        return 1
    elif is_division_2(passed, credit):
        return 2
    elif is_division_3(passed, credit):
        return 3
    elif is_division_4(passed, credit):
        return 4
    elif is_division_9(passed, credit):
        return 9
    else:
        return 7


def best8(passed):
    return sum(sorted(passed.values())[:8])


# XXX: use set intersection
def is_division_1(passed, credit):
    """
    Pass a minimum of eight subjects
    which must include English Language (with credit),
    a Humanity subject,
    mathematics and,
    a science subject.
    At least seven of the subjects must be a credit level or better.
    The aggregate for the best eight done subjects must not exceed 32
    """
    return (
        len(passed) >= 8 and
        [code for code in credit.keys() if code in ENGLISH] and
        [code for code in passed.keys() if code in HUMANITIES] and
        [code for code in passed.keys() if code in MATHEMATICS] and
        [code for code in passed.keys() if code in SCIENCE] and
        len(credit) >= 7 and
        best8(passed) <= 32
    )


def is_division_2(passed, credit):
    """
    Pass a minimum of eight subjects
    including English Language.
    Six of the subjects must be at a credit level or better.
    The aggregate for the best eight done subjects must not exceed 45.
    """
    return (
        len(passed) >= 8 and
        [code for code in passed.keys() if code in ENGLISH] and
        len(credit) >= 6 and
        best8(passed) <= 45
    )


def is_division_3(passed, credit):
    """
    Either
    (i) Pass a minimum of eight subjects (with at least 3 credits or better)
    OR (i) Pass a minimum of seven subjects (with at least 4 credits or better)
    OR (ii) Pass a minimum of five subjects with credits or better. 
    The aggregate for the best done subjects must not exceed 58.
    """
    return (
        (
            (len(passed) >= 8 and len(credit) >= 3) or
            (len(passed) >= 7 and len(credit) >= 4) or
            (len(passed) >= 5 and credit)
        ) and
        best8(passed) <= 58
    )


def is_division_4(passed, credit):
    """
    Either
    (i) Pass at least one subject with credit or better 
    OR (ii) Pass at least two subjects with pass 7 
    OR (iii) Pass at least three subjects with pass 8 or better
    """
    pass7 = [v for v in passed.values() if v == 7]
    pass8 = [v for v in passed.values() if v == 8]
    return (
        (
            credit or
            len(pass7) >= 2 or
            len(pass8) >= 3
        ) and
        best8(passed) <= 69
    )


def is_division_9(passed, credit):
    return best8(passed) <= 72


class IUCEResult(Interface):

    aggregate = Int(
        title=_(u'Aggregate'),
        required=False)

    division = TextLine(
        title=_(u'Division'),
        required=False)


class UCEResult(Persistent, Contained):

    aggregate = None
    division = None


def getUCEResult(person):
    annotations = IAnnotations(person)
    name = 'uce_result'
    try:
        return annotations[name]
    except KeyError:
        annotations[name] = UCEResult()
        annotations[name].__name__ = name
        annotations[name].__parent__ = person
        return annotations[name]
