#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2015 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
PEAS assessment.
"""

from decimal import Decimal
from urllib import urlencode
from collections import OrderedDict

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import implements
from zope.interface import Interface
from zope.i18n import translate
from zope.traversing.api import getName
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import ValidationError
from zope.schema.vocabulary import SimpleVocabulary
from zope.schema import Choice
from zope.security.proxy import removeSecurityProxy
from reportlab.lib import units

import zc.resourcelibrary
from z3c.form import button
from z3c.form import field
from z3c.form import form
from z3c.form import validator
from z3c.form.browser.checkbox import SingleCheckBoxFieldWidget

from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.membership import Membership
from schooltool.basicperson.browser.person import PersonTerm
from schooltool.basicperson.interfaces import IDemographics
from schooltool.basicperson.person import StudentLevels
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.course.interfaces import ISection
from schooltool.course.interfaces import ISectionContainer
from schooltool.course.interfaces import ICourseContainer
from schooltool.export.export import RequestXLSReportDialog
from schooltool.gradebook.interfaces import IActivities
from schooltool.gradebook.interfaces import IGradebook
from schooltool.group.browser.group import GroupAwarePersonTableFilter
from schooltool.group.browser.group import GroupJSONSearchView
from schooltool.group.group import defaultGroups
from schooltool.group.interfaces import IGroupContainer
from schooltool.level.interfaces import ILevelContainer
from schooltool.person.interfaces import IPerson
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.report.report import ReportLinkViewlet
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.requirement.interfaces import IDiscreteValuesScoreSystem
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.skin import flourish
from schooltool.term.interfaces import IDateManager
from schooltool.term.interfaces import ITerm

from schooltool.peas.assessment import IAssessmentPreferences
from schooltool.peas.assessment import IUCEResult
from schooltool.peas.assessment import AssessmentSheetDeployBase
from schooltool.peas.assessment import default_deploy_term
from schooltool.peas.assessment import get_peas_scoresystem
from schooltool.peas.assessment import find_division
from schooltool.peas.assessment import best8
from schooltool.peas.assessment import ASSESSMENT_SHEETS
from schooltool.peas.assessment import MOCK_SHEET_ID
from schooltool.peas.assessment import MOCK_OTHER_SHEET_ID
from schooltool.peas.assessment import TARGET_SHEET_ID
from schooltool.peas.assessment import UCE_SHEET_ID
from schooltool.peas.assessment import IWeightField
from schooltool.peas.browser.report import ICurrentTerm
from schooltool.peas.browser.report import term_id_getter
from schooltool.peas.browser.report import OverallReportBase
from schooltool.peas import PeasMessage as _


class AssessmentPreferencesLinkViewlet(flourish.page.LinkViewlet,
                                       ActiveSchoolYearContentMixin):

    @property
    def enabled(self):
        return self.schoolyear

    @property
    def url(self):
        preferences = IAssessmentPreferences(self.schoolyear)
        return absoluteURL(preferences, self.request)


class AssessmentPreferencesView(flourish.form.Form,
                                form.EditForm,
                                AssessmentSheetDeployBase):

    template = flourish.templates.Inherit(flourish.page.Page.template)
    content_template = ViewPageTemplateFile(
        'templates/assessment_preferences_form.pt')
    page_template = ViewPageTemplateFile(
        'templates/assessment_preferences_page.pt')

    fields = field.Fields(IAssessmentPreferences)
    fields['bot'].widgetFactory = SingleCheckBoxFieldWidget
    fields['mot'].widgetFactory = SingleCheckBoxFieldWidget
    fields['eot'].widgetFactory = SingleCheckBoxFieldWidget
    fields['mock'].widgetFactory = SingleCheckBoxFieldWidget
    fields['mock_other'].widgetFactory = SingleCheckBoxFieldWidget
    label = None
    legend = _('Assessment Sheets')

    @property
    def subtitle(self):
        return ISchoolYear(self.context).title

    def update(self):
        form.EditForm.update(self)

    @button.buttonAndHandler(STMessage('Submit'), name='apply')
    def handle_edit_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
        else:
            for sheet in ['bot', 'mot', 'eot']:
                if not data[sheet]:
                    data['%s_weight' % sheet] = None
            changes = self.applyChanges(data)
            if changes:
                self.status = self.successMessage
            else:
                self.status = self.noChangesMessage
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            if self.status == self.successMessage:
                self.update_sections()
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(STMessage('Cancel'))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        app = ISchoolToolApplication(None)
        schoolyear = ISchoolYear(self.context)
        return '%s/manage?schoolyear_id=%s' % (
            absoluteURL(app, self.request),
            schoolyear.__name__)

    def updateActions(self):
        super(AssessmentPreferencesView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @Lazy
    def schoolyear(self):
        return ISchoolYear(self.context)

    def update_sections(self):
        for term in self.schoolyear.values():
            for section in ISectionContainer(term).values():
                self.deploy_assessment_sheets(section)


class MockTermRequired(ValidationError):
    __doc__ = _('Term for deploying the MOCK - PEAS sheet is required')


class MockTermValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        super(MockTermValidator, self).validate(value)
        if value:
            widget = self.view.widgets['mock_term']
            name = widget.name
            if widget.noValueToken in self.request.get(name):
                raise MockTermRequired(value)


validator.WidgetValidatorDiscriminators(
    MockTermValidator,
    view=AssessmentPreferencesView,
    field=IAssessmentPreferences['mock'])


class MockOtherTermRequired(ValidationError):
    __doc__ = _('Term for deploying the MOCK - Other sheet is required')


class MockOtherTermValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        super(MockOtherTermValidator, self).validate(value)
        if value:
            widget = self.view.widgets['mock_other_term']
            name = widget.name
            if widget.noValueToken in self.request.get(name):
                raise MockOtherTermRequired(value)


validator.WidgetValidatorDiscriminators(
    MockOtherTermValidator,
    view=AssessmentPreferencesView,
    field=IAssessmentPreferences['mock_other'])


class SchoolYearTermsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        schoolyear = ISchoolYear(self.context)
        for term in sorted(schoolyear.values(), key=lambda t: t.first):
            result.append(self.createTerm(
                removeSecurityProxy(term),
                term.__name__,
                term.title
            ))
        return result


def SchoolYearTermsVocabularyFactory():
    return SchoolYearTermsVocabulary


class UceExamLinkViewlet(flourish.page.LinkViewlet,
                         ActiveSchoolYearContentMixin):

    @property
    def enabled(self):
        preferences = IAssessmentPreferences(self.schoolyear)
        return (preferences.bot is not None or
                preferences.mot is not None or
                preferences.eot is not None or
                preferences.mock is not None or
                preferences.mock_other is not None)

    @property
    def url(self):
        groups = IGroupContainer(self.schoolyear)
        students = groups['students']
        return '%s/uce_exam.html' % (
            absoluteURL(students, self.request))


class UceExamView(flourish.page.Page):

    content_template = flourish.templates.Inline('''
    <div tal:content="structure context/schooltool:content/ajax/view/context/members/group_aware_person_table" />
    ''')

    def update(self):
        zc.resourcelibrary.need('schooltool.table')

    @property
    def json_url(self):
        return '%s/json' % absoluteURL(self.context, self.request)


class UceGroupTableFilter(GroupAwarePersonTableFilter):

    template = ViewPageTemplateFile('templates/uce_search.pt')


class UCEViewBase(flourish.page.Page,
                  ActiveSchoolYearContentMixin):

    container_class = 'container extra-wide-container'
    prefix = 'grade.'

    @property
    def title(self):
        return self.context.title

    @Lazy
    def uce_term(self):
        return default_deploy_term(self.schoolyear)

    @Lazy
    def sections(self):
        result = []
        for section in ISectionContainer(self.uce_term).values():
            if self.context in section.members.all():
                for course in section.courses:
                    if course.government_id is not None:
                        try:
                            int(course.government_id)
                            result.append({
                                'section': section,
                                'course': course,
                                'code': course.government_id,
                            })
                        except:
                            pass
        return sorted(result, key=lambda i: int(i['code']))

    @property
    def gender(self):
        return self.context.gender[0].upper()

    def age(self):
        if self.context.birth_date is not None:
            today = getUtility(IDateManager).today
            return today.year - self.context.birth_date.year

    def name(self):
        return '%s %s' % (self.context.last_name.upper(),
                          self.context.first_name.upper())


class UceExamEditView(UCEViewBase):

    content_template = ViewPageTemplateFile('templates/uce_edit.pt')

    def nextURL(self):
        groups = IGroupContainer(self.schoolyear)
        students = groups['students']
        return '%s/uce_exam.html' % absoluteURL(students, self.request)

    def update(self):
        if 'data' in self.request:
            decoder = flourish.tal.JSONDecoder()
            data = decoder.decode(self.request['data'])
            for i, item in enumerate(sorted(data.items(),
                                            key=lambda i: int(i[0]))):
                k, v = item
                self.request.form[self.prefix+str(i)+'-code'] = k
                self.request.form[self.prefix+str(i)+'-grade'] = v
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
            return
        if 'SUBMIT' in self.request:
            encoder = flourish.tal.JSONEncoder()
            data = {}
            for key, value in self.request.items():
                if key.startswith(self.prefix):
                    code, grade = value
                    if code:
                        data[code] = grade
            params = {
                'schoolyear_id': self.schoolyear.__name__,
                'data': encoder.encode(data),
                'aggregate': self.request.get('aggregate'),
                'division': self.request.get('division'),
            }
            url = '%s/uce_display.html?%s' % (
                absoluteURL(self.context, self.request), urlencode(params))
            self.request.response.redirect(url)


class StudentUceExamView(UCEViewBase):

    content_template = ViewPageTemplateFile('templates/uce_view.pt')
    errors = ()
    warnings = ()
    data = {}

    @Lazy
    def division(self):
        return find_division(self.data.items())

    def save(self, data):
        evaluator = getName(IPerson(self.request.principal))
        sections_by_code = dict([(section['code'], section)
                                 for section in self.sections])
        for code, grade in data.items():
            section = sections_by_code.get(code)['section']
            if section is not None:
                activities = IActivities(section)
                sheet = activities.get(UCE_SHEET_ID)
                if sheet is not None:
                    activity = sheet.get(UCE_SHEET_ID)
                    if activity is not None:
                        gradebook = IGradebook(sheet)
                        gradebook.evaluate(
                            self.context, activity, grade, evaluator)
        result = IUCEResult(self.context)
        try:
            result.division = int(self.request.get('division'))
        except:
            pass
        try:
            result.aggregate = int(self.request.get('aggregate'))
        except:
            pass

    def update(self):
        if 'EDIT' in self.request:
            params = {
                'schoolyear_id': self.schoolyear.__name__,
                'data': self.request.get('original_data'),
                'aggregate': self.request.get('aggregate'),
                'division': self.request.get('division'),
            }
            url = '%s/uce_input.html?%s' % (
                absoluteURL(self.context, self.request),
                urlencode(params))
            self.request.response.redirect(url)
            return
        errors, data = self.extract_data()
        if 'CONFIRM' in self.request:
            self.save(data)
            students = IGroupContainer(self.schoolyear)['students']
            url = '%s/uce_exam.html?schoolyear_id=%s' % (
                absoluteURL(students, self.request),
                self.schoolyear.__name__)
            self.request.response.redirect(url)
            return
        items = sorted(data.items(), key=lambda i: int(i[0]))
        self.data = OrderedDict(items)
        if errors:
            self.errors = errors

    @Lazy
    def warnings(self):
        result = []
        errors, data = self.extract_data()
        bestscore = sum(sorted([int(v) for v in data.values()])[:8])
        try:
            aggregate = int(self.request.get('aggregate'))
        except:
            aggregate = None
        if aggregate != bestscore:
            result.append(
                translate(
                    _("Submitted aggregate value of ${submitted} "
                      "doesn't match calculated value of ${value}",
                      mapping={'submitted': self.request.get('aggregate'),
                               'value': bestscore}),
                    context=self.request))
        try:
            division = int(self.request.get('division'))
        except:
            division = None
        if division != self.division:
            result.append(
                translate(
                    _("Submitted division value of ${submitted} "
                      "doesn't match calculated value of ${value}",
                      mapping={'submitted': self.request.get('division'),
                               'value': self.division}),
                    context=self.request))
        return result

    def extract_data(self):
        result = {}
        sections_by_code = dict([(section['code'], section)
                                 for section in self.sections])
        errors = []
        scoresystem = get_peas_scoresystem()
        decoder = flourish.tal.JSONDecoder()
        data = decoder.decode(self.request['data'])
        if not data:
            errors.append(
                translate(_('No data was submitted'),
                          context=self.request))
        for code, grade in data.items():
            error = False
            if isinstance(grade, list):
                errors.append(
                    translate(
                        _('Subject ${code} is repeated',
                          mapping={'code': code}),
                        context=self.request))
                continue
            if code and code not in sections_by_code:
                errors.append(
                    translate(
                        _('Student is not enrolled in subject ${code}',
                          mapping={'code': code}),
                        context=self.request))
                error = True
            if grade not in scoresystem.scoresDict:
                errors.append(
                    translate(
                        _('Invalid grade for subject ${code}',
                          mapping={'code': code}),
                        context=self.request))
                error = True
            if not error:
                result[code] = grade
        return errors, result

    def courses(self):
        result = []
        sections_by_code = dict([(section['code'], section['course'].title)
                                 for section in self.sections])
        for code, grade in self.data.items():
            result.append({
                'code': code,
                'title': sections_by_code[code],
                'score': grade,
            })
        return result


class UceJSONSearchView(GroupJSONSearchView):

    def encode(self, item):
        label = '%s, %s' % (item.title, item.username)
        return {
            'label': label,
            'value': label,
            'url': self.get_url(item),
        }

    def get_url(self, item):
        return '%s/uce_input.html?schoolyear_id=%s' % (
            absoluteURL(item, self.request),
            ISchoolYear(self.context.__parent__).__name__)


class RequestReportCardView(RequestRemoteReportDialog):

    report_builder = 'peas_report_card.pdf'

    fields = field.Fields(ICurrentTerm)

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)


class ReportCardPDFView(flourish.report.PlainPDFPage):

    content_template = flourish.templates.XMLFile('rml/report_card.pt')

    name = _('Report Card')
    title_column_width = 7 * units.cm

    @property
    def title(self):
        return self.context.title

    @property
    def subtitles_left(self):
        return [
            STMessage(u'Username: ${username}',
                      mapping={'username': self.context.username})
        ]

    @Lazy
    def schoolyear(self):
        return ISchoolYear(self.term)

    @Lazy
    def term(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear in ISchoolYearContainer(app).values():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return term

    def scope(self):
        return '%s | %s' % (self.term.title, self.schoolyear.title)

    @property
    def base_filename(self):
        return 'peas_report_card_%s' % self.context.__name__

    @property
    def message_title(self):
        return _('${student} report card',
                 mapping={'student': self.context.title})

    def column_widths(self, score_columns_count):
        widths = [self.title_column_width]
        doc_w, doc_h = self.page_size
        score_columns_available_width = (
            doc_w -
            self.margin.left -
            self.margin.right -
            self.title_column_width)
        score_columns_width = (
            float(score_columns_available_width) /
            score_columns_count)
        widths.extend([score_columns_width
                       for i in range(score_columns_count)])
        return ','.join(map(str, widths))

    def update(self):
        super(ReportCardPDFView, self).update()
        preferences = IAssessmentPreferences(self.schoolyear)
        names = ['bot', 'mot', 'eot']
        self.term_sheets = [name for name in names
                            if getattr(preferences, name)]
        self.colWidths = self.column_widths(len(self.term_sheets))

    @Lazy
    def learner_sections(self):
        sections = []
        r = Membership.bind(member=self.context).relationships
        for link_info in r:
            target = removeSecurityProxy(link_info.target)
            if (ISection.providedBy(target) and ITerm(target) == self.term):
                sections.append(target)
        return sections

    def rows(self):
        result = []
        for section in sorted(
                self.learner_sections,
                key=lambda s: ', '.join([c.title for c in s.courses])):
            cols = [', '.join([c.title for c in section.courses])]
            activities = IActivities(section)
            for sheet in self.term_sheets:
                grade = ''
                ws = activities.get(sheet)
                if ws is not None:
                    gb = IGradebook(ws)
                    total, average = gb.getWorksheetTotalAverage(
                        ws, self.context)
                    if (average is not None and
                        average is not UNSCORED):
                        grade = unicode('%.1f' % average)
                cols.append(grade)
            result.append(cols)
        return result


class UCEScoresViewlet(flourish.viewlet.Viewlet, UCEViewBase):

    template = ViewPageTemplateFile('templates/uce_scores.pt')

    def render(self, *args, **kw):
        result = IUCEResult(self.context)
        if (result.aggregate is None and result.division is None):
            return ''
        return self.template(*args, **kw)

    def grades(self):
        result = []
        for section_data in self.sections:
            section = section_data['section']
            course = section_data['course']
            code = section_data['code']
            activities = IActivities(section)
            ws = activities[UCE_SHEET_ID]
            activity = ws[UCE_SHEET_ID]
            gradebook = IGradebook(ws)
            grade = None
            score = gradebook.getScore(self.context, activity)
            if (score is not None and
                score is not UNSCORED and
                score.value is not None and
                score.value is not UNSCORED):
                grade = score.value
            if grade is not None:
                result.append({
                    'code': code,
                    'course': course.title,
                    'grade': grade,
                })
        return result


class SchoolYearReportLink(ReportLinkViewlet,
                           ActiveSchoolYearContentMixin):

    @property
    def report_link(self):
        url = absoluteURL(self.schoolyear, self.request)
        return '%s/%s' % (url, self.link)


class BasicClassroomReportLink(SchoolYearReportLink):

    @property
    def enabled(self):
        preferences = IAssessmentPreferences(self.schoolyear)
        return (preferences.bot is not None or
                preferences.mot is not None or
                preferences.eot is not None or
                preferences.mock is not None or
                preferences.mock_other is not None)

    def render(self, *args, **kw):
        if not self.enabled:
            return ''
        return super(BasicClassroomReportLink, self).render(*args, **kw)


class CoursesVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        schoolyear = self.context['schoolyear']
        courses = ICourseContainer(schoolyear).values()
        for course in sorted(courses, key=lambda c: c.title):
            result.append(
                self.createTerm(course, course.__name__, course.title))
        return result


def CoursesVocabularyFactory():
    return CoursesVocabulary


class TeachersVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        schoolyear = self.context['schoolyear']
        teachers = IGroupContainer(schoolyear)['teachers']
        for person in sorted(teachers.members,
                             key=lambda p: (p.last_name, p.first_name)):
            result.append(PersonTerm(person))
        return result


def TeachersVocabularyFactory():
    return TeachersVocabulary


class IBasicClassroomReport(Interface):

    level = Choice(
        title=STMessage('Level'),
        vocabulary='schooltool.level.level.levelvocabulary')

    group = Choice(
        title=_('Stream'),
        source='schooltool.peas.groups')

    course = Choice(
        title=STMessage('Course'),
        source='schooltool.peas.courses')


class BasicClassroomReportRequestView(RequestXLSReportDialog):

    report_builder = 'basic_classroom_report.xls'

    fields = field.Fields(IBasicClassroomReport)

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['schoolyear'] = self.context

    def updateTaskParams(self, task):
        level = self.form_params.get('level')
        course = self.form_params.get('course')
        group= self.form_params.get('group')
        task.request_params['level'] = level.__name__
        task.request_params['course'] = course.__name__
        if group is not u'':
            task.request_params['group'] = group.__name__

    def updateWidgets(self):
        super(BasicClassroomReportRequestView, self).updateWidgets()
        self.widgets['level'].prompt = True
        self.widgets['course'].prompt = True


def get_average(ws, student):
    gradebook = IGradebook(ws)
    total = 0
    count = 0
    for activity in gradebook.getWorksheetActivities(ws):
        score = gradebook.getScore(student, activity)
        if not score:
            continue
        ss = score.scoreSystem
        value = score.value
        if IDiscreteValuesScoreSystem.providedBy(ss):
            delta = 100.0 / (len(ss.scores) - 1)
            steps = None
            for i in range(len(ss.scores)):
                if ss.scores[i][2] == ss.getNumericalValue(value):
                    steps = i
                    break
            if steps is not None:
                total += delta * (len(ss.scores) - steps - 1)
                count += 100
        else:
            total += value
            count += ss.max
    if count:
        return Decimal(100 * total) / Decimal(count)


def convert_average(average):
    ss = get_peas_scoresystem()
    for score, abbr, value, percent in ss.scores:
        if average >= percent:
            return score


class AssessmentReportBase(object):

    def print_legend_3(self, ws, row=None, col=None):
        title = 'Key: arrows'
        labels = [
            (100, 90, 'Above target'),
            (90, 100, 'Below target'),
            (100, 100, 'On target'),
        ]
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write_merge(row, row, col, col + 1, title, style)

        row += 1
        for i, label_info in enumerate(labels):
            average, target, label = label_info
            arrow = get_arrow(average, target)
            if arrow == unichr(0x2191):
                color = 'green'
            elif arrow == unichr(0x2193):
                color = 'red'
            else:
                color = 'gray_ega'
            if label_info == labels[-1]:
                borders = self.translate_borders([0, 0, 1, 1])
            else:
                borders = self.translate_borders([0, 0, 0, 1])
            style = self.easyxf(
                'font: bold on, color %s; '
                'pattern: pattern solid, fore_color white; '
                'borders: %s; '
                'alignment: horizontal center; ' % (color, borders))
            ws.write(row + i, col, arrow, style)
            if label_info == labels[-1]:
                borders = self.translate_borders([0, 1, 1, 0])
            else:
                borders = self.translate_borders([0, 1, 0, 0])
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: %s; ' % borders,
                num_format_str='0')
            ws.write(row + i, col + 1, label, style)

    def print_legend_1(self, ws, row=None, col=None):
        title = 'KEY: Colour coding'
        labels = [
            (_('Over 75%'), '1 and 2', 80),
            (_('65-74%'), '3 and 4', 70),
            (_('50-64%'), '5 and 6', 60),
            (_('40-49%'), '7 and 8', 45),
            (_('Under 40%'), '9', 0),
        ]
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write_merge(row, row, col, col + 1, title, style)

        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left medium; '
        )
        ws.write(row, col, '%', style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left no_line; '
        )
        ws.write(row, col + 1, 'UCE eqv.', style)

        row += 1
        for i, label_info in enumerate(labels):
            label_1, label_2, value = label_info
            background = self.get_background_color(value)
            if label_info == labels[-1]:
                borders = self.translate_borders([0, 0, 1, 1])
            else:
                borders = self.translate_borders([0, 0, 0, 1])
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row + i, col, label_1, style)
            if label_info == labels[-1]:
                borders = self.translate_borders([0, 1, 1, 0])
            else:
                borders = self.translate_borders([0, 1, 0, 0])
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row + i, col + 1, label_2, style)
        return row + len(labels)

    @Lazy
    def custom_colors(self):
        result = super(AssessmentReportBase, self).custom_colors
        result['summary'] = (30, 31, 73, 125)
        result['a_above75'] = (31, 0, 176, 80)
        result['a_above65'] = (32, 146, 208, 80)
        result['a_above50'] = (33, 255, 255, 0)
        result['a_above40'] = (34, 255, 153, 51)
        result['a_below40'] = (35, 255, 0, 0)
        return result

    def adjust_weights(self, enabled_sheets):
        result = OrderedDict(enabled_sheets)        
        if not filter(lambda x: x is not None, enabled_sheets.values()):
            same_weight = 1.0 / len(enabled_sheets)
            for sheet in enabled_sheets:
                result[sheet] = same_weight
        total_weight = sum(result.values())
        for sheet in enabled_sheets:
            adjusted_weight = result[sheet] / total_weight
            result[sheet] = float(adjusted_weight)
        return result

    @Lazy
    def enabled_sheets(self):
        preferences = IAssessmentPreferences(self.schoolyear)
        enabled_sheets = OrderedDict()
        for sheet in ['bot', 'mot', 'eot']:
            if getattr(preferences, sheet):
                weight = getattr(preferences, '%s_weight' % sheet)
                enabled_sheets[sheet] = weight
        return self.adjust_weights(enabled_sheets)

    def get_background_color(self, value):
        if value >= 75:
            return 'a_above75'
        elif value >= 65:
            return 'a_above65'
        elif value >= 50:
            return 'a_above50'
        elif value >= 40:
            return 'a_above40'
        else:
            return 'a_below40'

    def get_discrete_background_color(self, value):
        if 1 <= value <= 2:
            return 'a_above75'
        elif 3 <= value <= 4:
            return 'a_above65'
        elif 5 <= value <= 6:
            return 'a_above50'
        elif 7 <= value <= 8:
            return 'a_above40'
        else:
            return 'a_below40'

    @Lazy
    def enabled_mocks(self):
        preferences = IAssessmentPreferences(self.schoolyear)
        enabled_mocks = OrderedDict()
        for sheet in ['mock', 'mock-other']:
            # XXX: needed because of a typo in the MOCK - Other sheet __name__
            preference = sheet.replace('-', '_')
            if getattr(preferences, preference):
                enabled_mocks[sheet] = getattr(preferences,
                                               '%s_term' % preference)
        return enabled_mocks
                

class BasicClassroomReportView(AssessmentReportBase, OverallReportBase):

    message_title = _('Basic classroom report')

    sheet_1_title = _('Assessments - Classroom report for a specific subject')
    sheet_1_label = _('Basic classroom report')

    @Lazy
    def data(self):
        result = {}
        uce_term = default_deploy_term(self.context)
        for section in self.course.sections:
            activities = IActivities(section)
            term = ITerm(section)
            for student in section.members:
                if (self.level not in student.levels or
                    (self.group is not None and
                     student not in self.group.members)):
                    continue
                if student not in result:
                    result[student] = {
                        'last_name': student.last_name,
                        'first_name': student.first_name,
                        'gender': student.gender,
                    }
                student_info = result[student]
                if term not in student_info:
                    student_info[term] = {}
                student_term_info = student_info[term]
                for sheet in self.enabled_sheets:
                    ws = activities.get(sheet)
                    if ws is not None:
                        gradebook = IGradebook(ws)
                        total, average = gradebook.getWorksheetTotalAverage(
                            ws, student)
                        if average is not UNSCORED:
                            average = float(average)
                        student_term_info[sheet] = average
                for sheet, mock_term in self.enabled_mocks.items():
                    if mock_term == ITerm(section):
                        ws = activities.get(sheet)
                        if ws is not None:
                            gradebook = IGradebook(ws)
                            total, average = gradebook.getWorksheetTotalAverage(
                                ws, student)
                            if average is not UNSCORED:
                                average = float(average)
                            student_term_info[sheet] = average
                if uce_term == ITerm(section):
                    sheet = UCE_SHEET_ID
                    ws = activities.get(sheet)
                    if ws is not None:
                        activity = ws.get(UCE_SHEET_ID)
                        if activity is not None:
                            gradebook = IGradebook(ws)
                            score = gradebook.getScore(student, activity)
                            if score:
                                ss = score.scoreSystem
                                value = ss.getNumericalValue(score.value)
                                student_term_info[sheet] = float(value)
                    sheet = TARGET_SHEET_ID
                    ws = activities.get(sheet)
                    if ws is not None:
                        average = get_average(ws, student)
                        if average is not None:
                            student_term_info[sheet] = float(average)
        return sorted(
            result.values(),
            key=lambda info: (info['last_name'], info['first_name']))

    def addExporters(self, progress):
        super(BasicClassroomReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Basic classroom report sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'basic_classroom_report_%s' % self.schoolyear.title

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def level(self):
        app = ISchoolToolApplication(None)
        container = ILevelContainer(app)
        return container.get(self.request.get('level'))

    @Lazy
    def course(self):
        container = ICourseContainer(self.context)
        return container.get(self.request.get('course'))

    @Lazy
    def group(self):
        groups = IGroupContainer(self.schoolyear)
        return groups.get(self.request.get('group'))

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, STMessage('Level'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.level.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, STMessage('Course'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.course.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Stream'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.group.title if self.group else 'All', style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    @Lazy
    def terms(self):
        return sorted(self.schoolyear.values(),
                      key=lambda term: term.first)

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)

        style = self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color report_title_bg; '
        )
        row = 8
        col = 2
        merge = len(self.terms) - 1
        ws.write_merge(row, row, col, col + merge, _('Summary'), style)

        col += merge + 2

        for term in self.terms:
            merge = len(self.enabled_sheets)
            ws.write_merge(row, row, col, col + merge, term.title,
                           self.styles['table_title'])
            col += merge + 2

        mock_titles = {
            'mock': 'MOCK - PEAS',
            'mock-other': 'MOCK - Other',
        }

        uce_term = default_deploy_term(self.context)
        show_mocks = self.level.__name__.lower() in ['s4']
        if show_mocks:
            for mock in self.enabled_mocks:
                ws.write(row, col, mock_titles[mock], self.styles['table_title'])
                style = self.easyxf(
                    'alignment: horizontal center; '
                )
                ws.write(row + 1, col, '%', style)
                col += 1

            if self.enabled_mocks:
                col += 1

            ws.write_merge(row, row, col, col + 1,
                           'S4 - actual UCE', self.styles['table_title'])
            style = self.easyxf(
                'font: bold on; '
                'alignment: horizontal center; '
            )
            ws.write(row + 1, col, 'Result', style)
            ws.write(row + 1, col + 1, 'Target', style)

        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Last name'), style)
        ws.write(row, 1, _('First name'), style)

        col = 2
        for term in self.terms:
            ws.write(row, col, term.title, self.styles['column_header'])
            style = self.easyxf(
                'alignment: horizontal center; '
            )
            ws.write(row + 1, col, '%', style)
            col += 1

        col += 1

        for term in self.terms:
            for sheet in self.enabled_sheets:
                ws.write(row, col, sheet.upper(), self.styles['column_header'])
                style = self.easyxf(
                    'alignment: horizontal center; '
                )
                ws.write(row + 1, col, '%', style)
                col += 1
            ws.write(row, col, _('Term ave'), self.styles['column_header'])
            style = self.easyxf(
                'alignment: horizontal center; '
            )
            ws.write(row + 1, col, '%', style)
            col += 2            

        row += 2
        borders = self.translate_borders([0, 0, 0, 0])
        for student_info in self.data:
            col = 0
            ws.write(row, col, student_info['last_name'])
            col += 1
            ws.write(row, col, student_info['first_name'])
            col += 1
            for term in self.terms:                
                value = self.term_average(student_info,
                                          term,
                                          self.enabled_sheets)
                if value is not None:
                    background = self.get_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
            col += 1
            for term in self.terms:
                for sheet in self.enabled_sheets:
                    value = self.sheet_average(student_info, term, sheet)
                    if value is UNSCORED:
                        style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'alignment: horizontal center; '
                        'borders: %s' % (borders),
                        num_format_str='0')
                        ws.write(row, col, '', style)
                    else:
                        background = self.get_background_color(value)
                        style = self.easyxf(
                            'font: bold on, color black; '
                            'pattern: pattern solid, fore_color %s; '
                            'alignment: horizontal center; '
                            'borders: %s' % (background, borders),
                            num_format_str='0')
                        ws.write(row, col, value, style)
                    col += 1
                value = self.term_average(student_info,
                                          term,
                                          self.enabled_sheets)
                if value is not None:
                    background = self.get_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                
                col += 2

            if show_mocks:
                for sheet, mock_term in self.enabled_mocks.items():
                    value = self.sheet_average(student_info, mock_term, sheet)
                    if value is UNSCORED:
                        style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'alignment: horizontal center; '
                        'borders: %s' % (borders),
                        num_format_str='0')
                        ws.write(row, col, '', style)
                    else:
                        background = self.get_background_color(value)
                        style = self.easyxf(
                            'font: bold on, color black; '
                            'pattern: pattern solid, fore_color %s; '
                            'alignment: horizontal center; '
                            'borders: %s' % (background, borders),
                            num_format_str='0')
                        ws.write(row, col, value, style)
                    col += 1

                if self.enabled_mocks:
                    col += 1

                value = self.sheet_average(student_info, uce_term, UCE_SHEET_ID)
                if value is not None:
                    background = self.get_discrete_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)

                col += 1
                value = self.sheet_average(
                    student_info, uce_term, TARGET_SHEET_ID)
                if value is not None:
                    background = self.get_background_color(value)
                    value = convert_average(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)

            row += 1

        row += 1
        is_female = lambda student_info: student_info['gender'] == 'female'
        is_male = lambda student_info: student_info['gender'] == 'male'

        classroom_rows = [
            (_('Classroom ave'), lambda student_info: True),
            (_('Girls ave'), is_female),
            (_('Boys ave'), is_male),
        ]

        for row_info in classroom_rows:
            col = 0
            title, filter_func = row_info
            ws.write(row, col, title, self.styles['table_title'])
            col = 2
            for term in self.terms:
                value = self.classroom_term_average(self.data,
                                                    term,
                                                    self.enabled_sheets,
                                                    filter_func)
                if value is not None:
                    background = self.get_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1

            col += 1
            for term in self.terms:
                for sheet in self.enabled_sheets:
                    value = self.classroom_sheet_average(self.data,
                                                         term,
                                                         sheet,
                                                         filter_func)
                    if value is not None:
                        background = self.get_background_color(value)
                    else:
                        background = 'white'
                    style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color %s; '
                        'alignment: horizontal center; '
                        'borders: %s' % (background, borders),
                        num_format_str='0')
                    ws.write(row, col, value, style)
                    col += 1
                value = self.classroom_term_average(self.data,
                                                    term,
                                                    self.enabled_sheets,
                                                    filter_func)
                if value is not None:
                    background = self.get_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 2

            for sheet, mock_term in self.enabled_mocks.items():
                value = self.classroom_sheet_average(self.data,
                                                     mock_term,
                                                     sheet,
                                                     filter_func)
                if value is not None:
                    background = self.get_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1

            row += 1

        row += 1
        self.print_legend_1(ws, row, 0)

        self.finish('sheet1')

    def term_average(self, student_info, term, sheets):
        term_grades = student_info[term]
        averages = {}
        for sheet in sheets:
            average = term_grades.get(sheet)
            if average not in [UNSCORED, None]:
                averages[sheet] = average
        if averages:
            total = 0
            for sheet, sheet_average in averages.items():
                total += sheet_average * sheets[sheet]
            return total

    def sheet_average(self, student_info, term, sheet):
        return student_info[term].get(sheet)

    def classroom_term_average(self, students_info, term, sheets, filter_func):
        averages = []
        count = 0
        for student_info in students_info:
            if filter_func(student_info):
                average = self.term_average(student_info, term, sheets)
                if average is not None:
                    count += 1
                    averages.append(average)
        if count:
            return self.div(sum(averages), count)

    def classroom_sheet_average(self, students_info, term, sheet, filter_func):
        averages = []
        count = 0
        for student_info in students_info:
            if filter_func(student_info):
                average = self.sheet_average(student_info, term, sheet)
                if average not in [UNSCORED, None]:
                    count += 1
                    averages.append(average)
        if count:
            return self.div(sum(averages), count)


class IStudentReport(Interface):

    level = Choice(
        title=STMessage('Level'),
        vocabulary='schooltool.level.level.levelvocabulary')

    group = Choice(
        title=_('Stream'),
        source='schooltool.peas.groups')

    term = Choice(
        title=_('Term / Exam'),
        source='schooltool.peas.term_exams')


class StudentReportRequestView(RequestXLSReportDialog):

    report_builder = 'student_report.xls'

    fields = field.Fields(IStudentReport)

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['schoolyear'] = self.context

    def updateTaskParams(self, task):
        level = self.form_params.get('level')
        group= self.form_params.get('group')
        term = self.form_params.get('term')
        task.request_params['level'] = level.__name__
        if group is not u'':
            task.request_params['group'] = group.__name__
        preferences = IAssessmentPreferences(self.context)
        if term == MOCK_SHEET_ID:
            term = preferences.mock_term
            sheet = MOCK_SHEET_ID
        elif term == MOCK_OTHER_SHEET_ID:
            term = preferences.mock_other_term
            sheet = MOCK_OTHER_SHEET_ID
        elif term == UCE_SHEET_ID:
            term = default_deploy_term(self.context)
            sheet = UCE_SHEET_ID
        else:
            sheet = ''
        task.request_params['term'] = term.__name__
        task.request_params['sheet'] = sheet

    def updateWidgets(self):
        super(StudentReportRequestView, self).updateWidgets()
        self.widgets['level'].prompt = True
        self.widgets['term'].prompt = True


class GroupsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        result.append(self.createTerm(
            u'',
            '__all_groups__',
            _('All'),
        ))
        schoolyear = self.context['schoolyear']
        groups = IGroupContainer(schoolyear)
        for group_id, group in sorted(groups.items(),
                                      key=lambda i: i[1].title):
            if group_id in defaultGroups:
                continue
            result.append(self.createTerm(group, group_id, group.title))
        return result


def GroupsVocabularyFactory():
    return GroupsVocabulary


class TermExamsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        schoolyear = self.context['schoolyear']
        for term in sorted(schoolyear.values(),
                           key=lambda t: t.first):
            result.append(self.createTerm(term, term.__name__, term.title))
        mocks = [MOCK_SHEET_ID, MOCK_OTHER_SHEET_ID]
        preferences = IAssessmentPreferences(schoolyear)
        for sheet in mocks + [UCE_SHEET_ID]:
            if sheet in mocks:
                # XXX: needed because of a typo in MOCK_OTHER_SHEET_ID
                preference = sheet.replace('-', '_')
                if not getattr(preferences, preference):
                    continue
            title = ASSESSMENT_SHEETS[sheet]['title']
            result.append(self.createTerm(sheet, sheet, title))
        return result


def TermExamsVocabularyFactory():
    return TermExamsVocabulary


COURSE_CODES = [
    '456',
    '112',
    '535',
    '545',
    '553',
    '273',
    '241',
    '527',
    '845',
    '223',
    '800',
    '810',
    '208',
    '610',
    '335',
]


def get_arrow(average, target):
    if average >= (target + 5):
        return unichr(0x2191)
    elif average <= (target - 5):
        return unichr(0x2193)
    return unichr(0x2194)


class StudentReportView(AssessmentReportBase, OverallReportBase):

    message_title = _('Student report')

    sheet_1_title = _('Assessment - Student attainment for all subjects '
                      'and comparison to target ')
    sheet_1_label = _('Student report')

    @property
    def title_merge(self):
        return (len([code for code in COURSE_CODES
                     if self.courses_by_code.get(code) is not None]) * 2) + 2

    @property
    def base_filename(self):
        return 'student_report_%s' % self.schoolyear.title

    def print_report_title(self, ws, title, row, merge):
        style = self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color report_title_bg; '
        )
        ws.write_merge(row, row, 0, self.title_merge, title, style)

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, STMessage('Level'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.level.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Stream'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.group.title if self.group else 'All', style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Term / Exam'), style)
        if not self.sheet:
            title = self.term.title
        else:
            title = ASSESSMENT_SHEETS[self.sheet]['title']
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    @Lazy
    def level(self):
        app = ISchoolToolApplication(None)
        container = ILevelContainer(app)
        return container.get(self.request.get('level'))

    @Lazy
    def group(self):
        groups = IGroupContainer(self.schoolyear)
        return groups.get(self.request.get('group'))

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def term(self):
        for term in self.schoolyear.values():
            if term.__name__ == self.request.get('term'):
                return term

    @Lazy
    def sheet(self):
        return self.request.get('sheet') or None

    @Lazy
    def courses_by_code(self):
        courses = ICourseContainer(self.context)
        return dict([(course.government_id, course)
                     for course in courses.values()])

    @Lazy
    def data(self):
        result = {}
        rel = StudentLevels.bind(level=self.level).relationships
        uce_results = {}
        for link_info in rel:
            student = link_info.target
            if self.group is not None and student not in self.group.members:
                continue
            if student not in result:
                result[student] = [
                    student.last_name,
                    student.first_name,
                    student.gender,
                ]
            if student not in uce_results:
                uce_results[student] = [
                    student.last_name,
                    student.first_name,
                    IUCEResult(student).aggregate,
                    IUCEResult(student).division,
                ]
            student_info = result[student]
            for course_code in COURSE_CODES:
                course = self.courses_by_code.get(course_code)
                course_average = None
                arrow = None
                target_section = None
                if course is not None:
                    for section in course.sections:
                        if student not in section.members:
                            continue
                        term = ITerm(section)
                        if term == default_deploy_term(self.context):
                            target_section = section
                        activities = IActivities(section)
                        if term == self.term:
                            if self.sheet is not None:
                                sheets = [self.sheet]
                            else:
                                sheets = self.enabled_sheets
                            averages = {}
                            for sheet in sheets:
                                ws = activities.get(sheet)
                                if ws is not None:
                                    gradebook = IGradebook(ws)
                                    if UCE_SHEET_ID in sheets:
                                        activity = ws.get(UCE_SHEET_ID)
                                        if activity is not None:
                                            score = gradebook.getScore(student, activity)
                                            if score:
                                                ss = score.scoreSystem
                                                for s, abbr, value, percent in ss.scores:
                                                    if s == score.value:
                                                        averages[sheet] = float(percent)
                                    else:
                                        total, average = gradebook.getWorksheetTotalAverage(
                                            ws, student)
                                        if average is not UNSCORED:
                                            averages[sheet] = float(average)
                            if averages:
                                if sheets == self.enabled_sheets:
                                    total = 0
                                    for sheet, sheet_average in averages.items():
                                        total += sheet_average * sheets[sheet]
                                    course_average = total
                                else:
                                    course_average = sum(averages.values())
                    if target_section is not None:
                        activities = IActivities(target_section)
                        sheet = TARGET_SHEET_ID
                        ws = activities.get(sheet)
                        if ws is not None:
                            target = get_average(ws, student)
                            if target is not None:
                                if course_average is not None:
                                    arrow = get_arrow(course_average, target)
                    student_info.append(course_average)
                    student_info.append(arrow)
        self.uce_results = sorted(
            uce_results.values(),
            key=lambda info: (info[0], info[1]))
        return sorted(
            result.values(),
            key=lambda info: (info[0], info[1]))

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)

        courses_count = len([code for code in COURSE_CODES
                             if self.courses_by_code.get(code) is not None])

        row = 8
        col = 3
        ws.write_merge(row, row, col, col + (courses_count * 2) - 1,
                       'Percentage score', self.styles['table_title'])

        uce_column = 4 + (courses_count * 2)
        show_uce_table = self.level.__name__.lower() in ['s3', 's4']
        if show_uce_table:
            ws.write_merge(row, row,
                           uce_column, uce_column + courses_count + 1,
                           'UCE equivalent grade (for S3 and S4 students only)',
                           self.styles['table_title'])

        row = 9
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )
        ws.write(row, 0, STMessage('Last name'), style)
        ws.write(row, 1, _('First name'), style)
        ws.write(row, 2, _('Gender'), style)
        col = 3
        for course_code in COURSE_CODES:
            course = self.courses_by_code.get(course_code)
            if course is None:
                continue
            title = course.course_id or course.goverment_id
            ws.write_merge(row, row, col, col + 1,
                           title,
                           self.styles['column_header'])
            col += 2
        if show_uce_table:
            ws.write(row, uce_column, 'Result', self.styles['column_header'])
            ws.write(row, uce_column + 1, 'Aggregate score',
                     self.styles['column_header'])
            col = uce_column + 2
            for course_code in COURSE_CODES:
                course = self.courses_by_code.get(course_code)
                if course is None:
                    continue
                title = course.course_id or course.goverment_id
                ws.write(row, col, title, self.styles['column_header'])
                col += 1

        row += 1
        uce_row = 0
        for student_info in self.data:
            for i, col in enumerate(student_info):
                if i < 3:
                    ws.write(row, i, col)
                else:
                    if i % 2 == 0:
                        if col == unichr(0x2191):
                            color = 'green'
                        elif col == unichr(0x2193):
                            color = 'red'
                        else:
                            color = 'gray_ega'
                        style = self.easyxf(
                            'font: bold on, color %s; '
                            'pattern: pattern solid, fore_color white; '
                            'alignment: horizontal center; ' % color)
                    else:
                        background = 'white'
                        if col is not None:
                            background = self.get_background_color(col)
                        style = self.easyxf(
                            'font: bold on, color black; '
                            'pattern: pattern solid, fore_color %s; '
                            'alignment: horizontal center; ' % background,
                            num_format_str='0')
                    ws.write(row, i, col, style)
            if show_uce_table:
                codes = [code for code in COURSE_CODES
                         if self.courses_by_code.get(code) is not None]
                grades = [convert_average(grade)
                          for i, grade in enumerate(student_info[3:])
                          if i % 2 == 0]
                items = zip(codes, grades)
                if (self.level.__name__.lower() == 's4' and
                    self.sheet == UCE_SHEET_ID):
                    division = self.uce_results[uce_row][3]
                    aggregate = self.uce_results[uce_row][2]
                else:
                    division = find_division(items)
                    aggregate = best8(dict([(k,int(v)) for k,v in items
                                            if v is not None]))
                col = uce_column
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'alignment: horizontal center; ')
                ws.write(row, col, division, style)
                col = uce_column + 1
                ws.write(row, col, aggregate, style)
                col = uce_column + 2
                for i, grade in enumerate(grades):
                    if grade is not None:
                        value = int(grade)
                        background = self.get_discrete_background_color(int(grade))
                    else:
                        value = ''
                        background = 'white'
                    style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color %s; '
                        'alignment: horizontal center; ' % background,
                        num_format_str='0')
                    ws.write(row, col + i, value, style)
            row += 1
            uce_row += 1

        row += 1
        row = self.print_legend_1(ws, row, 0)
        row += 1
        self.print_legend_3(ws, row, 0)

        self.finish('sheet1')


class OverallSchoolReportRequestView(StudentReportRequestView):

    report_builder = 'overall_school_report.xls'


class OverallSchoolReportView(StudentReportView):

    message_title = _('Overall school results')

    sheet_1_title = _('Assessment - Overall school results')
    sheet_1_label = _('Overall school results')

    @property
    def title_merge(self):
        return len([code for code in COURSE_CODES
                    if self.courses_by_code.get(code) is not None])

    @property
    def base_filename(self):
        return 'overall_school_report_%s' % self.schoolyear.title

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        row = 8
        courses_count = len([code for code in COURSE_CODES
                             if self.courses_by_code.get(code) is not None])
        self.print_legend_2(ws, row, courses_count + 2)
        ws.write_merge(row, row,
                       1, courses_count,
                       _('Overall subject results'),
                       self.styles['table_title'])
        row += 1
        ws.write(row, 0, _('Students achieving...'),
                 self.styles['column_header'])
        col = 1
        for course_code in COURSE_CODES:
            course = self.courses_by_code.get(course_code)
            if course is None:
                continue
            title = course.course_id or course.goverment_id
            ws.write(row, col, title, self.styles['column_header'])
            col += 1

        row += 1
        row_headers = [
            'D1 (80-100%)',
            'D2 (75-79%)',
            'C3 (70-74%)',
            'C4 (65-69%)',
            'C5 (60-64%)',
            'C6 (50%-59%)',
            'P7 (45%-49%)',
            'P8 (40%-44%)',
            'F9 (0%-39%)',
        ]

        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        filter_funcs = [
            lambda v: v >= 80,
            lambda v: 75 <= v < 80,
            lambda v: 70 <= v < 75,
            lambda v: 65 <= v < 70,
            lambda v: 60 <= v < 65,
            lambda v: 50 <= v < 60,
            lambda v: 45 <= v < 50,
            lambda v: 40 <= v < 45,
            lambda v: v < 40,
        ]

        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None]
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = self.div(len(filtered) * 100, len(grades))
                    background = self.get_gradient_color(value)
                else:
                    value = ''
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'borders: %s; '
                    'alignment: horizontal center; ' % (
                        background,
                        self.translate_borders([0, 1, 0, 1])),
                    num_format_str='0')
                ws.write(row + i, col + j, value, style)

        row = 20
        row_headers = [
            '% students above 70%',
            '% students above 50%',
            '% students 49% or less',
        ]

        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        average_borders = self.translate_borders([1, 0, 1, 0])
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: %s; ' % (average_borders))
        ws.write(row + len(row_headers), 0, 'Average %', style)

        filter_funcs = [
            lambda v: v > 70,
            lambda v: v > 50,
            lambda v: v <= 50,
        ]
        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None]
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = self.div(len(filtered) * 100, len(grades))
                    background = self.get_gradient_color(value)
                    style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color %s; '
                        'alignment: horizontal center; ' % background,
                        num_format_str='0')
                    ws.write(row + i, col + j, value, style)
            if grades:
                value = self.div(sum(grades), len(grades))
                background = self.get_gradient_color(value)
            else:
                value = ''
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s; '
                'alignment: horizontal center; ' % (background,
                                                    average_borders),
                num_format_str='0')
            ws.write(row + i + 1, col + j, value, style)

        row = 25
        row_headers = [
            '# students that took exam',
            '# students above 70%',
            '# students above 50%',
            '# students 49% or less',
        ]
        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        filter_funcs = [
            lambda v: True,
            lambda v: v > 70,
            lambda v: v > 50,
            lambda v: v <= 50,
        ]

        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None]
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = len(filtered)
                    ws.write(row + i, col + j, value)

        row = 30
        ws.write_merge(row, row,
                       0, len([code for code in COURSE_CODES
                               if self.courses_by_code.get(code) is not None]),
                       'Girls', self.styles['table_title'])
        row += 1
        row_headers = [
            '% girls above 70%',
            '% girls above 50%',
            '% girls 49% or less',
        ]
        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: %s; ' % (average_borders))
        ws.write(row + len(row_headers), 0, 'Average % for girls', style)

        filter_funcs = [
            lambda v: v > 70,
            lambda v: 50 < v <= 70,
            lambda v: v <= 50,
        ]

        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None and r[2] == 'female']
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = self.div(len(filtered) * 100, len(grades))
                    background = self.get_gradient_color(value)
                    style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color %s; '
                        'alignment: horizontal center; ' % background,
                        num_format_str='0')
                    ws.write(row + i, col + j, value, style)
            if grades:
                value = self.div(sum(grades), len(grades))
                background = self.get_gradient_color(value)
            else:
                value = ''
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s; '
                'alignment: horizontal center; ' % (background,
                                                    average_borders),
                num_format_str='0')
            ws.write(row + i + 1, col + j, value, style)

        row = 36
        row_headers = [
            '# girls that took exam',
            '# girls above 70%',
            '# girls above 50%',
            '# girls 49% or less',
        ]
        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        filter_funcs = [
            lambda v: True,
            lambda v: v > 70,
            lambda v: 50 < v <= 70,
            lambda v: v <= 50,
        ]

        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None and r[2] == 'female']
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = len(filtered)
                    ws.write(row + i, col + j, value)

        row = 41
        ws.write_merge(row, row,
                       0, len([code for code in COURSE_CODES
                               if self.courses_by_code.get(code) is not None]),
                       'Boys', self.styles['table_title'])
        row += 1
        row_headers = [
            '% boys above 70%',
            '% boys above 50%',
            '% boys 49% or less',
        ]
        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: %s; ' % (average_borders))
        ws.write(row + len(row_headers), 0, 'Average % for boys', style)

        filter_funcs = [
            lambda v: v > 70,
            lambda v: 50 < v <= 70,
            lambda v: v <= 50,
        ]

        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None and r[2] == 'male']
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = self.div(len(filtered) * 100, len(grades))
                    background = self.get_gradient_color(value)
                    style = self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color %s; '
                        'alignment: horizontal center; ' % background,
                        num_format_str='0')
                    ws.write(row + i, col + j, value, style)
            if grades:
                value = self.div(sum(grades), len(grades))
                background = self.get_gradient_color(value)
            else:
                value = ''
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s; '
                'alignment: horizontal center; ' % (background,
                                                    average_borders),
                num_format_str='0')
            ws.write(row + i + 1, col + j, value, style)

        row = 47
        row_headers = [
            '# boys that took exam',
            '# boys above 70%',
            '# boys above 50%',
            '# boys 49% or less',
        ]
        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header)

        filter_funcs = [
            lambda v: True,
            lambda v: v > 70,
            lambda v: 50 < v <= 70,
            lambda v: v <= 50,
        ]

        col = 1
        for j, course_code in enumerate([
                code for code in COURSE_CODES
                if self.courses_by_code.get(code) is not None]):
            grades = [r[3 + (j * 2)] for r in self.data
                      if r[3 + (j * 2)] is not None and r[2] == 'male']
            for i, filter_func in enumerate(filter_funcs):
                filtered = [grade for grade in grades if filter_func(grade)]
                if grades:
                    value = len(filtered)
                    ws.write(row + i, col + j, value)

        show_division = (self.level.__name__.lower() in ['s4'] and
                         self.sheet in (MOCK_SHEET_ID, MOCK_OTHER_SHEET_ID,
                                        UCE_SHEET_ID))
        if show_division:
            division_data = []
            for student_info in self.data:
                grades = {}
                for i, course_code in enumerate([
                        code for code in COURSE_CODES
                        if self.courses_by_code.get(code) is not None]):
                    grade = student_info[3 + (i * 2)]
                    if grade is None:
                        continue
                    grades[course_code] = convert_average(grade)
                division = find_division(grades.items())
                division_data.append((student_info[2], division))

            row = 52
            merge = 7
            title = 'Overall division results for this year'
            ws.write_merge(row, row, 0, merge,
                           title, self.styles['table_title'])
            row += 1
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            )
            ws.write(row, 0, 'Students achieving:', style)
            ws.write(row, 1, 'Total (%)', self.styles['column_header'])
            ws.write(row, 2, 'Girls (%)', self.styles['column_header'])
            ws.write(row, 3, 'Boys (%)', self.styles['column_header'])
            ws.write(row, 5, 'Total (#)', self.styles['column_header'])
            ws.write(row, 6, 'Girls (#)', self.styles['column_header'])
            ws.write(row, 7, 'Boys (#)', self.styles['column_header'])

            row += 1
            row_headers = [
                'Division 1',
                'Division 2',
                'Division 3',
                'Division 4',
                'Division 7',
                'Division 9',
                'X',
            ]
            for i, header in enumerate(row_headers):
                ws.write(row + i, 0, header, style)

            style = self.easyxf('', num_format_str='0')

            female = [r for r in division_data if r[0] == 'female']
            male = [r for r in division_data if r[0] == 'male']

            col = 1
            for i, division in enumerate([1, 2, 3, 4, 7, 9, 0]):
                division_students = [r[1] for r in division_data
                                     if r[1] == division]
                value = self.div(len(division_students) * 100,
                                 len(division_data))
                ws.write(row + i, col, value, style)

            col = 2
            for i, division in enumerate([1, 2, 3, 4, 7, 9, 0]):
                division_students = [r[1] for r in division_data
                                     if r[1] == division and r[0] == 'female']
                value = self.div(len(division_students) * 100,
                                 len(female))
                ws.write(row + i, col, value, style)

            col = 3
            for i, division in enumerate([1, 2, 3, 4, 7, 9, 0]):
                division_students = [r[1] for r in division_data
                                     if r[1] == division and r[0] == 'male']
                value = self.div(len(division_students) * 100,
                                 len(male))
                ws.write(row + i, col, value, style)

            col = 5
            for i, division in enumerate([1, 2, 3, 4, 7, 9, 0]):
                division_students = [r[1] for r in division_data
                                     if r[1] == division]
                value = len(division_students)
                ws.write(row + i, col, value, style)

            col = 6
            for i, division in enumerate([1, 2, 3, 4, 7, 9, 0]):
                division_students = [r[1] for r in division_data
                                     if r[1] == division and r[0] == 'female']
                value = len(division_students)
                ws.write(row + i, col, value, style)

            col = 7
            for i, division in enumerate([1, 2, 3, 4, 7, 9, 0]):
                division_students = [r[1] for r in division_data
                                     if r[1] == division and r[0] == 'male']
                value = len(division_students)
                ws.write(row + i, col, value, style)

            row = 62
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            )
            row_headers = [
                'Division 1 and 2',
                'Division 1, 2, and 3',
                'Division 1, 2, 3 and 4',
            ]
            for i, header in enumerate(row_headers):
                ws.write(row + i, 0, header, style)

            style = self.easyxf('', num_format_str='0')

            division_groups = [
                [1, 2],
                [1, 2, 3],
                [1, 2, 3, 4],
            ]

            for i, divisions in enumerate(division_groups):
                col = 1
                division_students = [r[1] for r in division_data
                                     if r[1] in divisions]
                value = self.div(len(division_students) * 100,
                                 len(division_data))
                ws.write(row + i, col, value, style)

                col = 2
                division_students = [r[1] for r in division_data
                                     if r[1] in divisions and r[0] == 'female']
                value = self.div(len(division_students) * 100,
                                 len(female))
                ws.write(row + i, col, value, style)

                col = 3
                division_students = [r[1] for r in division_data
                                     if r[1] in divisions and r[0] == 'male']
                value = self.div(len(division_students) * 100,
                                 len(male))
                ws.write(row + i, col, value, style)

                col = 5
                division_students = [r[1] for r in division_data
                                     if r[1] in divisions]
                value = len(division_students)
                ws.write(row + i, col, value, style)

                col = 6
                division_students = [r[1] for r in division_data
                                     if r[1] in divisions and r[0] == 'female']
                value = len(division_students)
                ws.write(row + i, col, value, style)

                col = 7
                division_students = [r[1] for r in division_data
                                     if r[1] in divisions and r[0] == 'male']
                value = len(division_students)
                ws.write(row + i, col, value, style)


            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            )

            row += (len(row_headers) + 1)
            ws.write(row, 0, 'Total', style)
            ws.write(row, 5, len(female) + len(male))
            ws.write(row, 6, len(female))
            ws.write(row, 7, len(male))

        self.finish('sheet1')


class UCEResultsReportRequestView(RequestXLSReportDialog):

    report_builder = 'uce_results_report.xls'


UCE_COURSE_CODES = [
    '112',
    '208',
    '223',
    '241',
    '273',
    '456',
    '535',
    '545',
    '553',
    '610',
    '335',
    '527',
    '800',
    '845',
    '810',
]


class UCEResultsReportView(StudentReportView):

    message_title = _('UCE Results table')

    sheet_1_label = _('UCE Results table')

    @property
    def base_filename(self):
        return 'uce_results_report_%s' % self.schoolyear.title

    @Lazy
    def level(self):
        app = ISchoolToolApplication(None)
        levels = ILevelContainer(app)
        for level in levels.values():
            if level.__name__.lower() == 's4':
                return level

    @Lazy
    def data(self):
        result = {}
        rel = StudentLevels.bind(level=self.level).relationships
        app = ISchoolToolApplication(None)
        preferences = IApplicationPreferences(app)
        for link_info in rel:
            student = link_info.target
            demographics = IDemographics(student)
            uce_result = IUCEResult(student)
            if student not in result:
                result[student] = {
                    'school_code': preferences.code,
                    'region': preferences.peas_region,
                    'year': demographics.get('enrollment_year'),
                    'gender': student.gender,
                    'boarding': demographics.get('day_boarding'),
                    'last_name': student.last_name,
                    'first_name': student.first_name,
                    'ple_score': demographics.get('ple_score'),
                    'uce_score': uce_result.aggregate,
                    'division': uce_result.division,
                    'obj': student,
                }
        return sorted(result.values(),
                      key=lambda i: (i['last_name'], i['first_name']))

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)

        row = 0
        headers = [
            'School Code',
            'PEAS Region',
            'Cohort',
            'Year Joined',
            'Gender',
            'Section',
            'Last name',
            'First name',
            'PLE score',
            'PLE: Div',
            'UCE Score',
            'Division',
            'ENG',
            'LIT',
            'CRE',
            'HIST',
            'GEO',
            'MTC',
            'PHY',
            'CHE',
            'BIO',
            'ART',
            'LUG',
            'AGRI',
            'COM',
            'ENT',
            'ACC',
            'Difference between PLE and UCE',
            'Check',
            'Check Statement',
        ]
        col = 0
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        uce_term = default_deploy_term(self.context)
        row += 1
        for student_info in self.data:
            col = 0
            ws.write(row, col, student_info['school_code'])
            col += 1
            ws.write(row, col, student_info['region'])
            col += 2
            ws.write(row, col, student_info['year'])
            col += 1
            ws.write(row, col, student_info['gender'])
            col += 1
            ws.write(row, col, student_info['boarding'])
            col += 1
            ws.write(row, col, student_info['last_name'])
            col += 1
            ws.write(row, col, student_info['first_name'])
            col += 1
            ws.write(row, col, student_info['ple_score'])
            col += 2
            ws.write(row, col, student_info['uce_score'])
            col += 1
            ws.write(row, col, student_info['division'])
            col += 1
            grades = {}
            for course_code in UCE_COURSE_CODES:
                course = self.courses_by_code.get(course_code)
                if course is not None:
                    for section in course.sections:
                        if ITerm(section) == uce_term:
                            if student_info['obj'] in section.members:
                                activities = IActivities(section)
                                worksheet = activities.get(UCE_SHEET_ID)
                                if worksheet is not None:
                                    activity = worksheet.get(UCE_SHEET_ID)
                                    if activity is not None:
                                        gradebook = IGradebook(worksheet)
                                        score = gradebook.getScore(student_info['obj'], activity)
                                        if score:
                                            ws.write(row, col, score.value)
                                            grades[course_code] = score.value
                col += 1
            col += 1
            try:
                bestscores = best8(dict([(k,int(v)) for k,v in grades.items()
                                         if v is not None]))
            except:
                bestscores = None
            ws.write(row, col, bestscores)
            col += 1
            try:
                uce_score =  int(student_info['uce_score'])
            except:
                uce_score = None
            ws.write(row, col, 'true' if bestscores == uce_score else 'false')
            row += 1

        self.finish('sheet1')


class IStudentDetailReport(Interface):

    level = Choice(
        title=STMessage('Level'),
        vocabulary='schooltool.level.level.levelvocabulary')

    group = Choice(
        title=_('Stream'),
        source='schooltool.peas.groups')

    term = Choice(
        title=_('Term'),
        source='schooltool.peas.year_terms')


class YearTermsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        schoolyear = self.context['schoolyear']
        for term in sorted(schoolyear.values(), key=lambda t: t.first):
            result.append(self.createTerm(term, term.__name__, term.title))
        return result


def YearTermsVocabularyFactory():
    return YearTermsVocabulary


class StudentDetailReportRequestView(RequestXLSReportDialog):

    report_builder = 'student_detail_report.xls'

    fields = field.Fields(IStudentDetailReport)

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['schoolyear'] = self.context

    def updateTaskParams(self, task):
        level = self.form_params.get('level')
        group= self.form_params.get('group')
        term = self.form_params.get('term')
        task.request_params['level'] = level.__name__
        if group is not u'':
            task.request_params['group'] = group.__name__
        task.request_params['term'] = term.__name__

    def updateWidgets(self):
        super(StudentDetailReportRequestView, self).updateWidgets()
        self.widgets['level'].prompt = True
        self.widgets['term'].prompt = True


class StudentDetailReportView(StudentReportView):

    message_title = _('Student - MORE DETAIL report')

    sheet_1_title = _('Assessment - Student attainment across subjects - '
                      'additional detail')
    sheet_1_label = _('Student - MORE DETAIL')

    sheet = None

    @property
    def title_merge(self):
        return (len([code for code in COURSE_CODES
                    if self.courses_by_code.get(code) is not None]) *
                len(self.enabled_sheets)) + 2

    @property
    def base_filename(self):
        return 'student_detail_report_%s' % self.schoolyear.title

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, STMessage('Level'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.level.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Stream'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.group.title if self.group else 'All', style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Term'), style)
        if not self.sheet:
            title = self.term.title
        else:
            title = ASSESSMENT_SHEETS[self.sheet]['title']
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    @Lazy
    def data(self):
        result = {}
        rel = StudentLevels.bind(level=self.level).relationships
        for link_info in rel:
            student = link_info.target
            if self.group is not None and student not in self.group.members:
                continue
            if student not in result:
                result[student] = {
                    'last_name': student.last_name,
                    'first_name': student.first_name,
                    'gender': student.gender,
                }
            student_info = result[student]
            for course_code in COURSE_CODES:
                course = self.courses_by_code.get(course_code)
                if course is not None:
                    for section in course.sections:
                        if student not in section.members:
                            continue
                        if course_code not in student_info:
                            student_info[course_code] = {}
                        course_student_info = student_info[course_code]
                        term = ITerm(section)
                        if term not in course_student_info:
                            course_student_info[term] = {}
                        student_term_info = course_student_info[term]
                        activities = IActivities(section)
                        sheets = self.enabled_sheets
                        for sheet in sheets:
                            ws = activities.get(sheet)
                            if ws is not None:
                                gradebook = IGradebook(ws)
                                total, average = gradebook.getWorksheetTotalAverage(
                                    ws, student)
                                if average is not UNSCORED:
                                    average = float(average)
                                student_term_info[sheet] = average
        return sorted(
            result.values(),
            key=lambda info: (info['last_name'], info['first_name']))

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        row = 8
        col = 0
        title = 'Term break down by student (selected term shown)'
        merge = (len([code
                      for code in COURSE_CODES
                      if self.courses_by_code.get(code) is not None]) *
                 len(self.enabled_sheets)) + 2
        ws.write_merge(row, row, col, col + merge,
                       title, self.styles['table_title'])
        row += 2
        col = 3
        merge = len(self.enabled_sheets) - 1
        for i, course_code in enumerate(COURSE_CODES):
            course = self.courses_by_code.get(course_code)
            if course is None:
                continue
            title = course.title
            ws.write_merge(row, row, col, col + merge,
                           title, self.styles['column_header'])
            col += merge + 1

        row += 1
        col = 0
        headers = [
            'Last name',
            'First name',
            'Gender',
        ]
        for header in headers:
            ws.write(row, col, header, self.styles['column_header'])
            col += 1

        for course_code in COURSE_CODES:
            course = self.courses_by_code.get(course_code)
            if course is None:
                continue
            for sheet in self.enabled_sheets:
                ws.write(row, col, sheet.upper(), self.styles['column_header'])
                col += 1

        row += 1
        for student_info in self.data:
            col = 0
            ws.write(row, col, student_info['last_name'])
            col += 1
            ws.write(row, col, student_info['first_name'])
            col += 1
            ws.write(row, col, student_info['gender'])
            col += 1
            for course_code in COURSE_CODES:
                course = self.courses_by_code.get(course_code)
                if course is None:
                    continue
                course_student_info =  student_info.get(course_code)
                if course_student_info is not None:
                    term_student_info = course_student_info.get(self.term)
                    if term_student_info is not None:
                        for sheet in self.enabled_sheets:
                            value = term_student_info.get(sheet)
                            if value is not UNSCORED:
                                background = self.get_background_color(value)
                                style = self.easyxf(
                                    'font: bold on, color black; '
                                    'pattern: pattern solid, fore_color %s; '
                                    'alignment: horizontal center; ' % background,
                                    num_format_str='0')
                                ws.write(row, col, value, style)
                            col += 1
                    else:
                        col += len(self.enabled_sheets)
                else:
                    col += len(self.enabled_sheets)
            row += 1

        row += 1
        col = 0
        title = 'Year break down by student'
        merge = (len([code
                      for code in COURSE_CODES
                      if self.courses_by_code.get(code) is not None]) *
                 len(self.enabled_sheets)) + 2
        ws.write_merge(row, row, col, col + merge,
                       title, self.styles['table_title'])
        row += 2
        col = 3
        merge = len(self.terms) - 1
        for i, course_code in enumerate(COURSE_CODES):
            course = self.courses_by_code.get(course_code)
            if course is None:
                continue
            title = course.title
            ws.write_merge(row, row, col, col + merge,
                           title, self.styles['column_header'])
            col += merge + 1

        row += 1
        col = 0
        headers = [
            'Last name',
            'First name',
            'Gender',
        ]
        for header in headers:
            ws.write(row, col, header, self.styles['column_header'])
            col += 1

        for course_code in COURSE_CODES:
            course = self.courses_by_code.get(course_code)
            if course is None:
                continue
            for term in self.terms:
                ws.write(row, col, term.title, self.styles['column_header'])
                col += 1

        row += 1
        for student_info in self.data:
            col = 0
            ws.write(row, col, student_info['last_name'])
            col += 1
            ws.write(row, col, student_info['first_name'])
            col += 1
            ws.write(row, col, student_info['gender'])
            col += 1
            for course_code in COURSE_CODES:
                course = self.courses_by_code.get(course_code)
                if course is None:
                    continue
                course_student_info =  student_info.get(course_code)
                if course_student_info is not None:
                    for term in self.terms:
                        term_student_info = course_student_info.get(term)
                        if term_student_info is not None:
                            averages = {}
                            for sheet in self.enabled_sheets:
                                value = term_student_info.get(sheet)
                                if value is not UNSCORED:
                                    averages[sheet] = float(value)
                            if averages:
                                total = 0
                                for sheet, sheet_average in averages.items():
                                    total += sheet_average * self.enabled_sheets[sheet]
                                value = total
                                background = self.get_background_color(value)
                                style = self.easyxf(
                                    'font: bold on, color black; '
                                    'pattern: pattern solid, fore_color %s; '
                                    'alignment: horizontal center; ' % background,
                                    num_format_str='0')
                                ws.write(row, col, value, style)
                            col += 1
                        else:
                            col += 1
                else:
                    col += len(self.terms)
            row += 1

        row += 1
        self.print_legend_1(ws, row, 0)

        self.finish('sheet1')

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def terms(self):
        return sorted(self.schoolyear.values(),
                      key=lambda term: term.first)


class WeightsMustAddOneHundred(ValidationError):

    __doc__ = _('Weights must add up to 100')


class WeightFieldValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        super(WeightFieldValidator, self).validate(value)
        if value:
            total = []
            for sheet in ['bot', 'mot', 'eot']:
                sheet_enabled = self.view.widgets[sheet].value
                if sheet_enabled:
                    weight = self.view.widgets['%s_weight' % sheet].value
                    if weight:
                        total.append(Decimal(weight))
            if not (abs(sum(total) - Decimal(100)) <= 0.0005):
                raise WeightsMustAddOneHundred(value)


validator.WidgetValidatorDiscriminators(
    WeightFieldValidator,
    view=AssessmentPreferencesView,
    field=IWeightField)
