#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
PEAS demographics initialization
"""

from collections import OrderedDict

from schooltool.app.app import InitBase
from schooltool.basicperson.demographics import BoolFieldDescription
from schooltool.basicperson.demographics import DescriptionFieldDescription
from schooltool.basicperson.demographics import EnumFieldDescription
from schooltool.basicperson.demographics import IntFieldDescription
from schooltool.basicperson.demographics import TextFieldDescription
from schooltool.basicperson.interfaces import IDemographicsFields

from schooltool.peas import PeasMessage as _


ATTAINMENT_FIELDS = OrderedDict([
    ('peas_ple_english'           , _('PLE English')),
    ('peas_ple_maths'             , _('PLE Maths')),
    ('peas_ple_sciences'          , _('PLE Sciences')),
    ('peas_ple_socialstudies'     , _('PLE Social Studies')),
    ('peas_uce_math'              , _('UCE Math')),
    ('peas_uce_english'           , _('UCE English')),
    ('peas_uce_biology'           , _('UCE Biology')),
    ('peas_uce_physics'           , _('UCE Physics')),
    ('peas_uce_chemistry'         , _('UCE Chemistry')),
    ('peas_uce_agriculture'       , _('UCE Agriculture')),
    ('peas_uce_entrepreneurship'  , _('UCE Entrepreneurship')),
    ('peas_uce_history'           , _('UCE History')),
    ('peas_uce_religiouseducation', _('UCE Religious Education (CRE/IRE)')),
    ('peas_uce_geography'         , _('UCE Geography')),
])


def setUpAttainmentFields(app):
    dfs = IDemographicsFields(app)
    scores = [unicode(score) for score in range(1, 10)]
    for ID, title in ATTAINMENT_FIELDS.items():
        dfs[ID] = EnumFieldDescription(ID, title, limit_keys=['students'])
        dfs[ID].items = scores[:]


def setUpDemographics(app):
    dfs = IDemographicsFields(app)

    # student fields
    if 'day_boarding' not in dfs:
        dfs['day_boarding'] = EnumFieldDescription(
            'day_boarding', _('Day/Boarding'),
            limit_keys=['students'])
        dfs['day_boarding'].items = [_('Day'),
                                     _('Boarding')]

    if 'use' not in dfs:
        dfs['use'] = EnumFieldDescription(
            'use', _('USE'),
            limit_keys=['students'])
        dfs['use'].items = [_('USE'),
                            _('non-USE')]

    if 'ple_year' not in dfs:
        dfs['ple_year'] = IntFieldDescription(
            'ple_year', _('Year of sitting PLE'),
            limit_keys=['students'])

    if 'ple_score' not in dfs:
        dfs['ple_score'] = IntFieldDescription(
            'ple_score', _('PLE aggregate score'),
            limit_keys=['students'])
        dfs['ple_score'].min_value = 4
        dfs['ple_score'].max_value = 32

    if 'enrollment_year' not in dfs:
        dfs['enrollment_year'] = IntFieldDescription(
            'enrollment_year', _('Year of joining PEAS/Year of enrollment'),
            limit_keys=['students'])

    if 'address_district' not in dfs:
        dfs['address_district'] = TextFieldDescription(
            'address_district', _('Address: District'),
            limit_keys=['students'])

    if 'address_subcounty' not in dfs:
        dfs['address_subcounty'] = TextFieldDescription(
            'address_subcounty', _('Address: Sub-county'),
            limit_keys=['students'])

    if 'address_parish' not in dfs:
        dfs['address_parish'] = TextFieldDescription(
            'address_parish', _('Address: Parish'),
            limit_keys=['students'])

    if 'parental_status' not in dfs:
        dfs['parental_status'] = EnumFieldDescription(
            'parental_status', _('Parental Status'),
            limit_keys=['students'])
        dfs['parental_status'].items = [_('Both alive'),
                                        _('Mother only'),
                                        _('Father only'),
                                        _('Both dead')]

    if 'disability' not in dfs:
        dfs['disability'] = EnumFieldDescription(
            'disability', _('Disability?'),
            limit_keys=['students'])
        dfs['disability'].items = [_('No'),
                                   _('Yes - sight'),
                                   _('Yes - hearing'),
                                   _('Yes - mobility'),
                                   _('Yes - other')]

    if 'out_of_school' not in dfs:
        dfs['out_of_school'] = BoolFieldDescription(
            'out_of_school',
            _('Out of school for more than 1 year before coming to PEAS?'),
            limit_keys=['students'])

    if 'out_of_school_reason' not in dfs:
        dfs['out_of_school_reason'] = EnumFieldDescription(
            'out_of_school_reason', _('Reason for being out of school?'),
            limit_keys=['students'])
        dfs['out_of_school_reason'].items = [
            _('Left to have a baby (pregnancy)'),
            _('Work'),
            _('Family death or illness'),
            _('Fees'),
            _('Other')]

    if 'language' not in dfs:
        dfs['language'] = EnumFieldDescription(
            'language', _('Language'),
            limit_keys=['students'])

    if 'distance' not in dfs:
        dfs['distance'] = IntFieldDescription(
            'distance', _('Distance from the school (min)'),
            limit_keys=['students'])

    # teacher fields
    if 'subject_1' not in dfs:
        dfs['subject_1'] = TextFieldDescription(
            'subject_1', _('Subject taught (1)'),
            limit_keys=['teachers'])

    if 'subject_2' not in dfs:
        dfs['subject_2'] = TextFieldDescription(
            'subject_2', _('Subject taught (2)'),
            limit_keys=['teachers'])

    if 'subject_3' not in dfs:
        dfs['subject_3'] = TextFieldDescription(
            'subject_3', _('Subject taught (3)'),
            limit_keys=['teachers'])

    if 'other_responsabilities' not in dfs:
        dfs['other_responsabilities'] = EnumFieldDescription(
            'other_responsabilities', _('Other Responsabilities'),
            limit_keys=['teachers'])

    if 'qualifications' not in dfs:
        dfs['qualifications'] = EnumFieldDescription(
            'qualifications', _('Qualifications'),
            limit_keys=['teachers'])
        dfs['qualifications'].items = [_('Graduate'),
                                       _('Diploma'),
                                       _('S6 graduate'),
                                       _('Unqualified')]

    if 'joining_year' not in dfs:
        dfs['joining_year'] = IntFieldDescription(
            'joining_year', _('Year of joining PEAS'),
            limit_keys=['teachers'])

    if 'cpd_attended' not in dfs:
        dfs['cpd_attended'] = DescriptionFieldDescription(
            'cpd_attended', _('CPD attended'),
            limit_keys=['teachers'])

    if 'teacher_accomodation' not in dfs:
        dfs['teacher_accomodation'] = BoolFieldDescription(
            'teacher_accomodation', _('Teacher accomodation'),
            limit_keys=['teachers'])

    if 'child_bursary' not in dfs:
        dfs['child_bursary'] = BoolFieldDescription(
            'child_bursary', _('Child on Bursary'),
            limit_keys=['teachers'])

    if 'shift_type' not in dfs:
        dfs['shift_type'] = EnumFieldDescription(
            'shift_type', _('Shift type'),
            limit_keys=['teachers'])
        dfs['shift_type'].items = [_('Full time'),
                                   _('Part time')]


class PeasDemographicsInit(InitBase):

    after = ('schooltool.basicperson.demographics',)

    def __call__(self):
        self.deleteDefaultFields()
        setUpDemographics(self.app)

    def deleteDefaultFields(self):
        dfs = IDemographicsFields(self.app)
        names = ['ID', 'ethnicity', 'language', 'placeofbirth', 'citizenship']
        for name in names:
            if name in dfs:
                del dfs[name]


class PeasAttainmentInit(InitBase):

    after = ('schooltool.peas.demographics')

    def __call__(self):
        setUpAttainmentFields(self.app)
