#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2015 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Note report views
"""

from datetime import datetime
from datetime import timedelta
from collections import Counter

from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import Interface
from zope.schema import Choice

from z3c.form import field

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.interfaces import IDemographics
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.export.export import RequestXLSReportDialog
from schooltool.group.interfaces import IGroupContainer
from schooltool.group.group import defaultGroups
from schooltool.level.interfaces import ILevelContainer
from schooltool.lyceum.journal.browser.journal import makeSchoolAttendanceMeeting
from schooltool.task.progress import normalized_progress
from schooltool.term.interfaces import IDateManager

from schooltool.peas import PeasMessage as _
from schooltool.peas.browser.assessment import SchoolYearReportLink
from schooltool.peas.browser.report import OverallReportBase
from schooltool.peas.browser.report import TeacherOverallReportBase
from schooltool.peas.note.interfaces import INote
from schooltool.peas.note.interfaces import INoteContainer
from schooltool.peas.note.browser.note import single_line


class NotesReportLink(SchoolYearReportLink):

    @property
    def enabled(self):
        has_terms = bool(self.context)
        return has_terms


class INoteReport(Interface):

    term = Choice(
        title=STMessage('Term'),
        source='schooltool.peas.year_terms')

    level = Choice(
        title=STMessage('Level'),
        vocabulary='schooltool.level.level.levelvocabulary',
        required=False)


class NotesReportRequestView(RequestXLSReportDialog):

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['schoolyear'] = self.context

    def updateWidgets(self):
        super(NotesReportRequestView, self).updateWidgets()
        self.widgets['term'].prompt = True
        widget = self.widgets.get('level')
        if widget is not None:
            widget.prompt = True
            widget.promptMessage = _('All')

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term.__name__
        level = self.form_params.get('level')
        if level is not None:
            task.request_params['level'] = level.__name__


class StudentNotesReportRequestView(NotesReportRequestView):

    report_builder = 'student_notes_report.xls'

    fields = field.Fields(INoteReport)


class TeacherNotesReportRequestView(NotesReportRequestView):

    report_builder = 'teacher_notes_report.xls'

    fields = field.Fields(INoteReport).select('term')


def daterange(start, end):
    result = []
    for n in range(int((end - start).days)):
        result.append(start + timedelta(n))
    return result


class NotesReportViewBase(OverallReportBase):

    @Lazy
    def custom_colors(self):
        result = super(NotesReportViewBase, self).custom_colors
        result['attendance'] = (30, 85, 142, 213)
        result['attainment'] = (31, 146, 208, 80)
        result['behaviour'] = (32, 179, 162, 199)
        result['extra'] = (33, 147, 205, 221)
        result['fees'] = (34, 217, 150, 148)
        result['lesson'] = (35, 255, 255, 153)
        result['other'] = (36, 247, 150, 70)
        return result

    def get_background_color(self, value):
        return value

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def term(self):
        for term in self.schoolyear.values():
            if term.__name__ == self.request.get('term'):
                return term

    @Lazy
    def selected_level(self):
        app = ISchoolToolApplication(None)
        container = ILevelContainer(app)
        return container.get(self.request.get('level'))

    def print_headers(self, ws, title):
        row = 0
        # XXX: overwritten only for merge
        merge = self.title_merge
        self.print_report_title(ws, title, row=row, merge=merge)

        row = 2
        today = getUtility(IDateManager).today
        self.print_report_date(ws, today, row=row)

        row += 1
        self.print_report_filters(ws, row=row)

    @Lazy
    def meetings(self, ignore_future=True):
        result = []
        today = getUtility(IDateManager).today
        for date in daterange(self.schoolyear.first, self.schoolyear.last):
            if ignore_future and date > today:
                continue
            dt = self.tzinfo.localize(
                datetime(date.year, date.month, date.day)
            )
            result.append(makeSchoolAttendanceMeeting(dt))
        return result

    def meetings_by_term(self, term):
        result = {}
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)
        index = terms.index(term)
        start = term.first
        try:
            end = terms[index + 1].first
        except (IndexError,):
            end = self.schoolyear.last
        meetings_by_date = dict(
            [(meeting.dtstart.astimezone(self.tzinfo).date(), meeting)
             for meeting in self.meetings]
        )
        for date in daterange(start, end):
            meeting = meetings_by_date.get(date)
            if meeting is not None:
                result[meeting] = self.data[meeting]
        return result

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        today = getUtility(IDateManager).today
        target = today - timedelta(days=30)
        meetings = {}
        for meeting, meeting_data in self.data.items():
            if meeting.dtstart.astimezone(self.tzinfo).date() >= target:
                meetings[meeting] = meeting_data
        title = _('Notes from the last 30 days')
        legend = _('Number of reports by category in the last 30 days')
        legend_total = _('Total in the last 30 days')
        row = self.print_notes_table(ws, title, legend, legend_total,
                                     self.initial_row, meetings)
        row += 2
        meetings = self.meetings_by_term(self.term)
        title = _('Notes made this term')
        legend = _('Number of reports by category this term')
        legend_total = _('Total this term')
        self.print_notes_table(ws, title, legend, legend_total, row, meetings)
        self.finish('sheet1')

    def print_notes_table(self, ws, title, legend, legend_total, row, meetings):
        ws.write_merge(row, row, 0, self.title_merge,
                       title, self.styles['table_title'])
        row += 1
        for i, header in enumerate(self.table_headers):
            ws.write(row, i, header, self.styles['column_header'])
        row += 1
        table = []
        for meeting, meeting_data in meetings.items():
            for person, person_data in meeting_data.items():
                if (self.selected_level is not None and
                    person_data.get('level') != self.selected_level):
                    continue
                for note in person_data['notes']:
                    table.append(self.format_row(person, person_data, note))
        vocabulary = INote['category'].vocabulary
        table = sorted(table, key=self.sorting_key)
        categories = []
        for i, table_row in enumerate(table):
            for col_no, col in enumerate(table_row):
                if col_no == (len(table_row) - 2):
                    background = self.get_background_color(col)
                    categories.append(col)
                    style = self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color %s; ' % background)
                    category = vocabulary.getTerm(col).title
                    ws.write(row, col_no, category, style)
                elif col_no == (len(table_row) - 4):
                    ws.write(row, col_no, col.strftime('%d/%m/%Y'))
                else:
                    ws.write(row, col_no, col)
            row += 1
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom no_line, left medium; '
        )
        ws.write_merge(row, row, 0, 1, legend, style)
        row += 1
        counter = Counter(categories)
        cats = [
            'attendance',
            'attainment',
            'behaviour',
            'extra',
            'fees',
            'lesson',
            'other',
        ]
        for category in cats:
            background = self.get_background_color(category)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top no_line, right no_line, bottom no_line, left medium; ' % background
            )
            ws.write(row, 0, vocabulary.getTerm(category).title, style)
            style = self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right medium, bottom no_line, left no_line; ')
            ws.write(row, 1, counter[category], style)
            row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, legend_total, style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, len(categories), style)
        return row


class StudentNotesReportView(NotesReportViewBase):

    message_title = _('Student notes')

    sheet_1_title = _('Student notes')

    sheet_1_label = _('Student notes')

    title_merge = 9

    initial_row = 7

    table_headers = (
        _('Class'),
        _('Stream'),
        _('Surname'),
        _('Firstname'),
        _('Gender'),
        _('Date of note'),
        _('Author'),
        _('Category'),
        _('Note detail')
    )

    @property
    def sorting_key(self):
        return lambda row: (row[0], row[1], row[2], row[5])

    @Lazy
    def groups(self):
        result = []
        groups = IGroupContainer(self.schoolyear)
        for group in groups.values():
            if group.__name__ not in defaultGroups:
                result.append(group)
        return result

    def group(self, person, date):
        result = []
        for group in self.groups:
            if person in group.members.on(date):
                result.append(group.title)
        return ', '.join(sorted(result))

    def format_row(self, person, person_data, note):
        return [
            person_data['level'].title,
            self.group(person, note.date),
            person.last_name,
            person.first_name,
            person.gender,
            note.date,
            ', '.join([editor.title for editor in note.editors]),
            note.category,
            single_line(note.body),
        ]

    def addExporters(self, progress):
        super(StudentNotesReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Student notes sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'student_notes_'

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Year'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Term'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.term.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Class'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        level = self.selected_level.title if self.selected_level else 'All'
        ws.write(row, 1, level, style)

    @Lazy
    def data(self):
        result = {}
        app = ISchoolToolApplication(None)
        persons = app['persons']
        students = IGroupContainer(self.schoolyear).get('students')
        meetings_count = len(self.meetings)
        persons_count = len(app['persons'])
        for meeting_no, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            students_on_date = list(students.members.on(date))
            meeting_data = {}
            for person_no, person in enumerate(persons.values()):
                self.progress(
                    'data', normalized_progress(
                        meeting_no, meetings_count, person_no, persons_count)
                )
                level = self.level(person, date)
                if (level is not None and
                    (person in students_on_date or
                     self.is_learner(person, date))):
                    person_notes_on_date = []
                    notes = INoteContainer(person)
                    for note in notes.values():
                        if note.date == date:
                            person_notes_on_date.append(note)
                    if person_notes_on_date:
                        meeting_data[person] = {
                            'notes': person_notes_on_date,
                            'level': level,
                        }
            result[meeting] = meeting_data
        return result


class TeacherNotesReportView(NotesReportViewBase, TeacherOverallReportBase):

    message_title = _('Teacher notes')

    sheet_1_title = _('Teacher notes')

    sheet_1_label = _('Teacher notes')

    title_merge = 8

    initial_row = 6

    table_headers = (
        _('Surname'),
        _('Firstname'),
        _('Gender'),
        _('FT / PT'),
        _('Date of note'),
        _('Author'),
        _('Category'),
        _('Note detail'),
    )

    @property
    def sorting_key(self):
        return lambda row: (row[0], row[4])

    def format_row(self, person, person_data, note):
        demographics = IDemographics(person)
        return [
            person.last_name,
            person.first_name,
            person.gender,
            demographics['shift_type'],
            note.date,
            ', '.join([editor.title for editor in note.editors]),
            note.category,
            single_line(note.body),
        ]

    def addExporters(self, progress):
        super(TeacherNotesReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Teacher notes sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'teacher_notes_'

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Year'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Term'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.term.title, style)

    @Lazy
    def data(self):
        result = {}
        app = ISchoolToolApplication(None)
        persons = app['persons']
        teachers = IGroupContainer(self.schoolyear).get('teachers')
        meetings_count = len(self.meetings)
        persons_count = len(app['persons'])
        for meeting_no, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            teachers_on_date = list(teachers.members.on(date))
            meeting_data = {}
            for person_no, person in enumerate(persons.values()):
                self.progress(
                    'data', normalized_progress(
                        meeting_no, meetings_count, person_no, persons_count)
                )
                if (person in teachers_on_date or
                    self.is_instructor(person, date)):
                    person_notes_on_date = []
                    notes = INoteContainer(person)
                    for note in notes.values():
                        if note.date == date:
                            person_notes_on_date.append(note)
                    if person_notes_on_date:
                        meeting_data[person] = {
                            'notes': person_notes_on_date,
                        }
            result[meeting] = meeting_data
        return result
