#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
PEAS reports.
"""

from collections import OrderedDict
import calendar
from decimal import Decimal
import datetime
import pytz
from urllib import urlencode
from StringIO import StringIO

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts
from zope.component import getAdapters
from zope.component import getUtility
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import implements
from zope.schema import Choice
from zope.schema import Date
from zope.schema import Float
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

import xlwt
import zc.resourcelibrary
from z3c.form import field
from z3c.form import widget
from zc.table.column import GetterColumn

from schooltool import table
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.browser.cal import month_names
from schooltool.app.browser.cal import CalendarViewBase
from schooltool.app.membership import Membership
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.relationships import Instruction
from schooltool.basicperson.demographics import LEAVE_SCHOOL_FIELDS
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.basicperson.interfaces import IDemographics
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.basicperson.person import StudentLevels
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.contact.interfaces import IContactable
from schooltool.contact.interfaces import IContact
from schooltool.contact.contact import PARENT
from schooltool.course.interfaces import ISection
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.interfaces import ILearner
from schooltool.course.interfaces import ISectionContainer
from schooltool.course.section import is_student
from schooltool.export.export import ExcelExportView
from schooltool.export.export import RequestXLSReportDialog
from schooltool.fee.interfaces import IFeeContainer
from schooltool.fee.interfaces import ICredit
from schooltool.fee.interfaces import IDebit
from schooltool.gradebook.interfaces import IGradebookRoot
from schooltool.gradebook.interfaces import IActivities
from schooltool.gradebook.browser.report_card import FlourishLayoutReportCardView
from schooltool.gradebook.browser.gradebook import convertAverage
from schooltool.group.group import defaultGroups
from schooltool.group.interfaces import IGroupContainer
from schooltool.group.interfaces import IGroup
from schooltool.level.interfaces import ILevelContainer
from schooltool.lyceum.journal.browser.journal import FlourishSchoolAttendanceView
from schooltool.lyceum.journal.browser.journal import makeSchoolAttendanceMeeting
from schooltool.lyceum.journal.interfaces import IEvaluateRequirement
from schooltool.lyceum.journal.journal import TeacherAMRequirement
from schooltool.lyceum.journal.journal import StudentAMRequirement
from schooltool.lyceum.journal.journal import StudentPMRequirement
from schooltool.lyceum.journal.browser.journal import IRequestAbsencesByDayForm
from schooltool.person.interfaces import IPersonFactory
from schooltool.relationship.temporal import ACTIVE
from schooltool.report.report import ReportLinkViewlet
from schooltool.requirement.interfaces import IEvaluations
from schooltool.requirement.interfaces import IDiscreteValuesScoreSystem
from schooltool.requirement.interfaces import IRangedValuesScoreSystem
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.skin import flourish
from schooltool.task.progress import normalized_progress
from schooltool.term.interfaces import ITerm
from schooltool.term.interfaces import IDateManager

from schooltool.peas import PeasMessage as _


get_day_boarding = lambda p: IDemographics(p).get('day_boarding') or ''
get_shift_type = lambda p: IDemographics(p).get('shift_type') or ''
is_female = lambda p, person_data: p.gender == 'female'
is_male = lambda p, person_data: p.gender == 'male'
is_boarding = lambda p, person_data: get_day_boarding(p).lower() == 'boarding'
is_day = lambda p, person_data: get_day_boarding(p).lower() == 'day'
is_female_boarding = lambda person, person_data: (
    is_female(person, person_data) and
    is_boarding(person, person_data)
)
is_female_day = lambda person, person_data: (
    is_female(person, person_data) and
    is_day(person, person_data)
)
is_male_boarding = lambda person, person_data: (
    is_male(person, person_data) and
    is_boarding(person, person_data)
)
is_male_day = lambda person, person_data: (
    is_male(person, person_data) and
    is_day(person, person_data)
)
is_full_time = lambda p, person_data: get_shift_type(p).lower() == 'full time'
is_part_time = lambda p, person_data: get_shift_type(p).lower() == 'part time'


class MonthlyReportLinkViewlet(ReportLinkViewlet):

    @property
    def month(self):
        return self.view.month

    @property
    def year(self):
        return self.view.year

    @property
    def extra_params(self):
        return {
            'month': self.month,
            'year': self.year,
        }


class MonthsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        result.append(self.createTerm(
            None,
            widget.SequenceWidget.noValueToken,
            _('Select a month'),
        ))
        result.extend(
            [self.createTerm(month, month, title)
             for month, title in month_names.items()]
        )
        return result


def MonthsVocabularyFactory():
    return MonthsVocabulary


class YearsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms(self.context.get('years'))
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self, years):
        result = []
        result.append(self.createTerm(
            None,
            widget.SequenceWidget.noValueToken,
            _('Select a year'),
        ))
        for year in years:
            result.append(self.createTerm(
                year,
                year,
                year,
            ))
        return result


def YearsVocabularyFactory():
    return YearsVocabulary


class SchoolYearsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        result.append(self.createTerm(
            None,
            widget.SequenceWidget.noValueToken,
            _('Select a school year'),
        ))
        app = ISchoolToolApplication(None)
        container = ISchoolYearContainer(app)
        for schoolyear_id, schoolyear in container.items():
            result.append(self.createTerm(
                schoolyear,
                schoolyear_id.encode('utf-8'),
                schoolyear.title,
            ))
        return result


def SchoolYearsVocabularyFactory():
    return SchoolYearsVocabulary


class IRequestMonthlyReport(Interface):

    year = Choice(
        title=_(u'Year'),
        source='schooltool.peas.years',
        required=True)

    month = Choice(
        title=_(u'Month'),
        source='schooltool.peas.months',
        required=True)


class IRequestYearlyReport(Interface):

    schoolyear = Choice(
        title=_(u'School Year'),
        source='schooltool.peas.schoolyears',
        required=True)


RequestMonthlyReportMonthDefault = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.view.month,
    field=IRequestMonthlyReport['month']
    )


RequestMonthlyReportYearDefault = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.view.year,
    field=IRequestMonthlyReport['year']
    )


def default_schoolyear(adapter):
    selected_years = adapter.view.schoolyears
    if selected_years:
        return selected_years[-1]


RequestYearlyReportSchoolYearDefault = widget.ComputedWidgetAttribute(
    default_schoolyear,
    field=IRequestYearlyReport['schoolyear']
    )


class RequestTeacherAttendanceDisaggregatedByDayView(
        RequestXLSReportDialog,
        FlourishSchoolAttendanceView):

    report_builder = 'teacher_attendance_disaggregated_by_day.xls'
    update = RequestXLSReportDialog.update
    render = RequestXLSReportDialog.render

    fields = field.Fields(IRequestMonthlyReport)

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['years'] = self.years()

    def years(self):
        this_year = self.year
        terms = self.all_terms
        min_year = min([this_year] +
                       [term.first.year
                        for term in terms.values() if term.first])
        max_year = max([this_year] +
                       [term.last.year
                        for term in terms.values() if term.last])
        return range(min_year, max_year+1)

    def updateTaskParams(self, task):
        month = self.form_params.get('month')
        year = self.form_params.get('year')
        if month is not None:
            task.request_params['month'] = str(month)
        if year is not None:
            task.request_params['year'] = str(year)


class RequestTeacherAttendanceDisaggregatedByWeekView(
        RequestTeacherAttendanceDisaggregatedByDayView):

    report_builder = 'teacher_attendance_disaggregated_by_week.xls'


class RequestTeacherAttendanceDisaggregatedByMonthView(
        RequestTeacherAttendanceDisaggregatedByDayView):

    report_builder = 'teacher_attendance_disaggregated_by_month.xls'

    fields = field.Fields(IRequestYearlyReport)

    def updateTaskParams(self, task):
        schoolyear = self.form_params.get('schoolyear')
        if schoolyear is not None:
            task.request_params['schoolyear'] = schoolyear.__name__


class RequestStudentAttendanceDisaggregatedByDayView(
        RequestTeacherAttendanceDisaggregatedByDayView):

    report_builder = 'student_attendance_disaggregated_by_day.xls'


class RequestStudentAttendanceDisaggregatedByWeekView(
        RequestTeacherAttendanceDisaggregatedByWeekView):

    report_builder = 'student_attendance_disaggregated_by_week.xls'


class RequestStudentAttendanceDisaggregatedByMonthView(
        RequestTeacherAttendanceDisaggregatedByMonthView):

    report_builder = 'student_attendance_disaggregated_by_month.xls'


class TeacherAttendanceDisaggregatedByDayView(ExcelExportView,
                                              FlourishSchoolAttendanceView):

    message_title = _('teacher attendance disaggregated by day report')
    sheet_title = _('Teacher Report by Day')

    @property
    def base_filename(self):
        month = self.month
        year = self.year
        return 'teacher_attendance_disaggregated_by_day_%d_%d' % (
            year, month)

    def export_report(self, wb):
        self.task_progress.force('report', active=True)
        ws = wb.add_sheet(self.sheet_title)
        self.print_headers(ws)
        self.print_report_rows(ws)
        self.finish('report')

    def print_headers(self, ws):
        headers = [
            'Date',
            '% of female teachers present',
            '% male teachers present',
            '% total teachers present',
            '% blanks',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 0, index, header)

    def print_report_rows(self, ws):
        row = 1
        for data in self.report_rows():
            self.write(ws, row, 0, data['date'])
            self.write(ws, row, 1, data.get('female_present'))
            self.write(ws, row, 2, data.get('male_present'))
            self.write(ws, row, 3, data.get('teacher_present'))
            self.write(ws, row, 4, data.get('blank'))
            row += 1

    def report_rows(self):
        result = []
        total = len(self.persons)
        total_female = len(self.female)
        total_male = len(self.male)
        today = getUtility(IDateManager).today
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            data = {}
            data['date'] = date.isoformat()
            female = []
            male = []
            teachers = []
            blank = []
            for person in self.persons:
                score = self.getScores(person)[meeting]
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    blank.append(1)
                elif not score.scoreSystem.isAbsent(score):
                    teachers.append(1)
                    if person.gender == 'male':
                        male.append(1)
                    elif person.gender == 'female':
                        female.append(1)
            if total_female:
                data['female_present'] = self.formatPercent(sum(female), total_female)
            if total_male:
                data['male_present'] = self.formatPercent(sum(male), total_male)
            if total:
                data['teacher_present'] = self.formatPercent(
                    sum(teachers), total)
                data['blank'] = self.formatPercent(sum(blank), total)
            result.append(data)
        return result

    def formatPercent(self, count, total):
        return '%.0f' % ((count*100)/float(total))

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self._grade_cache = {}
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()

        year = self.year
        month = self.month

        dates = [self.tzinfo.localize(datetime.datetime(year, month, day))
                 for day in calendar.Calendar().itermonthdays(year, month)
                 if day]

        meetings = []

        for dt in dates:
            date = dt.date()
            for term in terms:
                if date in term and term.isSchoolday(date):
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    @Lazy
    def all_meetings(self):
        return self.allMeetings()

    def getScores(self, person):
        if person in self._grade_cache:
            return self._grade_cache[person]
        self._grade_cache[person] = result = {}
        unique_meetings = set()
        terms = self.terms
        events = list(self.all_meetings)
        sorted_events = sorted(events, key=lambda e: e.dtstart)
        unproxied_person = removeSecurityProxy(person)
        for event in sorted_events:
            date = event.dtstart.astimezone(self.tzinfo).date()
            if not any([date in term for term in terms]):
                continue
            unproxied_event = removeSecurityProxy(event)
            requirement = self.makeRequirement(unproxied_event)
            score = IEvaluateRequirement(requirement).getEvaluation(
                    unproxied_person, requirement, default=UNSCORED)
            if event.meeting_id not in unique_meetings:
                result[event] = score
                unique_meetings.add(event.meeting_id)
        return result

    def makeRequirement(self, meeting):
        return TeacherAMRequirement(meeting)

    def teacherGroups(self):
        result = []
        group_id = 'teachers'
        for year in self.schoolyears:
            groups = IGroupContainer(year)
            result.append(groups[group_id])
        return result

    @Lazy
    def persons(self):
        result = set()
        groups = self.teacherGroups()
        for group in groups:
            for teacher in group.members.all():
                result.add(removeSecurityProxy(teacher))
        return result

    @Lazy
    def female(self):
        result = []
        for teacher in self.persons:
            if teacher.gender == 'female':
                result.append(teacher)
        return result

    @Lazy
    def male(self):
        result = []
        for teacher in self.persons:
            if teacher.gender == 'male':
                result.append(teacher)
        return result


class TeacherAttendanceDisaggregatedByWeekView(
        TeacherAttendanceDisaggregatedByDayView,
        CalendarViewBase):

    message_title = _('teacher attendance disaggregated by week report')
    sheet_title = _('Teacher Report by Week')

    @property
    def base_filename(self):
        month = self.month
        year = self.year
        return 'teacher_attendance_disaggregated_by_week_%d_%d' % (
            year, month)

    def report_rows(self):
        result = []
        total = len(self.persons)
        total_female = len(self.female)
        total_male = len(self.male)
        weeks = []
        meeting_dates = {}
        today = getUtility(IDateManager).today
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            meeting_dates[date] = meeting
        for week in self.getMonth(datetime.date(self.year, self.month, 1)):
            for day in week:
                if day.date in meeting_dates:
                    weeks.append(week)
                    break
        for week in weeks:
            data = {}
            date = week[0].date
            data['date'] = translate(_('W/b on ${date}',
                                       mapping={'date': date.isoformat()}),
                                     context=self.request)
            meeting_days_in_week = []
            female_week = []
            male_week = []
            teachers_week = []
            blank_week = []
            for day in week:
                female = []
                male = []
                teachers = []
                blank = []
                meeting = meeting_dates.get(day.date)
                if meeting is not None:
                    meeting_days_in_week.append(1)
                    for person in self.persons:
                        score = self.getScores(person)[meeting]
                        if (score is None or
                            score is UNSCORED or
                            score.value is UNSCORED):
                            blank.append(1)
                        elif not score.scoreSystem.isAbsent(score):
                            teachers.append(1)
                            if person.gender == 'male':
                                male.append(1)
                            elif person.gender == 'female':
                                female.append(1)
                    female_week.append(sum(female))
                    male_week.append(sum(male))
                    teachers_week.append(sum(teachers))
                    blank_week.append(sum(blank))
            total_days_in_week = sum(meeting_days_in_week)
            if total_female:
                data['female_present'] = self.formatPercent(
                    sum(female_week), total_female * total_days_in_week)
            if total_male:
                data['male_present'] = self.formatPercent(
                    sum(male_week), total_male * total_days_in_week)
            if total:
                data['teacher_present'] = self.formatPercent(
                    sum(teachers_week), total * total_days_in_week)
                data['blank'] = self.formatPercent(
                    sum(blank_week), total * total_days_in_week)
            result.append(data)
        return result


class TeacherAttendanceDisaggregatedByMonthView(
        TeacherAttendanceDisaggregatedByWeekView):

    message_title = _('teacher attendance disaggregated by month report')
    sheet_title = _('Teacher Report by Month')

    @Lazy
    def schoolyear(self):
        app = ISchoolToolApplication(None)
        schoolyears = ISchoolYearContainer(app)
        return schoolyears.get(self.request.get('schoolyear'))

    @property
    def base_filename(self):
        year = self.schoolyear.title if self.schoolyear is not None else ''
        return 'teacher_attendance_disaggregated_by_month_%s' % year

    def allMeetings(self):
        meetings = {}

        if self.schoolyear is None:
            return meetings

        terms = sorted(self.schoolyear.values(), key=lambda t:t.first)

        if not terms:
            return meetings

        today = getUtility(IDateManager).today
        for term in terms:
            for date in term:
                if date <= today and term.isSchoolday(date):
                    key = (date.year, date.month)
                    if key not in meetings:
                        meetings[key] = []
                    meetings[key].append(makeSchoolAttendanceMeeting(date))

        return meetings

    def getScores(self, person):
        if person in self._grade_cache:
            return self._grade_cache[person]
        self._grade_cache[person] = result = {}
        unique_meetings = set()
        meetings = []
        for month_meetings in self.all_meetings.values():
            meetings.extend(month_meetings)
        sorted_events = sorted(meetings, key=lambda e: e.dtstart)
        unproxied_person = removeSecurityProxy(person)
        for event in sorted_events:
            unproxied_event = removeSecurityProxy(event)
            requirement = self.makeRequirement(unproxied_event)
            score = IEvaluateRequirement(requirement).getEvaluation(
                    unproxied_person, requirement, default=UNSCORED)
            if event.meeting_id not in unique_meetings:
                result[event] = score
                unique_meetings.add(event.meeting_id)
        return result

    def report_rows(self):
        result = []
        total = len(self.persons)
        total_female = len(self.female)
        total_male = len(self.male)
        for key, month_meetings in sorted(self.all_meetings.items()):
            year, month = key
            data = {}
            data['date'] = translate(
                _('${month} ${year}',
                  mapping={'month': month_names[month], 'year': year}))
            female_month = []
            male_month = []
            teachers_month = []
            blank_month = []
            total_days_in_month = len(month_meetings)
            for meeting in month_meetings:
                female = []
                male = []
                teachers = []
                blank = []
                for person in self.persons:
                    score = self.getScores(person)[meeting]
                    if (score is None or
                        score is UNSCORED or
                        score.value is UNSCORED):
                        blank.append(1)
                    elif not score.scoreSystem.isAbsent(score):
                        teachers.append(1)
                        if person.gender == 'male':
                            male.append(1)
                        elif person.gender == 'female':
                            female.append(1)
                female_month.append(sum(female))
                male_month.append(sum(male))
                teachers_month.append(sum(teachers))
                blank_month.append(sum(blank))
            if total_female:
                data['female_present'] = self.formatPercent(
                    sum(female_month), total_female * total_days_in_month)
            if total_male:
                data['male_present'] = self.formatPercent(
                    sum(male_month), total_male * total_days_in_month)
            if total:
                data['teacher_present'] = self.formatPercent(
                    sum(teachers_month), total * total_days_in_month)
                data['blank'] = self.formatPercent(
                    sum(blank_month), total * total_days_in_month)
            result.append(data)
        return result


class StudentAttendanceDisaggregatedByDayView(
        TeacherAttendanceDisaggregatedByDayView):

    message_title = _('student attendance disaggregated by week report')
    sheet_title = _('Student Report by Day')

    @property
    def base_filename(self):
        month = self.month
        year = self.year
        return 'student_attendance_disaggregated_by_day_%d_%d' % (
            year, month)

    def print_report_rows(self, ws):
        row = 1
        for data in self.report_rows():
            self.write(ws, row, 0, data['date'])
            self.write(ws, row, 1, data.get('am'))
            self.write(ws, row, 2, data.get('pm'))
            self.write(ws, row, 3, data.get('all_day'))
            self.write(ws, row, 4, data.get('am_blank'))
            self.write(ws, row, 5, data.get('pm_blank'))
            row += 1

    def print_headers(self, ws):
        headers = [
            'Date',
            '% of students present in the morning AM',
            '% of students present in the afternoon PM',
            '% of students present in both morning and afternoon',
            '% blank morning AM',
            '% blank afternoon PM',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 0, index, header)

    def report_rows(self):
        result = []
        total = len(self.persons)
        today = getUtility(IDateManager).today
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            am_requirement = StudentAMRequirement(meeting)
            pm_requirement = StudentPMRequirement(meeting)
            data = {}
            data['date'] = date.isoformat()
            am = []
            pm = []
            all_day = []
            am_blank = []
            pm_blank = []
            for person in self.persons:
                present_am = False
                present_pm = False
                score = IEvaluateRequirement(am_requirement).getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    am_blank.append(1)
                elif not score.scoreSystem.isAbsent(score):
                    am.append(1)
                    present_am = True
                score = IEvaluateRequirement(pm_requirement).getEvaluation(
                    person, pm_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pm_blank.append(1)
                elif not score.scoreSystem.isAbsent(score):
                    pm.append(1)
                    present_pm = True
                if present_am and present_pm:
                    all_day.append(1)
            if total:
                data['am'] = self.formatPercent(sum(am), total)
                data['pm'] = self.formatPercent(sum(pm), total)
                data['all_day'] = self.formatPercent(sum(all_day), total)
                data['am_blank'] = self.formatPercent(sum(am_blank), total)
                data['pm_blank'] = self.formatPercent(sum(pm_blank), total)
            result.append(data)
        return result

    @Lazy
    def persons(self):
        result = set()
        for username in self.context:
            person = removeSecurityProxy(self.context[username])
            relationships = Membership.bind(member=person).all().relationships
            for link_info in relationships:
                section = removeSecurityProxy(link_info.target)
                if IGroup.providedBy(section):
                    continue
                if ITerm(section) in self.terms:
                    result.add(person)
        return list(result)


class StudentAttendanceDisaggregatedByWeekView(
        StudentAttendanceDisaggregatedByDayView,
        CalendarViewBase):

    message_title = _('student attendance disaggregated by week report')
    sheet_title = _('Student Report by Week')

    @property
    def base_filename(self):
        month = self.month
        year = self.year
        return 'student_attendance_disaggregated_by_week_%d_%d' % (
            year, month)

    def report_rows(self):
        result = []
        total = len(self.persons)
        weeks = []
        meeting_dates = {}
        today = getUtility(IDateManager).today
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            meeting_dates[date] = meeting
        for week in self.getMonth(datetime.date(self.year, self.month, 1)):
            for day in week:
                if day.date in meeting_dates:
                    weeks.append(week)
                    break
        for week in weeks:
            data = {}
            date = week[0].date
            data['date'] = translate(_('W/b on ${date}',
                                       mapping={'date': date.isoformat()}),
                                     context=self.request)
            meeting_days_in_week = []
            am_week = []
            pm_week = []
            all_day_week = []
            am_blank_week = []
            pm_blank_week = []
            for day in week:
                am = []
                pm = []
                all_day = []
                am_blank = []
                pm_blank = []
                meeting = meeting_dates.get(day.date)
                if meeting is not None:
                    am_requirement = StudentAMRequirement(meeting)
                    pm_requirement = StudentPMRequirement(meeting)
                    meeting_days_in_week.append(1)
                    for person in self.persons:
                        present_am = False
                        present_pm = False
                        evaluations = IEvaluateRequirement(am_requirement)
                        score = evaluations.getEvaluation(
                            person, am_requirement, default=UNSCORED)
                        if (score is None or
                            score is UNSCORED or
                            score.value is UNSCORED):
                            am_blank.append(1)
                        elif not score.scoreSystem.isAbsent(score):
                            am.append(1)
                            present_am = True
                        evaluations = IEvaluateRequirement(pm_requirement)
                        score = evaluations.getEvaluation(
                            person, pm_requirement, default=UNSCORED)
                        if (score is None or
                            score is UNSCORED or
                            score.value is UNSCORED):
                            pm_blank.append(1)
                        elif not score.scoreSystem.isAbsent(score):
                            pm.append(1)
                            present_pm = True
                        if present_am and present_pm:
                            all_day.append(1)
                    am_week.append(sum(am))
                    pm_week.append(sum(pm))
                    all_day_week.append(sum(all_day))
                    am_blank_week.append(sum(am_blank))
                    pm_blank_week.append(sum(pm_blank))
            total_days_in_week = sum(meeting_days_in_week)
            if total:
                data['am'] = self.formatPercent(
                    sum(am_week), total * total_days_in_week)
                data['pm'] = self.formatPercent(
                    sum(pm_week), total * total_days_in_week)
                data['all_day'] = self.formatPercent(
                    sum(all_day_week), total * total_days_in_week)
                data['am_blank'] = self.formatPercent(
                    sum(am_blank_week), total * total_days_in_week)
                data['pm_blank'] = self.formatPercent(
                    sum(pm_blank_week), total * total_days_in_week)
            result.append(data)
        return result


class StudentAttendanceDisaggregatedByMonthView(
        StudentAttendanceDisaggregatedByWeekView):

    message_title = _('student attendance disaggregated by month report')
    sheet_title = _('Student Report by Month')

    @Lazy
    def schoolyear(self):
        app = ISchoolToolApplication(None)
        schoolyears = ISchoolYearContainer(app)
        return schoolyears.get(self.request.get('schoolyear'))

    @property
    def base_filename(self):
        year = self.schoolyear.title if self.schoolyear is not None else ''
        return 'student_attendance_disaggregated_by_month_%s' % year

    def allMeetings(self):
        meetings = {}

        if self.schoolyear is None:
            return meetings

        terms = sorted(self.schoolyear.values(), key=lambda t:t.first)

        if not terms:
            return meetings

        today = getUtility(IDateManager).today
        for term in terms:
            for date in term:
                if date <= today and term.isSchoolday(date):
                    key = (date.year, date.month)
                    if key not in meetings:
                        meetings[key] = []
                    meetings[key].append(makeSchoolAttendanceMeeting(date))

        return meetings

    def report_rows(self):
        result = []
        total = len(self.persons)
        for key, month_meetings in sorted(self.all_meetings.items()):
            year, month = key
            data = {}
            data['date'] = translate(
                _('${month} ${year}',
                  mapping={'month': month_names[month], 'year': year}))
            am_month = []
            pm_month = []
            all_day_month = []
            am_blank_month = []
            pm_blank_month = []
            total_days_in_month = len(month_meetings)
            for meeting in month_meetings:
                am = []
                pm = []
                all_day = []
                am_blank = []
                pm_blank = []
                am_requirement = StudentAMRequirement(meeting)
                pm_requirement = StudentPMRequirement(meeting)
                for person in self.persons:
                    present_am = False
                    present_pm = False
                    evaluations = IEvaluateRequirement(am_requirement)
                    score = evaluations.getEvaluation(
                        person, am_requirement, default=UNSCORED)
                    if (score is None or
                        score is UNSCORED or
                        score.value is UNSCORED):
                        am_blank.append(1)
                    elif not score.scoreSystem.isAbsent(score):
                        am.append(1)
                        present_am = True
                    evaluations = IEvaluateRequirement(pm_requirement)
                    score = evaluations.getEvaluation(
                        person, pm_requirement, default=UNSCORED)
                    if (score is None or
                        score is UNSCORED or
                        score.value is UNSCORED):
                        pm_blank.append(1)
                    elif not score.scoreSystem.isAbsent(score):
                        pm.append(1)
                        present_pm = True
                    if present_am and present_pm:
                        all_day.append(1)
                am_month.append(sum(am))
                pm_month.append(sum(pm))
                all_day_month.append(sum(all_day))
                am_blank_month.append(sum(am_blank))
                pm_blank_month.append(sum(pm_blank))
            if total:
                data['am'] = self.formatPercent(
                    sum(am_month), total * total_days_in_month)
                data['pm'] = self.formatPercent(
                    sum(pm_month), total * total_days_in_month)
                data['all_day'] = self.formatPercent(
                    sum(all_day_month), total * total_days_in_month)
                data['am_blank'] = self.formatPercent(
                    sum(am_blank_month), total * total_days_in_month)
                data['pm_blank'] = self.formatPercent(
                    sum(pm_blank_month), total * total_days_in_month)
            result.append(data)
        return result


class CustomExportView(flourish.page.Page):

    content_template = ViewPageTemplateFile(
        'templates/custom_export.pt')
    refine_container_id = 'refine_container'
    form_container_id = 'students_table'
    container_class = 'container widecontainer'
    dialog_container_id = 'dialog-container'

    @Lazy
    def dialog_title(self):
        return translate(_('Export as XLS'), context=self.request)

    @Lazy
    def dialog_url(self):
        params = {
            'file_type': translate(STMessage('xls'),
                                   context=self.request).upper(),
            'description': translate(_('User-defined spreadsheet'),
                                     context=self.request),
            }
        return '%s/request_custom_export.html?%s' % (
            absoluteURL(self.context, self.request), urlencode(params))

    @Lazy
    def filters(self):
        return {
            'group_filter_id': 'group_filter',
            'level_filter_id': 'level_filter',
            'gender_filter_id': 'gender_filter',
            'course_filter_id': 'course_filter',
            'instructor_filter_id': 'instructor_filter',
            'section_filter_id': 'section_filter',
        }

    def update(self):
        zc.resourcelibrary.need('schooltool.skin.flourish-scripts')
        zc.resourcelibrary.need('schooltool.skin.flourish-report')
        zc.resourcelibrary.need('schooltool.flourish.custom')
        zc.resourcelibrary.need('schooltool.peas.scripts')

    @Lazy
    def schoolyear(self):
        app = ISchoolToolApplication(None)
        return ISchoolYearContainer(app).getActiveSchoolYear()

    @Lazy
    def collator(self):
        return ICollator(self.request.locale)

    @Lazy
    def group_container(self):
        return IGroupContainer(self.schoolyear)

    @Lazy
    def level_container(self):
        return ILevelContainer(ISchoolToolApplication(None))

    @Lazy
    def gender_vocabulary(self):
        return IBasicPerson['gender'].vocabulary

    @Lazy
    def course_container(self):
        return ICourseContainer(self.schoolyear)

    @Lazy
    def instructors(self):
        result = set()
        for term in self.schoolyear.values():
            sections = ISectionContainer(term)
            for section in sections.values():
                result.update(section.instructors)
        return dict([(person.__name__, person) for person in result])

    def section_key(self, section):
        term = ITerm(section)
        return '%s.%s' % (term.__name__, section.__name__)

    @Lazy
    def columns(self):
        result = {}
        for name, provider in getAdapters((self,), IColumnProvider):
            result[name] = OrderedDict([(c.name, c)
                                        for c in provider.columns])
        return result


class TitleRefineViewlet(flourish.page.Refine):

    template = flourish.templates.Inline('''
      <div class="header"
           tal:condition="view/title"
           tal:content="view/title">
        [ Filter title ]
      </div>
    ''')


class TitledFilterViewlet(flourish.page.Refine):

    template = flourish.templates.Inline('''
      <div class="header custom-export-filter-header"
           tal:condition="view/title"
           tal:content="view/title">
        [ Filter title ]
      </div>
      <div class="body" tal:content="structure view/body_template">
        [ options ]
      </div>
    ''')


class SelectFilterViewlet(TitledFilterViewlet):

    empty_option = _('--Select--')
    options = None
    html_id = None
    css_class = 'report-filter'

    @Lazy
    def selected_option(self):
        return self.request.get(self.html_id)

    options_template = flourish.templates.Inline('''
      <tal:block repeat="option view/options">
        <option tal:attributes="value option/value;
                                selected option/selected|nothing"
                tal:content="option/title" />
      </tal:block>
    ''')

    body_template = flourish.templates.Inline('''
      <select tal:attributes="name view/html_id;
                              class view/css_class;">
        <option tal:content="view/empty_option" value="" />
        <tal:block content="structure view/options_template" />
      </select>
    ''')


class IColumnProvider(Interface):

    columns = Attribute('Iterable with zc.table columns')


class IReportColumn(Interface):
    """A getter column for the custom export"""


class AddColumnViewlet(SelectFilterViewlet):

    prefix = 'add_column'
    column_provider_name = None
    css_class = 'report-filter add-column-menu'

    @Lazy
    def html_id(self):
        return '%s.%s' % (self.prefix, self.column_provider_name)

    def options(self):
        result = []
        for column in self.view.columns[self.column_provider_name].values():
            result.append({
                'title': column.title,
                'value': column.name,
            })
        return result


class GroupFilterViewlet(SelectFilterViewlet):

    @property
    def html_id(self):
        return self.view.filters['group_filter_id']

    @Lazy
    def options(self):
        result = []
        for group in sorted(self.view.group_container.values(),
                            key=lambda g: self.view.collator.key(g.title)):
            result.append({
                'title': group.title,
                'value': group.__name__,
                'selected': group.__name__ == self.selected_option,
            })
        return result


class LevelFilterViewlet(SelectFilterViewlet):

    @property
    def html_id(self):
        return self.view.filters['level_filter_id']

    @Lazy
    def options(self):
        result = []
        for level in self.view.level_container.values():
            result.append({
                'title': level.title,
                'value': level.__name__,
                'selected': level.__name__ == self.selected_option,
            })
        return result


class GenderFilterViewlet(SelectFilterViewlet):

    @property
    def html_id(self):
        return self.view.filters['gender_filter_id']

    @Lazy
    def options(self):
        result = []
        for term in self.view.gender_vocabulary:
            result.append({
                'title': term.title,
                'value': term.token,
                'selected': term.token == self.selected_option,
            })
        return result


class CourseFilterViewlet(SelectFilterViewlet):

    @property
    def html_id(self):
        return self.view.filters['course_filter_id']

    @Lazy
    def options(self):
        result = []
        for course in sorted(self.view.course_container.values(),
                             key=lambda c: self.view.collator.key(c.title)):
            result.append({
                'title': course.title,
                'value': course.__name__,
                'selected': course.__name__ == self.selected_option,
            })
        return result


class InstructorFilterViewlet(SelectFilterViewlet):

    @property
    def html_id(self):
        return self.view.filters['instructor_filter_id']

    @Lazy
    def options(self):
        result = []
        factory = getUtility(IPersonFactory)
        sorting_key = lambda x: factory.getSortingKey(x, self.view.collator)
        for instructor in sorted(self.view.instructors.values(),
                                 key=sorting_key):
            result.append({
                'title': instructor.title,
                'value': instructor.__name__,
                'selected': instructor.__name__ == self.selected_option,
            })
        return result


class SectionFilterViewlet(SelectFilterViewlet):

    @property
    def html_id(self):
        return self.view.filters['section_filter_id']

    options_template = flourish.templates.Inline('''
      <tal:block repeat="term view/sections_by_term">
        <option disabled="disabled"
                class="separator"
                tal:content="term/title" />
        <option tal:repeat="section term/sections"
                tal:attributes="value section/value;
                                selected section/selected"
                tal:content="section/title" />
      </tal:block>
    ''')

    @Lazy
    def sections_by_term(self):
        result = []
        sections = set()
        for term in sorted(self.view.schoolyear.values(),
                           key=lambda t: self.view.collator.key(t.title)):
            sections = []
            for section in sorted(ISectionContainer(term).values(),
                                  key=lambda s: self.view.collator.key(s.title)):
                section_key = self.view.section_key(section)
                sections.append({
                    'title': section.title,
                    'value': section_key,
                    'selected': section_key == self.selected_option,
                })
            result.append({
                'title': term.title,
                'sections': sections,
            })
        return result


class CustomExportXLSView(ExcelExportView, CustomExportView):

    message_title = _('custom export')
    sheet_title = _('Custom Export')

    @property
    def base_filename(self):
        return 'custom_export'

    def export_report(self, wb):
        self.task_progress.force('report', active=True)
        ws = wb.add_sheet(self.sheet_title)
        manager = flourish.content.queryContentProvider(self.context,
                                                        self.request,
                                                        self,
                                                        'custom_export_table')
        manager.batch_size = 0
        manager.update()
        formatter = manager.makeFormatter()
        self.print_headers(ws, formatter)
        self.print_report_rows(ws, formatter)
        self.finish('report')

    def print_headers(self, ws, formatter):
        headers = [column.title for column in formatter.columns]
        for index, header in enumerate(headers):
            self.write_header(ws, 0, index, header)

    def print_report_rows(self, ws, formatter):
        row = 1
        for person in formatter.getItems():
            for i, column in enumerate(formatter.columns):
                content = column.renderCell(person, formatter)
                if column.name == 'title':
                    content = column.getter(person, formatter)
                self.write(ws, row, i, content)
            row += 1

    def report_rows(self):
        return []

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb


class PersonsManageLink(ReportLinkViewlet, ActiveSchoolYearContentMixin):

    view_name = None

    @property
    def report_link(self):
        persons_url = absoluteURL(self.context['persons'], self.request)
        return '%s/%s' % (persons_url, self.view_name)

    def render(self, *args, **kw):
        if self.schoolyear is None:
            return ''
        return super(PersonsManageLink, self).render(*args, **kw)


class CustomExportLink(PersonsManageLink):

    view_name = 'custom_export.html'


class DemographicsSnapshotLink(PersonsManageLink):

    view_name = 'request_demographics_snapshot.html'


class UnpaidStudentsLink(PersonsManageLink):

    view_name = 'request_fee_collection_rate.html'


class CollectionRateByWeekLink(PersonsManageLink):

    view_name = 'request_fee_collection_rate_by_week.html'


class CustomExportTableSortFormatter(table.ajax.AJAXFormSortFormatter):

    script_name = 'ST.report.on_section_report_sort'

    @Lazy
    def columns_by_name(self):
        res = {}
        for col in self.columns:
            res[col.name] = col
        return res


class CustomExportTable(table.ajax.Table):

    table_formatter = CustomExportTableSortFormatter
    batch_size = 0
    _state = None

    @property
    def html_id(self):
        return self.view.form_container_id

    def update(self):
        if 'DELETE_LAST_COLUMN' in self.request and self.state:
            self._state = self._state[:-1]
        super(CustomExportTable, self).update()

    @property
    def state(self):
        if self._state is None:
            if 'viewstate' in self.request:
                decoder = flourish.tal.JSONDecoder()
                self._state = decoder.decode(self.request['viewstate'])
            else:
                self._state = []
        return self._state

    def columns(self):
        default = super(CustomExportTable, self).columns()
        requested_columns = []
        for info in self.state:
            column_provider_name = info['provider']
            column_name = info['name']
            column = self.view.columns[column_provider_name][column_name]
            requested_columns.append(column)
        for key in self.request.form:
            if key.startswith('add_column') and self.request[key]:
                prefix, column_provider_name = key.split('.')
                column_name = self.request[key]
                column = self.view.columns[column_provider_name][column_name]
                requested_columns.append(column)
                self.state.append({
                    'provider': column_provider_name,
                    'name': column_name,
                })
                break
        return default + requested_columns


class CustomExportTableFilter(table.ajax.TableFilter):

    def filter(self, items):
        result = []
        filter_handlers = []
        for filter_id in self.view.filters.values():
            requested_value = self.request.get(filter_id)
            if requested_value:
                getter = getattr(self, 'get_' + filter_id.replace('_filter', ''))
                selected_value = getter(requested_value)
                if selected_value is not None:
                    filter_handlers.append(
                        (getattr(self, 'handle_' + filter_id), selected_value)
                    )
        for item in items:
            if is_student(item):
                ommit = False
                for handler, selected_value in filter_handlers:
                    if handler(item, selected_value) == False:
                        ommit = True
                        break
                if not ommit:
                    result.append(item)
        return result

    def render(self, *args, **kw):
        return ''

    def get_group(self, group_id):
        return self.view.group_container.get(group_id)

    def handle_group_filter(self, person, group):
        return person in group.members

    def get_level(self, level_id):
        return self.view.level_container.get(level_id)

    def handle_level_filter(self, person, level):
        return level in person.levels

    def get_gender(self, gender_id):
        return gender_id

    def handle_gender_filter(self, person, gender):
        return person.gender == gender

    def get_course(self, course_id):
        course = self.view.course_container.get(course_id)
        if course:
            result = set()
            terms = self.view.schoolyear.values()
            for section in course.sections:
                if ITerm(section) in terms:
                    result.update(section.members)
            if result:
                return result

    def handle_course_filter(self, person, section_members):
        return person in section_members

    def get_instructor(self, instructor_id):
        instructor = self.view.instructors.get(instructor_id)
        if instructor:
            result = set()
            terms = self.view.schoolyear.values()
            relationship = Instruction.bind(instructor=instructor)
            for link_info in relationship.relationships:
                section = removeSecurityProxy(link_info.target)
                if ITerm(section) in terms:
                    result.update(section.members)
            if result:
                return result

    def handle_instructor_filter(self, person, section_members):
        return person in section_members

    def get_section(self, section_key):
        section = None
        for term in self.view.schoolyear.values():
            sections = ISectionContainer(term)
            for section in sections.values():
                if self.view.section_key(section) == section_key:
                    break
        if section:
            return section.members

    def handle_section_filter(self, person, section_members):
        return person in section_members


class CustomExportScript(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile(
        'templates/custom_export_script.pt')


class LevelAddColumnViewlet(AddColumnViewlet):

    column_provider_name = 'level'


class DetailsAddColumnViewlet(AddColumnViewlet):

    column_provider_name = 'details'


class DemographicsAddColumnViewlet(AddColumnViewlet):

    column_provider_name = 'demographics'


class AttendanceAddColumnViewlet(AddColumnViewlet):

    column_provider_name = 'attendance'

    options_template = flourish.templates.Inline('''
      <tal:block repeat="period view/scores_by_period">
        <option disabled="disabled"
                class="separator"
                tal:content="period/title" />
        <option tal:repeat="score period/scores"
                tal:attributes="value score/value;"
                tal:content="score/title" />
      </tal:block>
    ''')

    @Lazy
    def scores_by_period(self):
        result = []
        periods = [
            (_('AM'), StudentAMRequirement),
            (_('PM'), StudentPMRequirement),
        ]
        for title, period in periods:
            scoresystem = period.score_system
            scores = []
            for score, abbr in scoresystem.scores:
                scores.append({
                    'title': '%s (%s)' % (score, abbr),
                    'value': '%s.%s' % (period.requirement_type, score),
                })
            result.append({
                'title': title,
                'scores': scores,
            })
        return result


class ReportSheetsAddColumnViewlet(AddColumnViewlet):

    column_provider_name = 'report_sheets'


class ColumnProvider(object):

    implements(IColumnProvider)
    adapts(flourish.page.Page)

    def __init__(self, view):
        self.context = view

    def columns(self):
        raise NotImplemented()


class ReportColumn(GetterColumn):

    implements(IReportColumn)


class LevelColumnProvider(ColumnProvider):

    @Lazy
    def columns(self):
        return [
            ReportColumn(
                name='level',
                title=_('Grade Level'),
                getter=self.get_person_level)
        ]

    def get_person_level(self, person, formatter):
        result = []
        today = getUtility(IDateManager).today
        for level in person.levels.on(today).any(ACTIVE):
            result.append(level)
        return ', '.join([level.title for level in result])


class DetailsColumnProvider(ColumnProvider):

    @Lazy
    def columns(self):
        result = []
        fields = ['gender', 'birth_date']
        for name in fields:
            result.append(ReportColumn(
                name=name,
                title=IBasicPerson[name].title,
                getter=self.get_person_detail(name)
                ))
        return result

    def get_person_detail(self, field):
        def getter(person, formatter):
            return getattr(person, field, None) or ''
        return getter


class DemographicsColumnProvider(ColumnProvider):

    @Lazy
    def columns(self):
        result = []
        limit_keys = ['students']
        dfs = IDemographicsFields(ISchoolToolApplication(None))
        for df in dfs.filter_keys(limit_keys):
            name = df.name
            if name not in LEAVE_SCHOOL_FIELDS:
                result.append(ReportColumn(
                    name=name,
                    title=df.title,
                    getter=self.get_person_demographics(name)
                    ))
        return result

    def get_person_demographics(self, field):
        def getter(person, formatter):
            return IDemographics(person).get(field) or ''
        return getter


class AttendanceColumnProvider(ColumnProvider):

    @Lazy
    def columns(self):
        result = []
        periods = [
            (_('AM'), StudentAMRequirement),
            (_('PM'), StudentPMRequirement),
        ]
        for title, period in periods:
            scoresystem = period.score_system
            for score, abbr in scoresystem.scores:
                result.append(ReportColumn(
                    name='%s.%s' % (period.requirement_type, score),
                    title='%s - %s' % (title, abbr),
                    getter=self.get_person_attendance_score(period, score)
                ))
        return result

    @Lazy
    def requirements(self):
        result = {
            StudentAMRequirement: [],
            StudentPMRequirement: [],
        }
        for term in self.context.schoolyear.values():
            for date in term:
                meeting = makeSchoolAttendanceMeeting(date)
                for requirement in result.keys():
                    result[requirement].append(requirement(meeting))
        return result

    def get_person_attendance_score(self, period, requested_score):
        def getter(person, formatter):
            scores = []
            for requirement in self.requirements[period]:
                score = IEvaluateRequirement(requirement).getEvaluation(
                    person, requirement, default=UNSCORED)
                if (score is UNSCORED or
                    score.value is UNSCORED or
                    score.value != requested_score):
                    continue
                scores.append(score)
            return len(scores) or ''
        return getter


def getMinMaxValue(score):
    ss = score.scoreSystem
    if IDiscreteValuesScoreSystem.providedBy(ss):
        return (ss.scores[-1][2], ss.scores[0][2],
            ss.getNumericalValue(score.value))
    elif IRangedValuesScoreSystem.providedBy(ss):
        return ss.min, ss.max, score.value
    return None, None, None


class ReportSheetsColumnProvider(ColumnProvider):

    @Lazy
    def current_columns(self):
        app = ISchoolToolApplication(None)
        root = IGradebookRoot(app)
        schoolyearKey = self.context.schoolyear.__name__
        columns = []
        if schoolyearKey in root.layouts:
            columns = root.layouts[schoolyearKey].columns
        result = []
        for column in columns:
            parts = column.source.split('|')
            if len(parts) != 3:
                continue
            term_id, worksheet_id, activity_id = parts
            if term_id not in self.context.schoolyear:
                continue
            worksheet = root.deployed.get(worksheet_id)
            if worksheet is not None:
                activity = worksheet.get(activity_id)
                if (activity is not None and
                    not IDiscreteValuesScoreSystem.providedBy(activity.scoresystem)):
                    result.append(column)
        return result

    @Lazy
    def columns(self):
        result = []
        app = ISchoolToolApplication(None)
        layout_view = FlourishLayoutReportCardView(app, self.context.request)
        for index, column in enumerate(self.current_columns):
            sourceName = layout_view.getSourceName(column.source)
            if sourceName is None:
                continue
            result.append(ReportColumn(
                name='Column%d' % index,
                title=column.heading or sourceName,
                getter=self.get_report_activity_average(index),
            ))
        return result

    def get_report_activity_average(self, column_index):
        def getter(person, formatter):
            evaluations = IEvaluations(person)
            try:
                column = self.current_columns[column_index]
            except (IndexError,):
                column = None
            if column is not None:
                source = column.source
                termName, worksheetName, activityName = source.split('|')
                term = self.context.schoolyear.get(termName)
                average_scoresystem = None
                if term is not None:
                    total = 0
                    count = 0
                    for section in ILearner(person).sections():
                        if ITerm(section) == term:
                            activities = IActivities(section)
                            worksheet = activities.get(worksheetName)
                            if worksheet is not None:
                                activity = worksheet.get(activityName)
                                if activity is not None:
                                    average_scoresystem = activity.scoresystem
                                    score = evaluations.get(activity)
                                    if not score:
                                        continue
                                    minimum, maximum, value = getMinMaxValue(score)
                                    if minimum is None:
                                        continue
                                    total += value
                                    count += maximum
                    if count:
                        average = Decimal(100 * total) / Decimal(count)
                        return convertAverage(average, average_scoresystem)
            return ''
        return getter


class CustomExportFields(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/custom_export_fields.pt')


class RequestCustomExportView(RequestXLSReportDialog,
                              CustomExportView):

    report_builder = 'custom_export.xls'
    update = RequestXLSReportDialog.update
    render = RequestXLSReportDialog.render

    template = ViewPageTemplateFile(
        'templates/request_custom_export_form.pt')

    @Lazy
    def hidden_fields(self):
        result = []
        for filter_id in self.filters.values():
            requested_value = self.request.get(filter_id)
            if requested_value:
                result.append({
                    'name': filter_id,
                    'value': requested_value,
                })
        if 'viewstate' in self.request:
            result.append({
                'name': 'viewstate',
                'value': self.request['viewstate'],
            })
        return result

    def updateTaskParams(self, task):
        for info in self.hidden_fields:
            task.request_params[info['name']] = info['value']


class CustomExportRefineFormManager(flourish.page.ContentViewletManager):

    template = flourish.templates.Inline("""
        <form method="get"
              tal:attributes="id view/view/refine_container_id">
          <tal:block repeat="viewlet view/viewlets">
            <div class="content"
                 tal:define="rendered viewlet;
                             stripped rendered/strip|nothing"
                 tal:condition="stripped"
                 tal:content="structure stripped">
            </div>
          </tal:block>
        </form>
    """)


class DemographicsSnapshotRequestView(RequestXLSReportDialog):

    report_builder = 'demographics_snapshot.xls'


class DemographicsSnapshotXLSView(ExcelExportView):

    message_title = _('demographics snapshot report')
    sheet_title = _('Demographics Snapshot')

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    @Lazy
    def style_column_title(self):
        style = xlwt.XFStyle()
        font = xlwt.Font()
        font.bold = True
        style.font = font
        alignment = xlwt.Alignment()
        alignment.horz = xlwt.Alignment.HORZ_CENTER
        alignment.vert = xlwt.Alignment.VERT_CENTER
        style.alignment = alignment
        borders = xlwt.Borders()
        black = xlwt.Style.colour_map['black']
        borders.left = xlwt.Borders.THIN
        borders.top = xlwt.Borders.THIN
        borders.right = xlwt.Borders.THIN
        borders.bottom = xlwt.Borders.THIN
        borders.left_colour = black
        borders.top_colour = black
        borders.right_colour = black
        borders.bottom_colour = black
        style.borders = borders
        return style

    @Lazy
    def style_row_title(self):
        style = xlwt.XFStyle()
        font = xlwt.Font()
        style.font = font
        alignment = xlwt.Alignment()
        alignment.horz = xlwt.Alignment.HORZ_LEFT
        alignment.vert = xlwt.Alignment.VERT_CENTER
        style.alignment = alignment
        borders = xlwt.Borders()
        black = xlwt.Style.colour_map['black']
        borders.left = xlwt.Borders.THIN
        borders.top = xlwt.Borders.THIN
        borders.right = xlwt.Borders.THIN
        borders.bottom = xlwt.Borders.THIN
        borders.left_colour = black
        borders.top_colour = black
        borders.right_colour = black
        borders.bottom_colour = black
        style.borders = borders
        return style

    @Lazy
    def style_cell(self):
        style = xlwt.XFStyle()
        font = xlwt.Font()
        style.font = font
        borders = xlwt.Borders()
        black = xlwt.Style.colour_map['black']
        borders.left = xlwt.Borders.THIN
        borders.top = xlwt.Borders.THIN
        borders.right = xlwt.Borders.THIN
        borders.bottom = xlwt.Borders.THIN
        borders.left_colour = black
        borders.top_colour = black
        borders.right_colour = black
        borders.bottom_colour = black
        style.borders = borders
        return style

    @property
    def base_filename(self):
        return 'demographics_snapshot'

    @Lazy
    def levels(self):
        app = ISchoolToolApplication(None)
        return ILevelContainer(app).values()

    def get_students_by_gender(self, level):
        result = {'boys': set(), 'girls': set()}
        rel = StudentLevels.bind(level=level).relationships
        for link_info in rel:
            person = link_info.target
            if person.gender == 'male':
                result['boys'].add(person)
            elif person.gender == 'female':
                result['girls'].add(person)
        return result

    level_total_title = _('Total School')

    @Lazy
    def levels_data(self):
        result = []
        all_boys = set()
        all_girls = set()
        for level in self.levels:
            by_gender = self.get_students_by_gender(level)
            all_boys = all_boys.union(by_gender['boys'])
            all_girls = all_girls.union(by_gender['girls'])
            result.append({
                'title': level.title,
                'boys': by_gender['boys'],
                'girls': by_gender['girls'],
                })
        result.append({
            'title': self.level_total_title,
            'boys': all_boys,
            'girls': all_girls,
            })
        return result

    def __call__(self):
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def export_report(self, wb):
        self.task_progress.force('report', active=True)
        ws = wb.add_sheet(self.sheet_title)
        self.print_headers(ws)
        self.print_row_titles(ws)
        self.print_report_cells(ws)
        self.finish('report')

    def print_headers(self, ws):
        self.print_levels(ws)
        self.print_column_titles(ws)

    def print_levels(self, ws):
        style = self.style_column_title
        merge = 2
        row = 0
        col = 1
        for info in self.levels_data:
            ws.write_merge(row, row, col, col + merge, info['title'], style)
            col += (merge + 1)

    def print_column_titles(self, ws):
        style = self.style_column_title
        row = 1
        col = 1
        for info in self.levels_data[:-1]:
            ws.write(row, col, _('Girls'), style)
            ws.write(row, col + 1, _('Boys'), style)
            title = translate(_('Total ${level}',
                                mapping={'level': info['title']}),
                              context=self.request)
            ws.write(row, col + 2, title, style)
            col += 3
        ws.write(row, col, _('All girls'), style)
        ws.write(row, col + 1, _('All boys'), style)
        ws.write(row, col + 2, _('Total School'), style)

    def print_row_titles(self, ws):
        style = self.style_row_title
        row = 2
        col = 0
        ws.write(row, col, _('Total number'), style)
        ws.write(row + 1, col, _('Gender breakdown (%)'), style)
        ws.write(row + 2, col, _('% Boarders'), style)
        ws.write(row + 3, col, _('Average PLE scores'), style)
        ws.write(row + 4, col, _('Average Age'), style)
        ws.write(row + 5, col, _('Average distance travelled to school'), style)
        ws.write(row + 6, col, _('% with Disability'), style)
        ws.write(row + 7, col, _('% orphans'), style)

    def print_report_cells(self, ws):
        style = self.style_cell
        row = 2
        col = 1
        self.print_row_total_number(ws, row, col, style)
        self.print_row_gender_breakdown(ws, row + 1, col, style)
        self.print_row_percent_boarders(ws, row + 2, col, style)
        self.print_row_average_ple_scores(ws, row + 3, col, style)
        self.print_row_average_age(ws, row + 4, col, style)
        self.print_row_average_distance(ws, row + 5, col, style)
        self.print_row_percent_disability(ws, row + 6, col, style)
        self.print_row_percent_orphans(ws, row + 7, col, style)

    def div(self, count, total):
        if not total:
            return 0.0
        return count/float(total)

    def print_row_total_number(self, ws, row, col, style):
        for info in self.levels_data:
            girls = len(info['girls'])
            boys = len(info['boys'])
            total = girls + boys
            ws.write(row, col, girls, style)
            ws.write(row, col + 1, boys, style)
            ws.write(row, col + 2, total, style)
            col += 3

    def print_row_gender_breakdown(self, ws, row, col, style):
        for info in self.levels_data:
            girls = len(info['girls'])
            boys = len(info['boys'])
            total = girls + boys
            ws.write(row, col, self.div(girls * 100, total), style)
            ws.write(row, col + 1, self.div(boys * 100, total) , style)
            ws.write(row, col + 2, self.div(total * 100, total), style)
            col += 3

    def print_row_percent_boarders(self, ws, row, col, style):
        get_boarding = lambda x: IDemographics(x).get('day_boarding') or ''
        is_boarding = lambda x: get_boarding(x).lower() == 'boarding'
        for info in self.levels_data:
            girls = len([p for p in info['girls'] if is_boarding(p)])
            boys = len([p for p in info['boys'] if is_boarding(p)])
            total = len(info['girls']) + len(info['boys'])
            boarding = girls + boys
            ws.write(row, col,
                     self.div(girls * 100, len(info['girls'])), style)
            ws.write(row, col + 1,
                     self.div(boys * 100, len(info['boys'])) , style)
            ws.write(row, col + 2,
                     self.div(boarding * 100, total), style)
            col += 3

    def print_row_average_ple_scores(self, ws, row, col, style):
        get_score = lambda x: IDemographics(x).get('ple_score') or 0
        for info in self.levels_data:
            girls = [get_score(p) for p in info['girls']]
            boys = [get_score(p) for p in info['boys']]
            total = girls + boys
            ws.write(row, col,
                     self.div(sum(girls), len(info['girls'])), style)
            ws.write(row, col + 1,
                     self.div(sum(boys), len(info['boys'])), style)
            ws.write(row, col + 2,
                     self.div(sum(total), len(total)), style)
            col += 3

    def print_row_average_age(self, ws, row, col, style):
        def get_age(birth_date):
            today = getUtility(IDateManager).today
            return (
                today.year -
                birth_date.year -
                ((today.month, today.day) < (birth_date.month, birth_date.day)))
        for info in self.levels_data:
            girls = [get_age(p.birth_date) for p in info['girls']
                           if p.birth_date is not None]
            boys = [get_age(p.birth_date) for p in info['boys']
                          if p.birth_date is not None]
            total = len(info['girls']) + len(info['boys'])
            ws.write(row, col,
                     self.div(sum(girls), len(info['girls'])), style)
            ws.write(row, col + 1,
                     self.div(sum(boys), len(info['boys'])), style)
            ws.write(row, col + 2,
                     self.div(sum(girls + boys), total), style)
            col += 3

    def print_row_average_distance(self, ws, row, col, style):
        get_distance = lambda x: IDemographics(x).get('distance') or 0
        for info in self.levels_data:
            girls = [get_distance(p) for p in info['girls']]
            boys = [get_distance(p) for p in info['boys']]
            total = girls + boys
            ws.write(row, col,
                     self.div(sum(girls), len(info['girls'])), style)
            ws.write(row, col + 1,
                     self.div(sum(boys), len(info['boys'])), style)
            ws.write(row, col + 2,
                     self.div(sum(total), len(total)), style)
            col += 3

    def print_row_percent_disability(self, ws, row, col, style):
        get_disability = lambda x: IDemographics(x).get('disability') or ''
        is_disabled = lambda x: 'yes' in get_disability(x).lower()
        for info in self.levels_data:
            girls = len([p for p in info['girls'] if is_disabled(p)])
            boys = len([p for p in info['boys'] if is_disabled(p)])
            total = len(info['girls']) + len(info['boys'])
            disabled = girls + boys
            ws.write(row, col,
                     self.div(girls * 100, len(info['girls'])), style)
            ws.write(row, col + 1,
                     self.div(boys * 100, len(info['boys'])) , style)
            ws.write(row, col + 2,
                     self.div(disabled * 100, total), style)
            col += 3

    def print_row_percent_orphans(self, ws, row, col, style):
        get_parental_status = lambda x: IDemographics(x).get('parental_status') or ''
        is_orphan = lambda x: 'both dead' in get_parental_status(x).lower()
        for info in self.levels_data:
            girls = len([p for p in info['girls'] if is_orphan(p)])
            boys = len([p for p in info['boys'] if is_orphan(p)])
            total = len(info['girls']) + len(info['boys'])
            orphan = girls + boys
            ws.write(row, col,
                     self.div(girls * 100, len(info['girls'])), style)
            ws.write(row, col + 1,
                     self.div(boys * 100, len(info['boys'])) , style)
            ws.write(row, col + 2,
                     self.div(orphan * 100, total), style)
            col += 3


class IDateRangeForm(Interface):

    start = Date(
        title=STMessage(u'Start date'),
        required=True)

    end = Date(
        title=STMessage(u'End date'),
        required=True)


class IBoardingForm(Interface):

    boarding = Choice(
        title=_('Day/Boarding'),
        source='schooltool.peas.boarding',
        required=False)


class BoardingVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        app = ISchoolToolApplication(None)
        dfs = IDemographicsFields(app)
        demo = dfs.get('day_boarding')
        if demo is not None:
            form_field = demo.makeField()['day_boarding']
            vocabulary = form_field.field.vocabulary
            for term in vocabulary:
                result.append(self.createTerm(
                    term.value,
                    term.token,
                    term.title,
                ))
        return result


def BoardingVocabularyFactory():
    return BoardingVocabulary


class AttendanceRateByDayReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(IDateRangeForm, IBoardingForm).select(
        'start', 'end', 'boarding')

    report_builder = 'attendance_rate_by_day.xls'

    def updateWidgets(self):
        super(AttendanceRateByDayReportRequestView, self).updateWidgets()
        self.widgets['boarding'].prompt = True
        self.widgets['boarding'].promptMessage = _('All')

    def updateTaskParams(self, task):
        start = self.form_params.get('start')
        end = self.form_params.get('end')
        if start > end:
            start, end = end, start
        if start is not None:
            day = '%d-%02d-%02d' % (start.year, start.month, start.day)
            task.request_params['start'] = day
        if end is not None:
            day = '%d-%02d-%02d' % (end.year, end.month, end.day)
            task.request_params['end'] = day
        task.request_params['boarding'] = self.form_params.get('boarding') or _('All')


class AttendanceRateByDayView(StudentAttendanceDisaggregatedByDayView,
                              CalendarViewBase):

    message_title = _('attendance rate by day report')
    sheet_title = _('Attendance Rate by Day')

    @property
    def base_filename(self):
        start = self.getDay('start')
        end = self.getDay('end')
        if start is None or end is None:
            filename = 'attendance_rate_by_day'
        else:
            filename = 'attendance_rate_by_day_%d_%02d_%02d_to_%d_%02d_%02d' % (
                start.year, start.month, start.day,
                end.year, end.month, end.day)
        return filename

    def getDay(self, name):
        day = self.request.get(name, None)
        if day is None:
            return datetime.date(datetime.now())
        try:
            year, month, day = [int(part) for part in day.split('-')]
            return datetime.date(year, month, day)
        except (TypeError, ValueError):
            pass
        try:
            return datetime.date(day)
        except (TypeError, ValueError):
            return None

    def print_headers(self, ws):
        self.write_header(ws, 0, 0, _('Day/Boarding'))
        self.write(ws, 0, 1, self.request.get('boarding'))
        headers = [
            'Term',
            'Week',
            'Date',
            'Attendance rate for girls (%)',
            'Attendance rate for boys (%)',
            'Attendance rate for all students (%)',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 2, index, header)

    def print_report_rows(self, ws):
        row = 3
        for data in self.report_rows():
            self.write(ws, row, 0, data.get('term'))
            self.write(ws, row, 1, data.get('week'))
            self.write(ws, row, 2, data.get('date'))
            self.write(ws, row, 3, data.get('girls'))
            self.write(ws, row, 4, data.get('boys'))
            self.write(ws, row, 5, data.get('all'))
            row += 1

    def getTerm(self, date):
        for term in self.terms:
            if date in term:
                return term

    def getWeekNumber(self, day, term):
        result = 0
        current_week = None
        for date in term:
            week = self.getWeek(date)
            if week != current_week:
                result += 1
                current_week = week
            if date == day:
                break
        return result

    def report_rows(self):
        result = []
        total = len(self.persons)
        today = getUtility(IDateManager).today
        boarding = self.request.get('boarding')
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            am_requirement = StudentAMRequirement(meeting)
            pm_requirement = StudentPMRequirement(meeting)
            data = {}
            term = self.getTerm(date)
            if term is not None:
                data['term'] = term.title
            data['week'] = self.getWeekNumber(date, term)
            data['date'] = date.isoformat()
            girls = []
            boys = []
            am = []
            pm = []
            all_day = []
            am_blank = []
            pm_blank = []
            for person in self.persons:
                present_am = False
                present_pm = False
                score = IEvaluateRequirement(am_requirement).getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    am_blank.append(1)
                elif not score.scoreSystem.isAbsent(score):
                    am.append(1)
                    present_am = True
                score = IEvaluateRequirement(pm_requirement).getEvaluation(
                    person, pm_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pm_blank.append(1)
                elif not score.scoreSystem.isAbsent(score):
                    pm.append(1)
                    present_pm = True
                if present_am and present_pm:
                    person_boarding = IDemographics(person).get('day_boarding')
                    if (boarding.lower() != 'all' and
                        person_boarding != boarding):
                        continue
                    all_day.append(1)
                    if person in self.female:
                        girls.append(1)
                    elif person in self.male:
                        boys.append(1)
            if total:
                data['girls'] = self.formatPercent(sum(girls), total)
                data['boys'] = self.formatPercent(sum(boys), total)
                data['all'] = self.formatPercent(sum(all_day), total)
            result.append(data)
        return result

    @Lazy
    def terms(self):
        selected = set()
        start = self.getDay('start')
        end = self.getDay('end')
        for term in self.all_terms.values():
            if start in term or end in term:
                selected.add(term)
        return sorted(selected, key=lambda term: term.first)

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        start = self.getDay('start')
        end = self.getDay('end')
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date) and (start <= date <= end):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings


class ICurrentTerm(Interface):

    term = Choice(
        title=_('Term'),
        source='schooltool.peas.terms',
        required=True)


def term_id_getter(schoolyear, term):
    return '%s.%s' % (
        schoolyear.__name__.encode('utf-8'), term.__name__.encode('utf-8'))


class TermsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        result.append(self.createTerm(
            None,
            widget.SequenceWidget.noValueToken,
            _('Select a term'),
        ))
        app = ISchoolToolApplication(None)
        container = ISchoolYearContainer(app)
        for schoolyear_id, schoolyear in container.items():
            for term in sorted(schoolyear.values(), key=lambda t: t.first):
                result.append(self.createTerm(
                    term,
                    term_id_getter(schoolyear, term),
                    '%s - %s' % (schoolyear.title, term.title),
                ))
        return result


def TermsVocabularyFactory():
    return TermsVocabulary


class AttendanceRateByWeekReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(ICurrentTerm, IBoardingForm).select(
        'term', 'boarding')

    report_builder = 'attendance_rate_by_week.xls'

    def updateWidgets(self):
        super(AttendanceRateByWeekReportRequestView, self).updateWidgets()
        self.widgets['boarding'].prompt = True
        self.widgets['boarding'].promptMessage = _('All')

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)
        task.request_params['boarding'] = self.form_params.get('boarding') or _('All')


class AttendanceRateByWeekView(AttendanceRateByDayView):

    message_title = _('attendance rate by week report')
    sheet_title = _('Attendance Rate by Week')

    @property
    def base_filename(self):
        return 'attendance_rate_by_week'

    def print_headers(self, ws):
        self.write_header(ws, 0, 0, _('Term'))
        self.write(ws, 0, 1, self.view_term.title)
        self.write_header(ws, 1, 0, _('Day/Boarding'))
        self.write(ws, 1, 1, self.request.get('boarding'))
        headers = [
            'Week',
            'Attendance rate for girls (%)',
            'Attendance rate for boys (%)',
            'Attendance rate for all students (%)',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 3, index, header)

    def print_report_rows(self, ws):
        row = 4
        for data in self.report_rows():
            self.write(ws, row, 0, data.get('week'))
            self.write(ws, row, 1, data.get('girls'))
            self.write(ws, row, 2, data.get('boys'))
            self.write(ws, row, 3, data.get('all'))
            row += 1

    @Lazy
    def terms(self):
        result = []
        requested = self.request.get('term')
        for term in self.all_terms.values():
            if term_id_getter(ISchoolYear(term), term) == requested:
                result.append(term)
        return result

    @Lazy
    def view_term(self):
        return self.terms[0]

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    def getWeeks(self, day, term):
        result = []
        current_week = None
        for date in term:
            week = self.getWeek(date)
            if week != current_week:
                result.append(week)
                current_week = week
            if date == day:
                break
        return result

    def report_rows(self):
        result = []
        total = len(self.persons)
        today = getUtility(IDateManager).today
        weeks = self.getWeeks(today, self.view_term)
        meeting_dates = {}
        boarding = self.request.get('boarding')
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            meeting_dates[date] = meeting
        for week_no, week in enumerate(weeks):
            data = {}
            date = week[0].date
            data['week'] = translate(_('Week ${week} (w/b on ${date})',
                                       mapping={'week': week_no + 1,
                                                'date': date.isoformat()}),
                                     context=self.request)
            meeting_days_in_week = []
            girls_week = []
            boys_week = []
            all_day_week = []
            for day in week:
                girls = []
                boys = []
                am = []
                pm = []
                all_day = []
                am_blank = []
                pm_blank = []
                meeting = meeting_dates.get(day.date)
                if meeting is not None:
                    am_requirement = StudentAMRequirement(meeting)
                    pm_requirement = StudentPMRequirement(meeting)
                    meeting_days_in_week.append(1)
                    for person in self.persons:
                        present_am = False
                        present_pm = False
                        evaluations = IEvaluateRequirement(am_requirement)
                        score = evaluations.getEvaluation(
                            person, am_requirement, default=UNSCORED)
                        if (score is None or
                            score is UNSCORED or
                            score.value is UNSCORED):
                            am_blank.append(1)
                        elif not score.scoreSystem.isAbsent(score):
                            am.append(1)
                            present_am = True
                        evaluations = IEvaluateRequirement(pm_requirement)
                        score = evaluations.getEvaluation(
                            person, pm_requirement, default=UNSCORED)
                        if (score is None or
                            score is UNSCORED or
                            score.value is UNSCORED):
                            pm_blank.append(1)
                        elif not score.scoreSystem.isAbsent(score):
                            pm.append(1)
                            present_pm = True
                        if present_am and present_pm:
                            person_boarding = IDemographics(person).get('day_boarding')
                            if (boarding.lower() != 'all' and
                                person_boarding != boarding):
                                continue
                            all_day.append(1)
                            if person in self.female:
                                girls.append(1)
                            elif person in self.male:
                                boys.append(1)
                    girls_week.append(sum(girls))
                    boys_week.append(sum(boys))
                    all_day_week.append(sum(all_day))
            total_days_in_week = sum(meeting_days_in_week)
            if total and total_days_in_week:
                data['girls'] = self.formatPercent(
                    sum(girls_week), total * total_days_in_week)
                data['boys'] = self.formatPercent(
                    sum(boys_week), total * total_days_in_week)
                data['all'] = self.formatPercent(
                    sum(all_day_week), total * total_days_in_week)
            result.append(data)
        return result


class AttendanceRateByTermReportRequestView(RequestXLSReportDialog,
                                            FlourishSchoolAttendanceView):

    fields = field.Fields(IRequestYearlyReport, IBoardingForm).select(
        'schoolyear', 'boarding')

    report_builder = 'attendance_rate_by_term.xls'

    update = RequestXLSReportDialog.update
    render = RequestXLSReportDialog.render

    def updateWidgets(self):
        super(AttendanceRateByTermReportRequestView, self).updateWidgets()
        self.widgets['boarding'].prompt = True
        self.widgets['boarding'].promptMessage = _('All')

    def updateTaskParams(self, task):
        schoolyear = self.form_params.get('schoolyear')
        if schoolyear is not None:
            task.request_params['schoolyear'] = schoolyear.__name__
        task.request_params['boarding'] = self.form_params.get('boarding') or _('All')


class AttendanceRateByTermView(AttendanceRateByWeekView):

    message_title = _('attendance rate by term report')
    sheet_title = _('Attendance Rate by Term')

    @Lazy
    def schoolyear(self):
        app = ISchoolToolApplication(None)
        schoolyears = ISchoolYearContainer(app)
        return schoolyears.get(self.request.get('schoolyear'))

    @property
    def base_filename(self):
        return 'attendance_rate_by_term_%s' % self.schoolyear.title

    def print_headers(self, ws):
        self.write_header(ws, 0, 0, _('School Year'))
        self.write(ws, 0, 1, self.schoolyear.title)
        self.write_header(ws, 1, 0, _('Day/Boarding'))
        self.write(ws, 1, 1, self.request.get('boarding'))
        headers = [
            'Term',
            'Attendance rate for girls (%)',
            'Attendance rate for boys (%)',
            'Attendance rate for all students (%)',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 3, index, header)

    def print_report_rows(self, ws):
        row = 4
        for data in self.report_rows():
            self.write(ws, row, 0, data.get('term'))
            self.write(ws, row, 1, data.get('girls'))
            self.write(ws, row, 2, data.get('boys'))
            self.write(ws, row, 3, data.get('all'))
            row += 1

    @Lazy
    def terms(self):
        return sorted(self.schoolyear.values(), key=lambda t: t.first)

    def allMeetings(self, term):
        meetings = []
        for date in term:
            if term.isSchoolday(date):
                dt = self.tzinfo.localize(
                    datetime.datetime(date.year, date.month, date.day))
                meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    def report_rows(self):
        result = []
        total = len(self.persons)
        today = getUtility(IDateManager).today
        boarding = self.request.get('boarding')
        for term in self.terms:
            term_meetings = self.allMeetings(term)
            data = {}
            data['term'] = term.title
            girls_term = []
            boys_term = []
            all_day_term = []
            total_days_in_term = len(term_meetings)
            for meeting in term_meetings:
                date = meeting.dtstart.astimezone(self.tzinfo).date()
                if date > today:
                    continue
                girls = []
                boys = []
                am = []
                pm = []
                all_day = []
                am_blank = []
                pm_blank = []
                am_requirement = StudentAMRequirement(meeting)
                pm_requirement = StudentPMRequirement(meeting)
                for person in self.persons:
                    present_am = False
                    present_pm = False
                    evaluations = IEvaluateRequirement(am_requirement)
                    score = evaluations.getEvaluation(
                        person, am_requirement, default=UNSCORED)
                    if (score is None or
                        score is UNSCORED or
                        score.value is UNSCORED):
                        am_blank.append(1)
                    elif not score.scoreSystem.isAbsent(score):
                        am.append(1)
                        present_am = True
                    evaluations = IEvaluateRequirement(pm_requirement)
                    score = evaluations.getEvaluation(
                        person, pm_requirement, default=UNSCORED)
                    if (score is None or
                        score is UNSCORED or
                        score.value is UNSCORED):
                        pm_blank.append(1)
                    elif not score.scoreSystem.isAbsent(score):
                        pm.append(1)
                        present_pm = True
                    if present_am and present_pm:
                        person_boarding = IDemographics(person).get('day_boarding')
                        if (boarding.lower() != 'all' and
                            person_boarding != boarding):
                            continue
                        all_day.append(1)
                        if person in self.female:
                            girls.append(1)
                        elif person in self.male:
                            boys.append(1)
                girls_term.append(sum(girls))
                boys_term.append(sum(boys))
                all_day_term.append(sum(all_day))
            if total and total_days_in_term:
                data['girls'] = self.formatPercent(
                    sum(girls_term), total * total_days_in_term)
                data['boys'] = self.formatPercent(
                    sum(boys_term), total * total_days_in_term)
                data['all'] = self.formatPercent(
                    sum(all_day_term), total * total_days_in_term)
            result.append(data)
        return result


class AbsenteeismReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(ICurrentTerm)

    report_builder = 'absenteeism_report.xls'

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)


class AbsenteeismReportView(ExcelExportView):

    message_title = _('absenteeism report')

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    @property
    def base_filename(self):
        return 'absenteeism_report'

    @Lazy
    def terms(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear in ISchoolYearContainer(app).values():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return [term]

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def export_report(self, wb):
        self.task_progress.force('report', active=True)

        ws1 = wb.add_sheet(_('Overall'))
        self.print_headers(ws1, _('Students with less than 75% attendance'))
        self.print_report_overall(ws1)

        ws2 = wb.add_sheet(_('3 absences in last 10 days'))
        self.print_headers(
            ws2, _('Students absent more than 3 days in the last 10 day period'))
        self.print_report_recent(ws2)

        self.finish('report')

    def print_headers(self, ws, title):
        ws.write_merge(0, 0, 0, 7, title)
        self.write_header(ws, 2, 0, _('Term'))
        self.write(ws, 2, 1, self.terms[0].title)
        headers = [
            'Class',
            'Student',
            'Day/Boarding',
            'Average attendance this term',
            'Number days missed in the last 10 days',
            'Fee collection rate',
            'Parent name',
            'Parent contact number',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 4, index, header)

    def print_report_overall(self, ws):
        row = 5
        today = getUtility(IDateManager).today
        meetings = [{'am': StudentAMRequirement(meeting),
                     'pm': StudentPMRequirement(meeting)}
                     for meeting in self.allMeetings()
                    if meeting.dtstart.astimezone(self.tzinfo).date() <= today]
        recent = meetings[-10:]
        total_meetings = len(meetings)
        for person in self.persons:
            person_all_day = []
            person_recent = []
            for info in meetings:
                am_requirement = info['am']
                pm_requirement = info['pm']
                present_am = False
                present_pm = False
                score = IEvaluateRequirement(am_requirement).getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_am = True
                score = IEvaluateRequirement(pm_requirement).getEvaluation(
                    person, pm_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_pm = True
                if present_am and present_pm:
                    person_all_day.append(1)
                    if info in recent:
                        person_recent.append(1)
            if total_meetings:
                average = (sum(person_all_day) * 100.0) / total_meetings
                missed_recent = len(recent) - sum(person_recent)
                if average < 75.0:
                    self.print_person(ws, row, person, average, missed_recent)
                    row += 1

    def getParents(self, person):
        result = set()
        for contact in IContactable(person).contacts.any(meaning=PARENT):
            result.add(contact)
        return sorted(result, key=lambda c: c.title)

    def print_person(self, ws, row, person, average, missed_recent):
        term = self.terms[0]
        credits = [fee.amount for fee in IFeeContainer(person).values()
                   if term.first <= fee.date <= term.last and
                   ICredit.providedBy(fee)]
        debits = [fee.amount for fee in IFeeContainer(person).values() 
                  if term.first <= fee.date <= term.last and
                  IDebit.providedBy(fee)]
        fee_collection_rate = 0
        if debits:
            fee_collection_rate = (sum(credits) * 100)/ sum(debits)
        self.write(ws, row, 1, person.title)
        self.write(ws, row, 2, IDemographics(person).get('day_boarding') or '')
        self.write(ws, row, 3, average)
        self.write(ws, row, 4, missed_recent)
        self.write(ws, row, 5, fee_collection_rate)
        parents = self.getParents(person)
        phones = lambda x: ', '.join(filter(None, [
            x.home_phone,
            x.work_phone,
            x.mobile_phone]))
        if parents:
            self.write(ws, row, 6, ', '.join([c.title for c in parents]))
            self.write(ws, row, 7, ', '.join([phones(c) for c in parents]))

    def print_report_recent(self, ws):
        row = 5
        today = getUtility(IDateManager).today
        meetings = [{'am': StudentAMRequirement(meeting),
                     'pm': StudentPMRequirement(meeting)}
                     for meeting in self.allMeetings()
                    if meeting.dtstart.astimezone(self.tzinfo).date() <= today]
        recent = meetings[-10:]
        total_meetings = len(meetings)
        for person in self.persons:
            person_all_day = []
            person_recent = []
            for info in meetings:
                am_requirement = info['am']
                pm_requirement = info['pm']
                present_am = False
                present_pm = False
                score = IEvaluateRequirement(am_requirement).getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_am = True
                score = IEvaluateRequirement(pm_requirement).getEvaluation(
                    person, pm_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_pm = True
                if present_am and present_pm:
                    person_all_day.append(1)
                    if info in recent:
                        person_recent.append(1)
            if total_meetings:
                average = (sum(person_all_day) * 100.0) / total_meetings
                missed_recent = len(recent) - sum(person_recent)
                if len(recent) >= 3 and missed_recent > 3:
                    self.print_person(ws, row, person, average, missed_recent)
                    row += 1

    @Lazy
    def persons(self):
        result = set()
        collator = ICollator(self.request.locale)
        factory = getUtility(IPersonFactory)
        sorting_key = lambda x: factory.getSortingKey(x, collator)
        for username in self.context:
            person = removeSecurityProxy(self.context[username])
            relationships = Membership.bind(member=person).all().relationships
            for link_info in relationships:
                section = removeSecurityProxy(link_info.target)
                if IGroup.providedBy(section):
                    continue
                if ITerm(section) in self.terms:
                    result.add(person)
        return sorted(result, key=sorting_key)


class TeacherAttendanceRateByDayReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(IDateRangeForm).select('start', 'end')

    report_builder = 'teacher_attendance_rate_by_day.xls'

    def updateTaskParams(self, task):
        start = self.form_params.get('start')
        end = self.form_params.get('end')
        if start > end:
            start, end = end, start
        if start is not None:
            day = '%d-%02d-%02d' % (start.year, start.month, start.day)
            task.request_params['start'] = day
        if end is not None:
            day = '%d-%02d-%02d' % (end.year, end.month, end.day)
            task.request_params['end'] = day


class TeacherAttendanceRateByDayView(TeacherAttendanceDisaggregatedByDayView,
                                     CalendarViewBase):

    message_title = _('teacher attendance rate by day report')
    sheet_title = _('Attendance Rate by Day')

    @property
    def base_filename(self):
        start = self.getDay('start')
        end = self.getDay('end')
        if start is None or end is None:
            filename = 'teacher_attendance_rate_by_day'
        else:
            filename = 'teacher_attendance_rate_by_day_%d_%02d_%02d_to_%d_%02d_%02d' % (
                start.year, start.month, start.day,
                end.year, end.month, end.day)
        return filename

    def getDay(self, name):
        day = self.request.get(name, None)
        if day is None:
            return datetime.date(datetime.now())
        try:
            year, month, day = [int(part) for part in day.split('-')]
            return datetime.date(year, month, day)
        except (TypeError, ValueError):
            pass
        try:
            return datetime.date(day)
        except (TypeError, ValueError):
            return None

    def print_headers(self, ws):
        headers = [
            'Term',
            'Week',
            'Date',
            'Attendance rate for all teachers (%)',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 0, index, header)

    def print_report_rows(self, ws):
        row = 1
        for data in self.report_rows():
            self.write(ws, row, 0, data.get('term'))
            self.write(ws, row, 1, data.get('week'))
            self.write(ws, row, 2, data.get('date'))
            self.write(ws, row, 3, data.get('teacher_present'))
            row += 1

    def getTerm(self, date):
        for term in self.terms:
            if date in term:
                return term

    def getWeekNumber(self, day, term):
        result = 0
        current_week = None
        for date in term:
            week = self.getWeek(date)
            if week != current_week:
                result += 1
                current_week = week
            if date == day:
                break
        return result

    def report_rows(self):
        result = []
        total = len(self.persons)
        today = getUtility(IDateManager).today
        for meeting in self.all_meetings:
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > today:
                continue
            data = {}
            term = self.getTerm(date)
            if term is not None:
                data['term'] = term.title
            data['week'] = self.getWeekNumber(date, term)
            data['date'] = date.isoformat()
            teachers = []
            for person in self.persons:
                score = self.getScores(person)[meeting]
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    teachers.append(1)
            if total:
                data['teacher_present'] = self.formatPercent(
                    sum(teachers), total)
            result.append(data)
        return result

    @Lazy
    def terms(self):
        selected = set()
        start = self.getDay('start')
        end = self.getDay('end')
        for term in self.all_terms.values():
            if start in term or end in term:
                selected.add(term)
        return sorted(selected, key=lambda term: term.first)

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        start = self.getDay('start')
        end = self.getDay('end')
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date) and (start <= date <= end):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings


class TeacherAbsenteeismReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(ICurrentTerm)

    report_builder = 'teacher_absenteeism_report.xls'

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)


class TeacherAbsenteeismReportView(ExcelExportView,
                                   FlourishSchoolAttendanceView,
                                   CalendarViewBase):

    message_title = _('teacher absenteeism report')

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    @property
    def base_filename(self):
        return 'teacher_absenteeism_report'

    @Lazy
    def terms(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear in ISchoolYearContainer(app).values():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return [term]

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def export_report(self, wb):
        self.task_progress.force('report', active=True)

        ws1 = wb.add_sheet(_('Overall'))
        self.print_overall_headers(ws1, _('Teachers attendance rate'))
        self.print_report_overall(ws1)

        ws2 = wb.add_sheet(_('3 absences in last 10 days'))
        self.print_headers(
            ws2, _('Teachers absent more than 3 days in the last 10 day period'))
        self.print_report_recent(ws2)

        self.finish('report')

    def print_overall_headers(self, ws, title):
        self.write(ws, 0, 0, title)
        self.write_header(ws, 2, 0, _('Term'))
        self.write(ws, 2, 1, self.terms[0].title)
        today = getUtility(IDateManager).today
        headers = [
            'Teacher',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 4, index, header)
        weeks = self.getWeeks(today, self.terms[0])
        col = 1
        for week_no, week in enumerate(weeks):
            label = translate(_('Week ${week}', mapping={'week': week_no + 1}), 
                              context=self.request)
            self.write_header(ws, 4, col, label)
            col += 1
        self.write_header(ws, 4, col, _('Total'))

    def getWeeks(self, day, term):
        result = []
        current_week = None
        for date in term:
            week = self.getWeek(date)
            if week != current_week:
                result.append(week)
                current_week = week
            if date == day:
                break
        return result

    def print_headers(self, ws, title):
        ws.write_merge(0, 0, 0, 3, title)
        self.write_header(ws, 2, 0, _('Term'))
        self.write(ws, 2, 1, self.terms[0].title)
        headers = [
            'Class',
            'Teacher',
            'Average attendance this term',
            'Number days missed in the last 10 days',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 4, index, header)

    def print_report_overall(self, ws):
        row = 5
        today = getUtility(IDateManager).today
        meetings = {}
        for meeting in self.allMeetings():
            d = meeting.dtstart.astimezone(self.tzinfo).date()
            if d <= today:
                meetings[d] = TeacherAMRequirement(meeting)
        weeks = self.getWeeks(today, self.terms[0])
        for person in self.persons:
            self.write(ws, row, 0, person.title)
            col = 1
            present_term = []
            for week in weeks:
                meeting_days_in_week = []
                present_week = []
                for day in week:
                    requirement = meetings.get(day.date)
                    if requirement is not None:
                        meeting_days_in_week.append(1)
                        score = IEvaluateRequirement(requirement).getEvaluation(
                            person, requirement, default=UNSCORED)
                        if (score is None or
                            score is UNSCORED or
                            score.value is UNSCORED):
                            pass
                        elif not score.scoreSystem.isAbsent(score):
                            present_week.append(1)
                present_term.append(sum(present_week))
                if meeting_days_in_week:
                    self.write(ws, row, col,
                               self.formatPercent(sum(present_week), sum(meeting_days_in_week)))
                col += 1
            if weeks:
                self.write(ws, row, col,
                           self.formatPercent(sum(present_term), len(weeks)))
            row += 1

    def getParents(self, person):
        result = set()
        for contact in IContactable(person).contacts.any(meaning=PARENT):
            result.add(contact)
        return sorted(result, key=lambda c: c.title)

    def print_person(self, ws, row, person, average, missed_recent):
        self.write(ws, row, 1, person.title)
        self.write(ws, row, 2, average)
        self.write(ws, row, 3, missed_recent)

    def print_report_recent(self, ws):
        row = 5
        today = getUtility(IDateManager).today
        meetings = [{'am': TeacherAMRequirement(meeting)}
                     for meeting in self.allMeetings()
                    if meeting.dtstart.astimezone(self.tzinfo).date() <= today]
        recent = meetings[-10:]
        total_meetings = len(meetings)
        for person in self.persons:
            person_all_day = []
            person_recent = []
            for info in meetings:
                am_requirement = info['am']
                present_am = False
                score = IEvaluateRequirement(am_requirement).getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_am = True
                if present_am:
                    person_all_day.append(1)
                    if info in recent:
                        person_recent.append(1)
            if total_meetings:
                average = (sum(person_all_day) * 100.0) / total_meetings
                missed_recent = len(recent) - sum(person_recent)
                if len(recent) >= 3 and missed_recent > 3:
                    self.print_person(ws, row, person, average, missed_recent)
                    row += 1

    def teacherGroups(self):
        result = []
        group_id = 'teachers'
        for year in self.schoolyears:
            groups = IGroupContainer(year)
            result.append(groups[group_id])
        return result

    @Lazy
    def persons(self):
        result = set()
        groups = self.teacherGroups()
        for group in groups:
            for teacher in group.members.all():
                result.add(removeSecurityProxy(teacher))
        return result

    def formatPercent(self, count, total):
        return '%.0f' % ((count*100)/float(total))


class IRateCollection(Interface):

    rate = Float(
        title=_('Payment threshold (%)'),
        min=0.0,
        max=100.0,
        default=75.0,
        required=True)


class FeeCollectionRateReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(ICurrentTerm, IRateCollection).select(
        'term', 'rate')

    report_builder = 'fee_collection_rate.xls'

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)
        task.request_params['rate'] = self.form_params.get('rate')


class FeeCollectionRateReportView(ExcelExportView):

    message_title = _('fee collection rate report')

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    @property
    def base_filename(self):
        return 'fee_collection_rate_report'

    @Lazy
    def terms(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear in ISchoolYearContainer(app).values():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return [term]

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def export_report(self, wb):
        self.task_progress.force('report', active=True)

        rate = self.request.get('rate')
        ws = wb.add_sheet(_('Overall'))
        self.print_headers(ws, translate(_('List students if their fee collection rate is less than ${rate}', mapping={'rate': rate}), context=self.request))
        self.print_report(ws)

        self.finish('report')

    def print_headers(self, ws, title):
        ws.write_merge(0, 0, 0, 6, title)
        self.write_header(ws, 2, 0, _('Term'))
        self.write(ws, 2, 1, self.terms[0].title)
        headers = [
            'Class',
            'Student',
            'Day/Boarding',
            'Average attendance this term',
            'Fee collection rate',
            'Parent name',
            'Parent contact number',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 4, index, header)

    def print_report(self, ws):
        row = 5
        today = getUtility(IDateManager).today
        meetings = [{'am': StudentAMRequirement(meeting),
                     'pm': StudentPMRequirement(meeting)}
                     for meeting in self.allMeetings()
                    if meeting.dtstart.astimezone(self.tzinfo).date() <= today]
        total_meetings = len(meetings)
        for person in self.persons:
            person_all_day = []
            for info in meetings:
                am_requirement = info['am']
                pm_requirement = info['pm']
                present_am = False
                present_pm = False
                score = IEvaluateRequirement(am_requirement).getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_am = True
                score = IEvaluateRequirement(pm_requirement).getEvaluation(
                    person, pm_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_pm = True
                if present_am and present_pm:
                    person_all_day.append(1)
            if total_meetings:
                average = (sum(person_all_day) * 100.0) / total_meetings
            else:
                average = 0
            printed = self.print_person(ws, row, person, average)
            if printed:
                row += 1

    def getParents(self, person):
        result = set()
        for contact in IContactable(person).contacts.any(meaning=PARENT):
            result.add(contact)
        return sorted(result, key=lambda c: c.title)

    def print_person(self, ws, row, person, average):
        term = self.terms[0]
        credits = [fee.amount for fee in IFeeContainer(person).values()
                   if term.first <= fee.date <= term.last and
                   ICredit.providedBy(fee)]
        debits = [fee.amount for fee in IFeeContainer(person).values() 
                  if term.first <= fee.date <= term.last and
                  IDebit.providedBy(fee)]
        fee_collection_rate = 0
        if debits:
            fee_collection_rate = (sum(credits) * 100)/ sum(debits)
        if fee_collection_rate > float(self.request.get('rate')):
            return False
        self.write(ws, row, 1, person.title)
        self.write(ws, row, 2, IDemographics(person).get('day_boarding') or '')
        self.write(ws, row, 3, average)
        self.write(ws, row, 4, fee_collection_rate)
        parents = self.getParents(person)
        phones = lambda x: ', '.join(filter(None, [
            x.home_phone,
            x.work_phone,
            x.mobile_phone]))
        if parents:
            self.write(ws, row, 5, ', '.join([c.title for c in parents]))
            self.write(ws, row, 6, ', '.join([phones(c) for c in parents]))
        return True

    @Lazy
    def persons(self):
        result = set()
        collator = ICollator(self.request.locale)
        factory = getUtility(IPersonFactory)
        sorting_key = lambda x: factory.getSortingKey(x, collator)
        for username in self.context:
            person = removeSecurityProxy(self.context[username])
            relationships = Membership.bind(member=person).all().relationships
            for link_info in relationships:
                section = removeSecurityProxy(link_info.target)
                if IGroup.providedBy(section):
                    continue
                if ITerm(section) in self.terms:
                    result.add(person)
        return sorted(result, key=sorting_key)


class FeeCollectionRateByWeekReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(ICurrentTerm)

    report_builder = 'fee_collection_rate_by_week.xls'

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)


class FeeCollectionRateByWeekReportView(ExcelExportView,
                                        CalendarViewBase):

    message_title = _('fee collection rate by week report')

    def addExporters(self, progress):
        progress.add('report', title=_('Report'), progress=0.0)

    @property
    def base_filename(self):
        return 'fee_collection_rate_by_week_report'

    @Lazy
    def terms(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear in ISchoolYearContainer(app).values():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return [term]

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def export_report(self, wb):
        self.task_progress.force('report', active=True)

        ws = wb.add_sheet(_('Fee Collection Rate By Week'))
        self.print_headers(ws)
        self.print_report(ws)

        self.finish('report')

    def print_headers(self, ws):
        self.write_header(ws, 0, 0, _('Term'))
        self.write(ws, 0, 1, self.terms[0].title)
        headers = [
            'Week',
            'Fee collection rate (%)',
        ]
        for index, header in enumerate(headers):
            self.write_header(ws, 2, index, header)

    def print_report(self, ws):
        row = 3
        today = getUtility(IDateManager).today
        weeks = self.getWeeks(today, self.terms[0])
        for week_no, week in enumerate(weeks):
            first = week[0].date
            last = week[-1].date
            if today < last:
                last = today
            week_credits = []
            week_debits = []
            for person in self.persons:
                credits = [fee.amount for fee in IFeeContainer(person).values()
                           if first <= fee.date <= last and
                           ICredit.providedBy(fee)]
                debits = [fee.amount for fee in IFeeContainer(person).values() 
                          if first <= fee.date <= last and
                          IDebit.providedBy(fee)]
                week_credits.extend(credits)
                week_debits.extend(debits)
            self.write(ws, row, 0,
                       translate(_('Week ${week} (w/b on ${date})',
                                   mapping={'week': week_no + 1,
                                            'date': first.isoformat()}),
                                 context=self.request))
            fee_collection_rate = None
            if week_debits:
                fee_collection_rate = (sum(week_credits) * 100)/ sum(week_debits)
            self.write(ws, row, 1, fee_collection_rate)
            row += 1

    @Lazy
    def persons(self):
        result = set()
        collator = ICollator(self.request.locale)
        factory = getUtility(IPersonFactory)
        sorting_key = lambda x: factory.getSortingKey(x, collator)
        for username in self.context:
            person = removeSecurityProxy(self.context[username])
            relationships = Membership.bind(member=person).all().relationships
            for link_info in relationships:
                section = removeSecurityProxy(link_info.target)
                if IGroup.providedBy(section):
                    continue
                if ITerm(section) in self.terms:
                    result.add(person)
        return sorted(result, key=sorting_key)

    def getWeeks(self, day, term):
        result = []
        current_week = None
        for date in term:
            week = self.getWeek(date)
            if week != current_week:
                result.append(week)
                current_week = week
            if date == day:
                break
        return result


class PEASUKReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(ICurrentTerm)
    report_builder = 'peas_uk_report.xls'

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)


class PEASUKReportView(DemographicsSnapshotXLSView, CalendarViewBase):

    message_title = _('PEAS UK report')
    level_total_title = _('Full School')

    @property
    def base_filename(self):
        return 'peas_uk_report'

    @Lazy
    def terms(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear in ISchoolYearContainer(app).values():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return [term]

    def getWeeks(self, day, term):
        result = []
        current_week = None
        for date in term:
            week = self.getWeek(date)
            if week != current_week:
                result.append(week)
                current_week = week
            if date == day:
                break
        return result

    @Lazy
    def week_groups(self):
        grouped_by = 4
        result = []
        today = getUtility(IDateManager).today
        weeks = self.getWeeks(today, self.terms[0])
        for i in range(0, len(weeks), grouped_by):
            group = weeks[i:i+grouped_by]
            weeks_count = len(group)
            meetings = []
            start = i + 1
            end = start + weeks_count - 1
            mapping = {'start': start, 'end': end}
            if weeks_count > 1:
                title = translate(_('Week ${start}-${end} %',
                                    mapping=mapping),
                                  context=self.request)
            else:
                title = translate(_('Week ${start} %',
                                    mapping=mapping),
                                    context=self.request)
            first = group[0][0].date
            last = group[-1][-1].date
            meetings = []
            for meeting in self.all_meetings:
                meeting_date = meeting.dtstart.astimezone(self.tzinfo).date()
                if first <= meeting_date <= last:
                    meetings.append(meeting)
            result.append({
                'group': group,
                'title': title,
                'meetings': meetings,
            })
        return result

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        return wb

    def export_report(self, wb):
        self.task_progress.force('report', active=True)

        ws = wb.add_sheet(_('For PEAS UK'))
        self.print_headers(ws)
        self.print_row_titles(ws)
        self.print_report(ws)

        self.finish('report')

    def print_headers(self, ws):
        self.write_header(ws, 0, 0, _('Term'))
        ws.write(0, 1, self.terms[0].title)
        self.print_levels(ws)
        self.print_column_titles(ws)

    def print_levels(self, ws):
        style = self.style_column_title
        merge = len(self.week_groups)
        row = 2
        col = 1
        for info in self.levels_data:
            ws.write_merge(row, row, col, col + merge, info['title'], style)
            col += (merge + 1)

    def print_column_titles(self, ws):
        style = self.style_column_title
        row = 3
        col = 0
        ws.write(row, col, _('Year Group'), style)
        col += 1
        for info in self.levels_data[:-1]:
            for week_group in self.week_groups:
                ws.write(row, col, week_group['title'], style)
                col += 1
            ws.write(row, col, _('Total %'), style)
            col += 1
        for week_group in self.week_groups:
            ws.write(row, col, week_group['title'], style)
            col += 1
        ws.write(row, col, _('Total %'), style)

    def print_row_titles(self, ws):
        style = self.style_row_title
        row = 4
        col = 0
        ws.write(row, col, _('Girls Boarder (GB)'), style)
        ws.write(row + 1, col, _('Girls Day (GD)'), style)
        ws.write(row + 2, col, _('Boys Boarder (BB)'), style)
        ws.write(row + 3, col, _('Boys Day (BD)'), style)
        ws.write(row + 4, col, _('Girls'), style)
        ws.write(row + 5, col, _('Boys'), style)
        ws.write(row + 6, col, _('Boarding'), style)
        ws.write(row + 7, col, _('Day'), style)

    def print_report(self, ws):
        style = self.style_cell
        row = 4
        col = 1
        self.print_row_girls_boarder(ws, row, col, style)
        self.print_row_girls_day(ws, row + 1, col, style)
        self.print_row_boys_boarder(ws, row + 2, col, style)
        self.print_row_boys_day(ws, row + 3, col, style)
        self.print_row_girls(ws, row + 4, col, style)
        self.print_row_boys(ws, row + 5, col, style)
        self.print_row_boarding(ws, row + 6, col, style)
        self.print_row_day(ws, row + 7, col, style)

    def allMeetings(self):
        terms = removeSecurityProxy(self.terms)
        if not terms:
            return ()
        meetings = []
        for term in terms:
            for date in term:
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day))
                    meetings.append(makeSchoolAttendanceMeeting(dt))
        return meetings

    @Lazy
    def all_meetings(self):
        return self.allMeetings()

    def get_attendance_for_week_group(self, persons, week_group):
        result = []
        total_meetings = len(week_group['meetings'])
        total = len(persons)
        for meeting in week_group['meetings']:
            am_requirement = StudentAMRequirement(meeting)
            pm_requirement = StudentPMRequirement(meeting)
            all_day = []
            for person in persons:
                present_am = False
                present_pm = False
                evaluations = IEvaluateRequirement(am_requirement)
                score = evaluations.getEvaluation(
                    person, am_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_am = True
                evaluations = IEvaluateRequirement(pm_requirement)
                score = evaluations.getEvaluation(
                    person, pm_requirement, default=UNSCORED)
                if (score is None or
                    score is UNSCORED or
                    score.value is UNSCORED):
                    pass
                elif not score.scoreSystem.isAbsent(score):
                    present_pm = True
                if present_am and present_pm:
                    all_day.append(1)
            result.append(sum(all_day))
        if total and total_meetings:
            return self.div(sum(result) * 100, total * total_meetings)
        return 0

    def print_row_girls_boarder(self, ws, row, col, style):
        get_boarding = lambda x: IDemographics(x).get('day_boarding') or ''
        is_boarding = lambda x: get_boarding(x).lower() == 'boarding'
        for info in self.levels_data:
            girls = [p for p in info['girls'] if is_boarding(p)]
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    girls, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_girls_day(self, ws, row, col, style):
        get_day = lambda x: IDemographics(x).get('day_boarding') or ''
        is_day = lambda x: get_day(x).lower() == 'day'
        for info in self.levels_data:
            girls = [p for p in info['girls'] if is_day(p)]
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    girls, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_boys_boarder(self, ws, row, col, style):
        get_boarding = lambda x: IDemographics(x).get('day_boarding') or ''
        is_boarding = lambda x: get_boarding(x).lower() == 'boarding'
        for info in self.levels_data:
            boys = [p for p in info['boys'] if is_boarding(p)]
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    boys, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_boys_day(self, ws, row, col, style):
        get_day = lambda x: IDemographics(x).get('day_boarding') or ''
        is_day = lambda x: get_day(x).lower() == 'day'
        for info in self.levels_data:
            boys = [p for p in info['boys'] if is_day(p)]
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    boys, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_girls(self, ws, row, col, style):
        for info in self.levels_data:
            girls = info['girls']
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    girls, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_boys(self, ws, row, col, style):
        for info in self.levels_data:
            boys = info['boys']
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    boys, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_boarding(self, ws, row, col, style):
        get_boarding = lambda x: IDemographics(x).get('day_boarding') or ''
        is_boarding = lambda x: get_boarding(x).lower() == 'boarding'
        for info in self.levels_data:
            girls = [p for p in info['girls'] if is_boarding(p)]
            boys = [p for p in info['boys'] if is_boarding(p)]
            girls_boys = girls + boys
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    girls_boys, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1

    def print_row_day(self, ws, row, col, style):
        get_boarding = lambda x: IDemographics(x).get('day_boarding') or ''
        is_boarding = lambda x: get_boarding(x).lower() == 'day'
        for info in self.levels_data:
            girls = [p for p in info['girls'] if is_boarding(p)]
            boys = [p for p in info['boys'] if is_boarding(p)]
            girls_boys = girls + boys
            all_attendance = []
            for week_group in self.week_groups:
                attendance = self.get_attendance_for_week_group(
                    girls_boys, week_group)
                all_attendance.append(attendance)
                ws.write(row, col, attendance, style)
                col += 1
            ws.write(row, col,
                     self.div(sum(all_attendance), len(all_attendance)),
                     style)
            col += 1


class OverallYearMonthReportRequestView(RequestXLSReportDialog,
                                        FlourishSchoolAttendanceView):

    report_builder = 'overall_year_month_report.xls'

    fields = field.Fields(IRequestYearlyReport)

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['years'] = self.years()

    def years(self):
        this_year = self.year
        terms = self.all_terms
        min_year = min([this_year] +
                       [term.first.year
                        for term in terms.values() if term.first])
        max_year = max([this_year] +
                       [term.last.year
                        for term in terms.values() if term.last])
        return range(min_year, max_year+1)

    def updateTaskParams(self, task):
        schoolyear = self.form_params.get('schoolyear')
        if schoolyear is not None:
            task.request_params['schoolyear'] = schoolyear.__name__


class OverallReportBase(ExcelExportView,
                        CalendarViewBase):

    overall_line_id = 'overall'
    _style_cache = {}

    def easyxf(self, s, num_format_str=None):
        key = s.strip(), num_format_str
        if key not in self._style_cache:
            self._style_cache[key] = xlwt.easyxf(s, num_format_str)
        return self._style_cache[key]

    @Lazy
    def schoolyear(self):
        app = ISchoolToolApplication(None)
        schoolyears = ISchoolYearContainer(app)
        return schoolyears.get(self.request.get('schoolyear'))

    def addExporters(self, progress):
        progress.add(
            'overall', title=_('Report'), progress=0.0)
        progress.add(
            'data', title=_('Gathering report data'),
            active=False, progress=0.0)

    def __call__(self):
        app = ISchoolToolApplication(None)
        self.tzinfo = pytz.timezone(IApplicationPreferences(app).timezone)
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.setup_book(wb)
        self.task_progress.force('data', active=True)
        self.data
        self.finish('data')

        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        self.task_progress.force('overall', progress=1.0)
        if 'DEBUG' in self.request:
            return self.renderize(wb)
        else:
            return wb

    def renderize(self, workbook):
        datafile = StringIO()
        workbook.save(datafile)
        data = datafile.getvalue()
        self.setUpHeaders(data)
        return data

    @Lazy
    def custom_colors(self):
        result = {
            'above95': (60, 146, 208, 80),
            'above90': (61, 255, 255, 0),
            'above80': (62, 255, 153, 51),
            'below80': (63, 255, 0, 0),
            'g_above90': (45, 55, 96, 146),
            'g_above80': (46, 77, 113, 158),
            'g_above70': (47, 99, 131, 170),
            'g_above60': (48, 121, 149, 182),
            'g_above50': (49, 143, 166, 194),
            'g_above40': (50, 166, 184, 206),
            'g_above30': (51, 188, 202, 218),
            'g_above20': (52, 210, 219, 230),
            'g_above10': (53, 232, 237, 242),
            'g_below10': (54, 255, 255, 255),
            'report_title_bg': (55, 31, 73, 125),
            'table_title_bg': (56, 217, 217, 217),
        }
        return result

    def setup_book(self, wb):
        for color, info in self.custom_colors.items():
            code, r, g, b = info
            xlwt.add_palette_colour(color, code)
            wb.set_colour_RGB(code, r, g, b)

    def get_background_color(self, value):
        if value >= 95:
            return 'above95'
        elif value >= 90:
            return 'above90'
        elif value >= 80:
            return 'above80'
        else:
            return 'below80'

    def get_gradient_color(self, value):
        if value >= 90:
            return 'g_above90'
        elif value >= 80:
            return 'g_above80'
        elif value >= 70:
            return 'g_above70'
        elif value >= 60:
            return 'g_above60'
        elif value >= 50:
            return 'g_above50'
        elif value >= 40:
            return 'g_above40'
        elif value >= 30:
            return 'g_above30'
        elif value >= 20:
            return 'g_above20'
        elif value >= 10:
            return 'g_above10'
        else:
            return 'g_below10'

    def get_reason_background_color(self, value):
        if 0 <= value <= 5:
            return 'above95'
        elif 5 < value <= 10:
            return 'above90'
        elif 10 < value <= 20:
            return 'above80'
        else:
            return 'below80'

    @Lazy
    def styles(self):
        result = {}
        pattern = 'pattern: pattern solid, fore_colour %s'
        for color in self.custom_colors:
            result[color] = self.easyxf(pattern % color)
        result.update({
            'block_top_row': self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right thin, bottom no_line, left no_line; '
            ),
            'block_bottom_row': self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right thin, bottom thin, left no_line; '
            ),
            'table_title': self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color table_title_bg; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ),
            'column_header': self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
            ),
        })
        return result

    @Lazy
    def levels(self):
        app = ISchoolToolApplication(None)
        return ILevelContainer(app)

    @Lazy
    def meetings(self, ignore_future=True):
        result = []
        today = getUtility(IDateManager).today        
        for term in sorted(self.schoolyear.values(), key=lambda t: t.first):
            for date in term:
                if ignore_future and date > today:
                    continue
                if term.isSchoolday(date):
                    dt = self.tzinfo.localize(
                        datetime.datetime(date.year, date.month, date.day)
                    )
                    result.append(makeSchoolAttendanceMeeting(dt))
        return result

    def meetings_by_term(self, term):
        result = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date in term:
                result[meeting] = attendance_data
        return result

    @Lazy
    def meetings_by_month(self):
        result = {}
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            for term in terms:
                if date in term:
                    key = term, date.year, date.month
                    if key not in result:
                        result[key] = {}
                    result[key][meeting] = attendance_data
        return result

    def is_learner(self, person, date):
        person = removeSecurityProxy(person)
        r = Membership.bind(member=person).on(date).any(ACTIVE).relationships
        for link_info in r:
            target = removeSecurityProxy(link_info.target)
            if (ISection.providedBy(target) and
                ISchoolYear(target) == self.schoolyear):
                return True
        return False

    def level(self, person, date):
        result = []
        person = removeSecurityProxy(person)
        for level in person.levels.on(date).any(ACTIVE):
            result.append(level)
        if result:
            return result[0]

    def isSkipped(self, score):
        if (score in (UNSCORED, None)
            or score.value is UNSCORED):
            return True
        ss = score.scoreSystem
        return ss.isSkipped(score)

    def isPresent(self, score):
        if (score in (UNSCORED, None)
            or score.value is UNSCORED):
            return False
        ss = score.scoreSystem
        not_present = ss.tag_absent + ss.tag_excused + ss.tag_skip
        return score.value not in not_present

    def reason(self, score):
        if (score in (UNSCORED, None)
            or score.value is UNSCORED):
            return
        return self.SCORE_TO_REASON.get(score.value)

    @Lazy
    def SCORE_TO_REASON(self):
        result = OrderedDict([
            ('As', 'sick'),
            ('Af', 'fees'),
            ('Ao', 'other'),
        ])
        return result

    @Lazy
    def data(self):
        result = {}
        app = ISchoolToolApplication(None)
        persons = app['persons']
        students = IGroupContainer(self.schoolyear).get('students')
        meetings_count = len(self.meetings)
        persons_count = len(app['persons'])
        for meeting_no, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            am_requirement = StudentAMRequirement(meeting)
            am_evaluations = IEvaluateRequirement(am_requirement)
            pm_requirement = StudentPMRequirement(meeting)
            pm_evaluations = IEvaluateRequirement(pm_requirement)
            students_on_date = list(students.members.on(date))
            am_attendance = {}
            pm_attendance = {}
            for person_no, person in enumerate(persons.values()):
                self.progress(
                    'data', normalized_progress(
                        meeting_no, meetings_count, person_no, persons_count)
                )
                level = self.level(person, date)
                if (level is not None and
                    (person in students_on_date or
                     self.is_learner(person, date))):
                    am_score = am_evaluations.getEvaluation(
                        person, am_requirement, default=UNSCORED)
                    if not self.isSkipped(am_score):
                        if person not in am_attendance:
                            am_attendance[person] = {}
                        is_present = self.isPresent(am_score)
                        am_attendance[person]['present'] = is_present
                        am_attendance[person]['level'] = level
                        am_attendance[person]['reason'] = self.reason(am_score)
                        am_attendance[person]['value'] = am_score.value
                        am_attendance[person]['score'] = am_score
                    pm_score = pm_evaluations.getEvaluation(
                        person, pm_requirement, default=UNSCORED)
                    if not self.isSkipped(pm_score):
                        if person not in pm_attendance:
                            pm_attendance[person] = {}
                        is_present = self.isPresent(pm_score)
                        pm_attendance[person]['present'] = is_present
                        pm_attendance[person]['level'] = level
                        pm_attendance[person]['reason'] = self.reason(pm_score)
                        pm_attendance[person]['value'] = pm_score.value
                        pm_attendance[person]['score'] = pm_score
            result[meeting] = {
                am_requirement: am_attendance,
                pm_requirement: pm_attendance,
            }
        return result

    def div(self, count, total):
        if not total:
            return 0.0
        return count/float(total)

    def print_legend_1(self, ws, row=None, col=None):
        labels = [
            (_('Above 95% present'), 95),
            (_('90% - 94% present'), 90),
            (_('80% - 89% present'), 80),
            (_('Less than 80% present'), 0),
        ]
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write(row, col, _('Legend'), style)
        row += 1
        borders = self.translate_borders([1, 1, 1, 1])
        for i, label_info in enumerate(labels):
            label, value = label_info
            background = self.get_background_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row + i, col, label, style)

    def print_legend_2(self, ws, row=None, col=None):
        labels = [
            (_('Dark blue = highest %'), 95),
            (_('Light blue = lowest %'), 10),
        ]
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write(row, col, _('Legend'), style)
        row += 1
        borders = self.translate_borders([1, 1, 1, 1])
        for i, label_info in enumerate(labels):
            label, value = label_info
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row + i, col, label, style)

    def print_headers(self, ws, title):
        row = 0
        merge = len(self.levels) + 1
        self.print_report_title(ws, title, row=row, merge=merge)

        row = 2
        today = getUtility(IDateManager).today
        self.print_report_date(ws, today, row=row)

        row += 1
        self.print_report_filters(ws, row=row)

    def print_report_title(self, ws, title, row, merge):
        style = self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color report_title_bg; '
        )
        ws.write_merge(row, row, 0, merge, title, style)

    def print_report_date(self, ws, today, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Date report produced'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, today.strftime('%d/%m/%Y'), style)

    def translate_borders(self, borders):
        result = []
        order = ['top', 'right', 'bottom', 'left']
        for i, enabled in  enumerate(borders):
            result.append('%s %s' % (order[i], ['no_line', 'thin'][enabled]))
        return ', '.join(result)        

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    @Lazy
    def weeks(self):
        result = []
        today = getUtility(IDateManager).today        
        for term in self.terms:
            current_week = None
            for date in term:
                if date > today:
                    continue
                if term.isSchoolday(date):
                    week = self.getWeek(date)
                    if week != current_week:
                        if week not in result:
                            result.append(week)
                        current_week = week
        return result

    @Lazy
    def days(self):
        result = []
        today = getUtility(IDateManager).today        
        for term in self.terms:
            for date in term:
                if date > today:
                    continue
                if term.isSchoolday(date):
                    result.append(date)
        return result


class OverallYearMonthReportView(OverallReportBase):

    message_title = _('Overall report: by year, term or month')

    sheet_1_title = _('Student attendance - overall reports - by year '
                      'and by term')
    sheet_1_label = _('Overall - year')

    sheet_2_title = _('Student attendance - overall reports - by month')
    sheet_2_label = _('Overall - montly')

    def addExporters(self, progress):
        super(OverallYearMonthReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Overall - year sheet'),
            active=False, progress=0.0)
        progress.add(
            'sheet2', title=_('Genearting Overall - monthly sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'overall_year_month_%s' % self.schoolyear.title

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_year_percentage_table(ws)
        self.print_year_rate_table(ws)
        self.task_progress.force('sheet1', progress=0.5)
        self.print_term_percentage_table(ws)
        self.print_term_rate_table(ws)
        self.print_legend_1(ws, row=5, col=len(self.levels) + 3)
        self.print_legend_2(ws, row=17, col=len(self.levels) + 3)
        self.finish('sheet1')

        self.task_progress.force('sheet2', active=True)
        ws = wb.add_sheet(self.sheet_2_label)
        self.print_headers(ws, self.sheet_2_title)
        self.print_month_percentage_table(ws)
        self.task_progress.force('sheet2', progress=0.5)
        self.print_month_rate_table(ws)
        col = (len(self.months) * (len(self.levels) + 1)) + 2
        self.print_legend_1(ws, row=5, col=col)
        self.print_legend_2(ws, row=19, col=col)
        self.finish('sheet2')

    @Lazy
    def months(self):
        result = set()
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)
        for meeting in self.data.keys():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            term_of_date = None
            for term in terms:
                if date in term:
                    term_of_date = term
            key = (term_of_date, date.year, date.month,
                   month_names[date.month][:3])
            result.add(key)
        return sorted(result, key=lambda k: (k[0].first, k[1], k[2]))

    def print_year_percentage_cell(self, ws, row, col, filter_func, borders):
        value = 0
        background = 'white'
        count = 0
        present = 0
        borders = self.translate_borders(borders)
        for meeting, attendance_data in self.data.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        count += 1
                        if person_data['present']:
                            present += 1
        if count:
            value = (self.div(present, count) * 100)
            if value:
                background = self.get_background_color(value)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

    def print_year_rate_cell(self, ws, row, col, limit, filter_func, borders):
        borders = self.translate_borders(borders)
        background = 'white'
        count = 0
        present = 0
        persons = {}
        for meeting, attendance_data in self.data.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        if person not in persons:
                            persons[person] = {'count': 0, 'present': 0}
                        persons[person]['count'] += 1
                        if person_data['present']:
                            persons[person]['present'] += 1
        for data in persons.values():
            value = self.div(data['present'], data['count']) * 100
            if limit(value):
                count += 1
                present += data['present']
        value = (self.div(count, len(persons)) * 100)
        if value:
            background = self.get_gradient_color(value)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

    def print_term_percentage_cell(self, ws, row, col, filter_func, borders,
                                   terms):
        by_term = dict([(term, {'count': 0, 'present': 0})
                        for term in terms])
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            term_data = None
            for term in terms:
                if date in term:
                    term_data = by_term[term]
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        term_data['count'] += 1
                        if person_data['present']:
                            term_data['present'] += 1

        borders = self.translate_borders(borders)
        year_to_date = {'count': 0, 'present': 0}
        for term in terms:
            value = 0
            background = 'white'
            term_data = by_term[term]
            if term_data['count']:
                year_to_date['count'] += term_data['count']
                year_to_date['present'] += term_data['present']
                value = self.div(term_data['present'], term_data['count']) * 100
                if value:
                    background = self.get_background_color(value)
            style = self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1

        value = 0
        background = 'white'
        if year_to_date['count']:
            value = self.div(year_to_date['present'],
                             year_to_date['count']) * 100
            if value:
                background = self.get_background_color(value)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)
        col += 1
        return col

    def print_term_rate_cell(self, ws, row, col, limit, filter_func, borders,
                             terms):
        borders = self.translate_borders(borders)
        by_term = dict([(term, {'count': 0, 'present': 0, 'persons': {}})
                        for term in terms])
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            term_data = None
            for term in terms:
                if date in term:
                    term_data = by_term[term]
            persons = term_data['persons']
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        if person not in persons:
                            persons[person] = {'count': 0, 'present': 0}
                        persons[person]['count'] += 1
                        if person_data['present']:
                            persons[person]['present'] += 1

        year_to_date = 0
        for term in terms:
            background = 'white'
            term_data = by_term[term]
            persons = term_data['persons']
            for data in persons.values():
                value = self.div(data['present'], data['count']) * 100
                if limit(value):
                    term_data['count'] += 1
                    term_data['present'] += data['present']

            value = self.div(
                term_data['count'], len(term_data['persons'])) * 100
            if value:
                background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
            year_to_date += value

        value = self.div(year_to_date, len(terms))
        background = 'white'
        if value:
            background = self.get_gradient_color(value)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)
        col += 1
        return col

    def print_month_percentage_cell(self, ws, row, col, filter_func, borders,
                                    terms):
        borders = self.translate_borders(borders)
        for term in terms:
            term_months = [k for k in self.months if k[0] is term]
            for term, year, month, month_name in term_months:
                value = 0
                background = 'white'
                count = 0
                present = 0
                meetings = self.meetings_by_month.get((term, year, month))
                for meeting, attendance_data in meetings.items():
                    requirements = [StudentAMRequirement(meeting),
                                    StudentPMRequirement(meeting)]
                    for requirement in requirements:
                        attendance = attendance_data[requirement]
                        for person, person_data in attendance.items():
                            if filter_func(person, person_data):
                                count += 1
                                if person_data['present']:
                                    present += 1
                if count:
                    value = self.div(present, count) * 100
                    if value:
                        background = self.get_background_color(value)
                style = self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
        return col

    def print_month_rate_cell(self, ws, row, col, limit, filter_func, borders,
                              terms):
        borders = self.translate_borders(borders)
        for term in terms:
            term_months = [k for k in self.months if k[0] is term]
            for term, year, month, month_name in term_months:
                value = 0
                background = 'white'
                count = 0
                present = 0
                persons = {}
                meetings = self.meetings_by_month.get((term, year, month))
                for meeting, attendance_data in meetings.items():
                    requirements = [StudentAMRequirement(meeting),
                                    StudentPMRequirement(meeting)]
                    for requirement in requirements:
                        attendance = attendance_data[requirement]
                        for person, person_data in attendance.items():
                            if filter_func(person, person_data):
                                if person not in persons:
                                    persons[person] = {'count': 0, 'present': 0}
                                persons[person]['count'] += 1
                                if person_data['present']:
                                    persons[person]['present'] += 1
                for data in persons.values():
                    value = self.div(data['present'], data['count']) * 100
                    if limit(value):
                        count += 1
                        present += data['present']
                value = (self.div(count, len(persons)) * 100)
                if value:
                    background = self.get_gradient_color(value)
                style = self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
        return col

    def print_year_percentage_table(self, ws):
        # table title
        title = _('Year to date - percentage of students present')
        row = 5
        merge = len(self.levels) + 1
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        # section row
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Section'), style)
        ws.write(row, 1, _('Full school'), self.styles['column_header'])
        for i, level in enumerate(self.levels.values()):
            ws.write(row, 2 + i, level.title, self.styles['column_header'])
        
        # all row
        row += 1
        ws.write(row, 0, _('All'), self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        ))
        borders = [1, 1, 1, 1]
        self.print_year_percentage_cell(
            ws, row, 1, lambda person, person_data: True, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: person_data['level'] is level,
                borders
            )

        # girls
        row += 1
        ws.write(row, 0, _('Girls'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(ws, row, 1, is_female, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_female(person, person_data)),
                borders
            )

        # boys
        row += 1
        ws.write(row, 0, _('Boys'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_male, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_male(person, person_data)),
                borders
            )

        # boarding
        row += 1
        ws.write(row, 0, _('Boarding'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(ws, row, 1, is_boarding, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_boarding(person, person_data)),
                borders
            )

        # day
        row += 1
        ws.write(row, 0, _('Day'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_day, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_day(person, person_data)),
                borders
            )

        # girls boarding
        row += 1
        ws.write(row, 0, _('Girls Boarder (GB)'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(ws, row, 1, is_female_boarding, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_boarding(person, person_data)),
                borders
            )

        # girls day
        row += 1
        ws.write(row, 0, _('Girls Day (GD)'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_female_day, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_day(person, person_data)),
                borders
            )

        # boys boarding
        row += 1
        ws.write(row, 0, _('Boys Boarder (BB)'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(ws, row, 1, is_male_boarding, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_boarding(person, person_data)),
                borders
            )

        # boys day
        row += 1
        ws.write(row, 0, _('Boys Day (BD)'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_male_day, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_day(person, person_data)),
                borders
            )

    def print_year_rate_table(self, ws):
        # table title
        title = _('Year to date - % students in different '
                  'attendance rate categories')
        row = 17
        merge = len(self.levels) + 1
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        # table headers
        row += 1
        ws.write(row, 1, _('Full school'), self.styles['column_header'])
        for i, level in enumerate(self.levels.values()):
            ws.write(row, 2 + i, level.title, self.styles['column_header'])

        # styles
        rate_row_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        borders = [0, 1, 0, 1]

        # above 95%
        row += 1
        ws.write(row, 0, _('Students above 95%'), rate_row_header)
        limit = lambda value: value >= 95
        self.print_year_rate_cell(
            ws, row, 1, limit, lambda person, person_data: True, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_rate_cell(
                ws, row, 2 + i, limit,
                lambda person, person_data: person_data['level'] is level,
                borders
            )

        # less than 95%
        row += 1
        ws.write(row, 0, _('Students less than 95%'), rate_row_header)
        limit = lambda value: 90 <= value < 95
        self.print_year_rate_cell(
            ws, row, 1, limit, lambda person, person_data: True, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_rate_cell(
                ws, row, 2 + i, limit,
                lambda person, person_data: person_data['level'] is level,
                borders
            )

        # less than 90%
        row += 1
        ws.write(row, 0, _('Students less than 90%'), rate_row_header)
        limit = lambda value: 80 <= value < 90
        self.print_year_rate_cell(
            ws, row, 1, limit, lambda person, person_data: True, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_rate_cell(
                ws, row, 2 + i, limit,
                lambda person, person_data: person_data['level'] is level,
                borders
            )

        # less than 80%
        row += 1
        ws.write(row, 0, _('Students less than 80%'),
                 self.styles['block_bottom_row'])
        limit = lambda value: value < 80
        borders = [0, 1, 1, 1]
        self.print_year_rate_cell(
            ws, row, 1, limit, lambda person, person_data: True, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_rate_cell(
                ws, row, 2 + i, limit,
                lambda person, person_data: person_data['level'] is level,
                borders
            )

    def print_term_percentage_table(self, ws):
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)

        # table title
        title = _('By term - % present')
        row = 24
        merge = (len(terms) + 1 ) * (len(self.levels) + 1)
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        # table header
        row += 1
        col =  1
        merge = len(terms)
        ws.write_merge(row, row, col, col + merge, _('Full school'),
                       self.styles['column_header'])
        col += merge + 1
        for i, level in enumerate(self.levels.values()):
            ws.write_merge(row, row, col, col + merge, level.title,
                           self.styles['column_header'])
            col +=  merge + 1

        # section row
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Section'), style)
        
        col = 1
        term_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
            'alignment: horizontal center; '
        )
        year_to_date_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left no_line; '
            'alignment: horizontal center; '
        )

        for term in terms:
            ws.write(row, col, term.title, term_header)
            col += 1
        ws.write(row, col, _('Year to date'), year_to_date_header)
        col += 1

        for i, level in enumerate(self.levels.values()):
            for term in terms:
                ws.write(row, col, term.title, term_header)
                col += 1
            ws.write(row, col, _('Year to date'), year_to_date_header)
            col += 1

        # all row
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left no_line; '
        )
        ws.write(row, 0, _('All'), style)

        borders = [1,1,1,1]
        col = self.print_term_percentage_cell(ws, row, 1,
                          lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # girls
        row += 1
        ws.write(row, 0, _('Girls'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_female, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_female(person, person_data)),
                borders, terms)

        # boys
        row += 1
        ws.write(row, 0, _('Boys'), self.styles['block_bottom_row'])
        borders = [0,1,1,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_male, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_male(person, person_data)),
                borders, terms)

        # boarding
        row += 1
        ws.write(row, 0, _('Boarding'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_boarding, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_boarding(person, person_data)),
                borders, terms)

        # day
        row += 1
        ws.write(row, 0, _('Day'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_day, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_day(person, person_data)),
                borders, terms)

        # girls boarding
        row += 1
        ws.write(row, 0, _('Girls Boarder (GB)'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_female_boarding, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_boarding(person, person_data)),
                borders, terms)

        # girls day
        row += 1
        ws.write(row, 0, _('Girls Day (GD)'), self.styles['block_bottom_row'])
        borders = [0,1,1,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_female_day, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_day(person, person_data)),
                borders, terms)

        # boys boarding
        row += 1
        ws.write(row, 0, _('Boys Boarder (BB)'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_male_boarding, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_boarding(person, person_data)),
                borders, terms)

        # boys day
        row += 1
        ws.write(row, 0, _('Boys Boarder (BB)'),
                 self.styles['block_bottom_row'])
        borders = [0,1,1,1]
        col = self.print_term_percentage_cell(
            ws, row, 1, is_male_day, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_day(person, person_data)),
                borders, terms)

    def print_term_rate_table(self, ws):
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)

        # table title
        title = _('Term to date - % students in different '
                  'attendance rate categories')
        row = 37
        merge = (len(terms) + 1) * (len(self.levels) + 1)
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        # table headers
        row += 1
        col =  1
        merge = len(terms)
        ws.write_merge(row, row, col, col + merge, _('Full school'),
                       self.styles['column_header'])

        col +=  merge + 1
        for i, level in enumerate(self.levels.values()):
            ws.write_merge(row, row, col, col + merge, level.title,
                           self.styles['column_header'])
            col +=  merge + 1

        # section row
        row += 1
        title = _('% student is different attendance boundaries')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, title, style)

        col = 1
        term_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
            'alignment: horizontal center; '
        )
        year_to_date_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left no_line; '
            'alignment: horizontal center; '
        )

        for term in terms:
            ws.write(row, col, term.title, term_header)
            col += 1
        ws.write(row, col, _('Year to date'), year_to_date_header)
        col += 1

        for i, level in enumerate(self.levels.values()):
            for term in terms:
                ws.write(row, col, term.title, term_header)
                col += 1
            ws.write(row, col, _('Year to date'), year_to_date_header)
            col += 1

        # styles
        rate_row_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        borders = [0, 1, 0, 1]

        # above 95%
        row += 1
        ws.write(row, 0, _('Students above 95%'), rate_row_header)
        limit = lambda value: value >= 95
        col = self.print_term_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # less than 95%
        row += 1
        ws.write(row, 0, _('Students less than 95%'), rate_row_header)
        limit = lambda value: 90 <= value < 95
        col = self.print_term_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # less than 90%
        row += 1
        ws.write(row, 0, _('Students less than 90%'), rate_row_header)
        limit = lambda value: 80 <= value < 90
        col = self.print_term_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # less than 80%
        row += 1
        ws.write(row, 0, _('Students less than 80%'),
                 self.styles['block_bottom_row'])
        limit = lambda value: value < 80
        borders = [0, 1, 1, 1]
        col = self.print_term_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_term_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

    def print_month_percentage_table(self, ws):
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)

        # table title
        title = _('By month - percentage present')
        row = 5
        merge = len(self.months) * (len(self.levels) + 1)
        ws.write_merge(row, row, 0,  merge, title, self.styles['table_title'])

        # table header
        row += 1
        col = 1
        merge = len(self.months) - 1
        ws.write_merge(row, row, col, col + merge, _('Full school'),
                       self.styles['column_header'])

        col += merge + 1
        for i, level in enumerate(self.levels.values()):
            ws.write_merge(row, row, col, col + merge, level.title,
                           self.styles['column_header'])
            col += merge + 1

        # table header 2
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, '', style)
        col = 1
        terms = [term for term in terms
                 if [k for k in self.months if k[0] is term]]
        for term in terms:
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right no_line, bottom thin, left no_line; '
                'alignment: horizontal center; '
            )
            term_months = [k for k in self.months if k[0] is term]
            merge = len(term_months) -  1
            if term is terms[0]:
                style.borders.left = xlwt.Borders.THIN
            if term is terms[-1]:
                style.borders.right = xlwt.Borders.THIN
            if merge > 0:
                ws.write_merge(row, row, col, col + merge, term.title, style)
            else:
                ws.write(row, col, term.title, style)
            col += merge + 1

        for i, level in enumerate(self.levels.values()):
            for term in terms:
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top thin, right no_line, bottom thin, left no_line; '
                    'alignment: horizontal center; '
                )
                term_months = [k for k in self.months if k[0] is term]
                merge = len(term_months) - 1
                if term is terms[-1]:
                    style.borders.right = xlwt.Borders.THIN
                if merge > 0:
                    ws.write_merge(row, row, col, col + merge, term.title, style)
                else:
                    ws.write(row, col, term.title, style)
                col += merge + 1

        # section row
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Section'), style)

        col = 1
        for term in terms:
            term_months = [k for k in self.months if k[0] is term]
            for parts in term_months:
                month_header = self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top thin, right no_line, bottom thin, left no_line; '
                    'alignment: horizontal center; '
                )
                term, year, month, month_name = parts
                if parts is term_months[-1]:
                    month_header.borders.right = xlwt.Borders.THIN
                ws.write(row, col, month_name, month_header)
                col += 1

        for i, level in enumerate(self.levels.values()):
            for term in terms:
                term_months = [k for k in self.months if k[0] is term]
                for parts in term_months:
                    month_header = self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'borders: top thin, right no_line, bottom thin, left no_line; '
                        'alignment: horizontal center; '
                    )
                    term, year, month, month_name = parts
                    if parts is term_months[-1]:
                        month_header.borders.right = xlwt.Borders.THIN
                    ws.write(row, col, month_name, month_header)
                    col += 1

        # all row
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left no_line; '
        )
        ws.write(row, 0, _('All'), style)

        borders = [1,1,1,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # girls
        row += 1
        ws.write(row, 0, _('Girls'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_female, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_female(person, person_data)),
                borders, terms)

        # boys
        row += 1
        ws.write(row, 0, _('Boys'), self.styles['block_bottom_row'])
        borders = [0,1,1,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_male, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_male(person, person_data)),
                borders, terms)

        # boarding
        row += 1
        ws.write(row, 0, _('Boarding'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_boarding, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_boarding(person, person_data)),
                borders, terms)

        # day
        row += 1
        ws.write(row, 0, _('Day'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_day, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_day(person, person_data)),
                borders, terms)

        # girls boarding
        row += 1
        ws.write(row, 0, _('Girls Boarder (GB)'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_female_boarding, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_boarding(person, person_data)),
                borders, terms)

        # girls day
        row += 1
        ws.write(row, 0, _('Girls Day (GD)'), self.styles['block_bottom_row'])
        borders = [0,1,1,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_female_day, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_day(person, person_data)),
                borders, terms)

        # boys boarding
        row += 1
        ws.write(row, 0, _('Boys Boarder (BB)'), self.styles['block_top_row'])
        borders = [1,1,0,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_male_boarding, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_boarding(person, person_data)),
                borders, terms)

        # boys day
        row += 1
        ws.write(row, 0, _('Boys Boarder (BB)'),
                 self.styles['block_bottom_row'])
        borders = [0,1,1,1]
        col = self.print_month_percentage_cell(
            ws, row, 1, is_male_day, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_percentage_cell(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_day(person, person_data)),
                borders, terms)

    def print_month_rate_table(self, ws):
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)

        # table title
        title = _('By month - % students in different attendance '
                  'rate categories')
        row = 19
        merge = len(self.months) * (len(self.levels) + 1)
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        # table header
        row += 1
        col = 1
        merge = len(self.months) - 1
        ws.write_merge(row, row, col, col + merge, _('Full school'),
                       self.styles['column_header'])

        col += merge + 1
        for i, level in enumerate(self.levels.values()):
            ws.write_merge(row, row, col, col + merge, level.title,
                           self.styles['column_header'])
            col += merge + 1

        # table header 2
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, '', style)
        col = 1
        terms = [term for term in terms
                 if [k for k in self.months if k[0] is term]]
        for term in terms:
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right no_line, bottom thin, left no_line; '
                'alignment: horizontal center; '
            )
            term_months = [k for k in self.months if k[0] is term]
            merge = len(term_months) -  1
            if term is terms[0]:
                style.borders.left = xlwt.Borders.THIN
            if term is terms[-1]:
                style.borders.right = xlwt.Borders.THIN
            if merge > 0:
                ws.write_merge(row, row, col, col + merge, term.title, style)
            else:
                ws.write(row, col, term.title, style)
            col += merge + 1

        for i, level in enumerate(self.levels.values()):
            for term in terms:
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top thin, right no_line, bottom thin, left no_line; '
                    'alignment: horizontal center; '
                )
                term_months = [k for k in self.months if k[0] is term]
                merge = len(term_months) - 1
                if term is terms[-1]:
                    style.borders.right = xlwt.Borders.THIN
                if merge > 0:
                    ws.write_merge(row, row, col, col + merge, term.title, style)
                else:
                    ws.write(row, col, term.title, style)
                col += merge + 1

        # section row
        row += 1
        title = _('% student is different attendance boundaries')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, title, style)

        col = 1
        for term in terms:
            term_months = [k for k in self.months if k[0] is term]
            for parts in term_months:
                month_header = self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top thin, right no_line, bottom thin, left no_line; '
                    'alignment: horizontal center; '
                )
                term, year, month, month_name = parts
                if parts is term_months[-1]:
                    month_header.borders.right = xlwt.Borders.THIN
                ws.write(row, col, month_name, month_header)
                col += 1

        for i, level in enumerate(self.levels.values()):
            for term in terms:
                term_months = [k for k in self.months if k[0] is term]
                for parts in term_months:
                    month_header = self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'borders: top thin, right no_line, bottom thin, left no_line; '
                        'alignment: horizontal center; '
                    )
                    term, year, month, month_name = parts
                    if parts is term_months[-1]:
                        month_header.borders.right = xlwt.Borders.THIN
                    ws.write(row, col, month_name, month_header)
                    col += 1

        # styles
        rate_row_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        borders = [0, 1, 0, 1]

        # above 95%
        row += 1
        ws.write(row, 0, _('Students above 95%'), rate_row_header)
        limit = lambda value: value >= 95
        col = self.print_month_rate_cell(
            ws, row, 1, limit, lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms
            )

        # less than 95%
        row += 1
        ws.write(row, 0, _('Students less than 95%'), rate_row_header)
        limit = lambda value: 90 <= value < 95
        col = self.print_month_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # less than 90%
        row += 1
        ws.write(row, 0, _('Students less than 90%'), rate_row_header)
        limit = lambda value: 80 <= value < 90
        col = self.print_month_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)

        # less than 80%
        row += 1
        ws.write(row, 0, _('Students less than 80%'),
                 self.styles['block_bottom_row'])
        limit = lambda value: value < 80
        borders = [0, 1, 1, 1]
        col = self.print_month_rate_cell(
            ws, row, 1, limit,
            lambda person, person_data: True, borders, terms)
        for i, level in enumerate(self.levels.values()):
            col = self.print_month_rate_cell(
                ws, row, col, limit,
                lambda person, person_data: person_data['level'] is level,
                borders, terms)


class IAllTerms(Interface):

    term = Choice(
        title=_('Term'),
        source='schooltool.peas.all_terms',
        required=True)


def schoolyear_id_getter(schoolyear):
    return '%s.all_terms' % schoolyear.__name__


class AllTermsVocabulary(SimpleVocabulary):

    implements(IContextSourceBinder)

    def __init__(self, context):
        self.context = context
        terms = self.createTerms()
        SimpleVocabulary.__init__(self, terms)

    def createTerms(self):
        result = []
        result.append(self.createTerm(
            None,
            widget.SequenceWidget.noValueToken,
            _('Select a term'),
        ))
        app = ISchoolToolApplication(None)
        container = ISchoolYearContainer(app)
        for schoolyear_id, schoolyear in container.items():

            result.append(self.createTerm(
                schoolyear,
                schoolyear_id_getter(schoolyear),
                '%s - All Terms' % schoolyear.title,
            ))

            
            for term in sorted(schoolyear.values(), key=lambda t: t.first):
                result.append(self.createTerm(
                    term,
                    term_id_getter(schoolyear, term),
                    '%s - %s' % (schoolyear.title, term.title),
                ))
        return result


def AllTermsVocabularyFactory():
    return AllTermsVocabulary


class OverallWeeklyDailyReportRequestView(OverallYearMonthReportRequestView):

    report_builder = 'overall_weekly_daily_report.xls'

    fields = field.Fields(IAllTerms)

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        if ISchoolYear.providedBy(term):
            schoolyear = term
            param = schoolyear_id_getter(schoolyear)
        else:
            param = term_id_getter(ISchoolYear(term), term)
        task.request_params['term'] = param


class OverallWeeklyDailyReportView(OverallReportBase):

    message_title = _('Overall report: by week and day')

    sheet_1_title = _('Student attendance - overall reports - by week')
    sheet_1_label = _('Overall - weekly')

    sheet_2_title = _('Student attendance - overall reports - by day')
    sheet_2_label = _('Overall - daily')

    def addExporters(self, progress):
        super(OverallWeeklyDailyReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Overall - weekly sheet'),
            active=False, progress=0.0)
        progress.add(
            'sheet2', title=_('Genearting Overall - daily sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'overall_weekly_daily_%s' % self.schoolyear.title

    @Lazy
    def schoolyear(self):
        return self._selected_year

    def __init__(self, *args, **kw):
        super(OverallWeeklyDailyReportView, self).__init__(*args, **kw)
        self.terms = []
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear_id, schoolyear in ISchoolYearContainer(app).items():
            if schoolyear_id_getter(schoolyear) == requested:
                self._selected_year = schoolyear
                self.terms = sorted(schoolyear.values(),
                                    key=lambda term: term.first)
                return
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    self._selected_year = schoolyear
                    self.terms.append(term)

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.task_progress.force('sheet1', progress=0.2)
        self.print_week_percentage_table(ws)
        self.print_legend_1(ws, row=6, col=5)
        self.finish('sheet1')

        self.task_progress.force('sheet2', active=True)
        ws = wb.add_sheet(self.sheet_2_label)
        self.print_headers(ws, self.sheet_2_title)
        self.task_progress.force('sheet2', progress=0.2)
        self.print_day_percentage_table(ws)
        self.print_legend_1(ws, row=6, col=5)
        self.finish('sheet2')

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Term selected'), style)
        if len(self.terms) > 1:
            title = 'All Terms'
        else:
            title = self.terms[0].title
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    def print_dates_percentage_cell(self, ws, row, col, filter_func, dates,
                                    borders):
        meetings = {}
        count = 0
        present = 0
        value = 0
        background = 'white'
        borders = self.translate_borders(borders)
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date in dates:
                meetings[meeting] = attendance_data
        for meeting, attendance_data in meetings.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        count += 1
                        if person_data['present']:
                            present += 1
        if count:
            value = (self.div(present, count) * 100)
            background = self.get_background_color(value)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

    def print_week_percentage_table(self, ws):
        # table title
        title = _('By week - percentage present')
        row = 6
        merge = 3
        ws.write_merge(row, row, 0,  merge, title, self.styles['table_title'])

        # table header
        row += 1
        col = 1
        headers = [_('All'), _('Girls'), _('Boys')]
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        # data rows
        row += 1
        borders = [1, 1, 1, 1]
        for week_no, week in enumerate(self.weeks):
            first = week[0].date
            title = translate(_('Week ${week} (w/b on ${date})',
                                mapping={'week': week_no + 1,
                                         'date': first.isoformat()}),
                              context=self.request)
            dates = [day.date for day in week]
            ws.write(row, 0, title)
            self.print_dates_percentage_cell(
                ws, row, 1, lambda person, person_data: True, dates, borders)
            self.print_dates_percentage_cell(
                ws, row, 2, is_female, dates, borders)
            self.print_dates_percentage_cell(
                ws, row, 3, is_male, dates, borders)
            row += 1

    def print_day_percentage_table(self, ws):
        # table title
        title = _('By day - percentage present')
        row = 6
        merge = 3
        ws.write_merge(row, row, 0,  merge, title, self.styles['table_title'])

        # table header
        row += 1
        col = 1
        headers = [_('All - % present'), _('Girls'), _('Boys')]
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        # data rows
        row += 1
        borders = [1, 1, 1, 1]
        for day in self.days:
            title = day.isoformat()
            ws.write(row, 0, title)
            dates = [day]
            self.print_dates_percentage_cell(
                ws, row, 1, lambda person, person_data: True, dates, borders)
            self.print_dates_percentage_cell(
                ws, row, 2, is_female, dates, borders)
            self.print_dates_percentage_cell(
                ws, row, 3, is_male, dates, borders)
            row += 1


class OverallReasonReportRequestView(OverallYearMonthReportRequestView):

    report_builder = 'overall_reason_report.xls'


class OverallReasonReportView(OverallReportBase):

    message_title = _('Overall report: reason for absence')

    sheet_1_title = _('Student attendance - Overall report - reason for absence')
    sheet_1_label = _('Overall - reason')

    def addExporters(self, progress):
        super(OverallReasonReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Overall - reason sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'overall_reason_%s' % self.schoolyear.title

    def print_headers(self, ws, title):
        row = 0
        # XXX: overwritten only for merge
        merge = (len(self.levels) + 1) + ((len(self.levels) + 1) * 3) + 1
        self.print_report_title(ws, title, row=row, merge=merge)

        row = 2
        today = getUtility(IDateManager).today
        self.print_report_date(ws, today, row=row)

        row += 1
        self.print_report_filters(ws, row=row)

    def print_legend_1(self, ws, row=None, col=None):
        labels = [
            (_('Below 5% absent (i.e. above 95% present)'), 95),
            (_('5% - 10% absent (i.e. between 94% and 90% present)'), 90),
            (_('10% - 20% absent (i.e. between 89% and 80% present)'), 80),
            (_('More than 20% absent (less than 80% absent)'), 0),
        ]
        merge = len(self.levels) + 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write_merge(row, row, col, col + merge, _('Legend'), style)
        row += 1
        borders = self.translate_borders([1, 1, 1, 1])
        for i, label_info in enumerate(labels):
            label, value = label_info
            background = self.get_background_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write_merge(row+i, row+i, col, col + merge, label, style)

    def print_legend_2(self, ws, row=None, col=None):
        labels = [
            (_('Dark blue = highest %'), 95),
            (_('Light blue = lowest %'), 10),
        ]
        merge = ((len(self.levels) + 1) * 3) - 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write_merge(row, row, col, col + merge, _('Legend'), style)
        row += 1
        borders = self.translate_borders([1, 1, 1, 1])
        for i, label_info in enumerate(labels):
            label, value = label_info
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write_merge(row+i, row+i, col, col + merge, label, style)

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)

        merge = (len(self.levels) + 1) + ((len(self.levels) + 1) * 3) + 1
        title = _('Year to date - Reason for absence')
        row = 5
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        merge = len(self.levels) + 1
        title = _('% students absent by class')
        row = 7
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
            'alignment: horizontal center; '
        )
        ws.write_merge(row, row, 0, merge, title, style)
        row = self.print_year_percentage_table(ws, 9, self.data)

        row = 7
        col = merge + 2
        merge = ((len(self.levels) + 1) * 3) - 1
        title = _('Reason for absence by class')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
            'alignment: horizontal center; '
        )
        ws.write_merge(row, row, col, col + merge, title, style)
        self.print_year_rate_table(ws, 8, col, self.data)

        self.print_legend_1(ws, row=20, col=0)
        self.print_legend_2(ws, row=20, col=len(self.levels) + 3)

        self.task_progress.force('sheet1', progress=0.5)

        row = 26
        for term in sorted(self.schoolyear.values(),
                           key=lambda t: t.first):
            original_row = row
            meetings = self.meetings_by_term(term)
            
            if not meetings:
                continue
            merge = (len(self.levels) + 1) + ((len(self.levels) + 1) * 3) + 1
            title = translate(_('${term} - Reason for absence',
                                mapping={'term': term.title}),
                              context=self.request)
            ws.write_merge(row, row, 0, merge, title,
                           self.styles['table_title'])

            merge = len(self.levels) + 1
            title = _('% students absent by class')
            row += 2
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
                'alignment: horizontal center; '
            )
            ws.write_merge(row, row, 0, merge, title, style)
            last_row = self.print_year_percentage_table(ws, row + 2, meetings)

            row = original_row + 2
            col = merge + 2
            merge = ((len(self.levels) + 1) * 3) - 1
            title = _('Reason for absence by class')
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
                'alignment: horizontal center; '
            )
            ws.write_merge(row, row, col, col + merge, title, style)
            last_row = self.print_year_rate_table(ws, row + 1, col, meetings)
            row = last_row + 2

        self.finish('sheet1')

    def print_year_percentage_cell(self, ws, row, col, filter_func, meetings,
                                   borders):
        value = 0
        background = 'white'
        count = 0
        present = 0
        borders = self.translate_borders(borders)
        for meeting, attendance_data in meetings.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        count += 1
                        if not person_data['present']:
                            present += 1
        value = (self.div(present, count) * 100)
        if count:
            background = self.get_reason_background_color(value)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

    def print_year_rate_cells(self, ws, row, col, filter_func, meetings,
                              last_row=False):
        count = 0
        sick = 0
        fees = 0
        other = 0
        for meeting, attendance_data in meetings.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        if not person_data['present']:
                            count += 1
                            reason = person_data['reason']
                            if reason == 'sick':
                                sick += 1
                            elif reason == 'fees':
                                fees += 1
                            else:
                                other += 1
        if last_row:
            borders = [0, 0, 1, 1]
        else:
            borders = [0, 0, 0, 1]
        borders = self.translate_borders(borders)
        value = (self.div(sick, count) * 100)
        background = self.get_gradient_color(value)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)
        col += 1
        if last_row:
            borders = [0, 0, 1, 0]
        else:
            borders = [0, 0, 0, 0]
        borders = self.translate_borders(borders)
        value = (self.div(fees, count) * 100)
        background = self.get_gradient_color(value)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)
        col += 1
        if last_row:
            borders= [0, 1, 1, 0]
        else:
            borders = [0, 1, 0, 0]
        borders = self.translate_borders(borders)
        value = (self.div(other, count) * 100)
        background = self.get_gradient_color(value)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)
        col += 1
        return col

    def print_year_percentage_table(self, ws, row, meetings):
        # section row
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Section'), style)
        ws.write(row, 1, _('Full school'), self.styles['column_header'])
        for i, level in enumerate(self.levels.values()):
            ws.write(row, 2 + i, level.title, self.styles['column_header'])
        
        # all row
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('All'), style)
        borders = [1, 1, 1, 1]
        self.print_year_percentage_cell(
            ws, row, 1, lambda person, person_data: True, meetings, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: person_data['level'] is level,
                meetings, borders
            )

        # girls boarding
        row += 1
        ws.write(row, 0, _('Girls Boarder (GB)'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(
            ws, row, 1, is_female_boarding, meetings, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_boarding(person, person_data)),
                meetings, borders
            )

        # girls day
        row += 1
        ws.write(row, 0, _('Girls Day (GD)'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(
            ws, row, 1, is_female_day, meetings, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_day(person, person_data)),
                meetings, borders
            )

        # boys boarding
        row += 1
        ws.write(row, 0, _('Boys Boarder (BB)'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(
            ws, row, 1, is_male_boarding, meetings, borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_boarding(person, person_data)),
                meetings, borders
            )

        # boys day
        row += 1
        ws.write(row, 0, _('Boys Day (BD)'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_male_day, meetings,
                                        borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_day(person, person_data)),
                meetings, borders
            )

        # girls
        row += 1
        ws.write(row, 0, _('Girls'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(ws, row, 1, is_female, meetings,
                                        borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_female(person, person_data)),
                meetings, borders
            )

        # boys
        row += 1
        ws.write(row, 0, _('Boys'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_male, meetings,
                                        borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_male(person, person_data)),
                meetings, borders
            )

        # boarding
        row += 1
        ws.write(row, 0, _('Boarding'), self.styles['block_top_row'])
        borders = [1, 1, 0, 1]
        self.print_year_percentage_cell(ws, row, 1, is_boarding, meetings,
                                        borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_boarding(person, person_data)),
                meetings, borders
            )

        # day
        row += 1
        ws.write(row, 0, _('Day'), self.styles['block_bottom_row'])
        borders = [0, 1, 1, 1]
        self.print_year_percentage_cell(ws, row, 1, is_day, meetings,
                                        borders)
        for i, level in enumerate(self.levels.values()):
            self.print_year_percentage_cell(
                ws, row, 2 + i,
                lambda person, person_data: (person_data['level'] is level and
                                             is_day(person, person_data)),
                meetings, borders
            )

        return row

    def print_year_rate_table(self, ws, row, col, meetings):
        original_col = col

        # table header
        merge = 2
        ws.write_merge(row, row, col, col + merge, _('Full school'),
                       self.styles['column_header'])
        col += merge + 1
        for i, level in enumerate(self.levels.values()):
            ws.write_merge(row, row, col, col + merge, level.title,
                           self.styles['column_header'])
            col +=  merge + 1

        # section row
        row += 1
        col = original_col
        reason_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left thin; '
            'alignment: horizontal center; '
        )
        reasons = [_('% sick'), _('% fees'), _('% other')]
        for title in reasons:
            ws.write(row, col, title, reason_header)
            col += 1
        for i, level in enumerate(self.levels.values()):
            for title in reasons:
                ws.write(row, col, title, reason_header)
                col += 1

        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, lambda person, person_data: True,
            meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: person_data['level'] is level,
                meetings)

        # girls boarding
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_female_boarding, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_boarding(person, person_data)),
                meetings
            )

        # girls day
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_female_day, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_female_day(person, person_data)),
                meetings
            )

        # boys boarding
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_male_boarding, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_boarding(person, person_data)),
                meetings
            )

        # boys day
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_male_day, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (
                    person_data['level'] is level and
                    is_male_day(person, person_data)),
                meetings
            )

        # girls
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_female, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_female(person, person_data)),
                meetings
            )

        # boys
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_male, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_male(person, person_data)),
                meetings
            )

        # boarding
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_boarding, meetings)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_boarding(person, person_data)),
                meetings
            )

        # day
        row += 1
        col = self.print_year_rate_cells(
            ws, row, original_col, is_day, meetings, last_row=True)
        for i, level in enumerate(self.levels.values()):
            col = self.print_year_rate_cells(
                ws, row, col,
                lambda person, person_data: (person_data['level'] is level and
                                             is_day(person, person_data)),
                meetings,
                last_row=True
            )
        return row


class OverallYearMonthWeeklyDailyReportRequestView(
        OverallYearMonthReportRequestView):

    report_builder = 'overall_year_month_weekly_daily_report.xls'


class TeacherOverallReportBase(OverallReportBase):

    def is_instructor(self, person, date):
        person = removeSecurityProxy(person)
        r = Instruction.bind(
            instructor=person).on(date).any(ACTIVE).relationships
        for link_info in r:
            target = removeSecurityProxy(link_info.target)
            if ISchoolYear(target) == self.schoolyear:
                return True
        return False

    @Lazy
    def SCORE_TO_REASON(self):
        result = OrderedDict([
            ('P', 'present'),
            ('Pl', 'present_late'),
            ('Pc', 'present_commitment'),
            ('As', 'sick'),
            ('Aa', 'authorised'),
            ('Au', 'unauthorised'),
            ('NT', 'not_timetabled'),
            ('Ss', 'suspended'),

        ])
        return result

    @Lazy
    def data(self):
        result = {}
        app = ISchoolToolApplication(None)
        persons = app['persons']
        teachers = IGroupContainer(self.schoolyear).get('teachers')
        meetings_count = len(self.meetings)
        persons_count = len(app['persons'])
        for meeting_no, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            am_requirement = TeacherAMRequirement(meeting)
            am_evaluations = IEvaluateRequirement(am_requirement)
            teachers_on_date = list(teachers.members.on(date))
            am_attendance = {}
            for person_no, person in enumerate(persons.values()):
                self.progress(
                    'data', normalized_progress(
                        meeting_no, meetings_count, person_no, persons_count)
                )
                if (person in teachers_on_date or
                    self.is_instructor(person, date)):
                    am_score = am_evaluations.getEvaluation(
                        person, am_requirement, default=UNSCORED)
                    if not self.isSkipped(am_score):
                        if person not in am_attendance:
                            am_attendance[person] = {}
                        is_present = self.isPresent(am_score)
                        am_attendance[person]['present'] = is_present
                        am_attendance[person]['reason'] = self.reason(am_score)
                        am_attendance[person]['value'] = am_score.value
                        am_attendance[person]['score'] = am_score
            result[meeting] = {
                am_requirement: am_attendance,
            }
        return result

    def instructor_courses(self, person):
        person = removeSecurityProxy(person)
        r = Instruction.bind(instructor=person).relationships
        result = set()
        for link_info in r:
            target = removeSecurityProxy(link_info.target)
            if ISchoolYear(target) == self.schoolyear:
                for course in target.courses:
                    result.add(course.title)
        return ', '.join(sorted(result))


class OverallYearMonthWeeklyDailyReportView(TeacherOverallReportBase):

    message_title = _('Overall report: by various time frames')

    sheet_1_title = _('Teacher attendance - overall reports - by '
                      'year and by term')
    sheet_1_label = _('Overall - Yr and Term')

    sheet_2_title = _('Teacher attendance - overall reports - by month')
    sheet_2_label = _('Overall - monthly')

    sheet_3_title = _('Teacher attendance - overall reports - by week')
    sheet_3_label = _('Overall - weekly')

    sheet_4_title = _('Teacher attendance - overall reports - by day')
    sheet_4_label = _('Overall - daily')

    def addExporters(self, progress):
        super(OverallYearMonthWeeklyDailyReportView, self).addExporters(
            progress)
        progress.add(
            'sheet1', title=_('Generating Overall - Yr and Term sheet'),
            active=False, progress=0.0)
        progress.add(
            'sheet2', title=_('Generating Overall - monthly sheet'),
            active=False, progress=0.0)
        progress.add(
            'sheet3', title=_('Generating Overall - weekly sheet'),
            active=False, progress=0.0)
        progress.add(
            'sheet4', title=_('Generating Overall - daily sheet'),
            active=False, progress=0.0)


    @property
    def base_filename(self):
        return 'overall_year_month_weekly_daily_%s' % self.schoolyear.title

    def export_report(self, wb):
        self.add_sheet_1(wb)
        self.add_sheet_2(wb)
        self.add_sheet_3(wb)
        self.add_sheet_4(wb)

    @Lazy
    def terms(self):
        result = []
        for term in sorted(self.schoolyear.values(),
                           key=lambda t: t.first):
            if self.meetings_by_term(term):
                result.append(term)
        return result

    def add_sheet_4(self, wb):
        self.task_progress.force('sheet4', active=True)
        ws = wb.add_sheet(self.sheet_4_label)
        self.print_headers(ws, self.sheet_4_title)
        self.print_legend_1(ws, row=6, col=5)

        row = 5
        merge = 3
        title = _('Daily')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])
        headers = [
            'Total % present',
            '% female present',
            '% male present',
        ]
        row += 1
        col = 1
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        row += 1
        meetings_by_day = {}
        for meeting, attendance in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date not in meetings_by_day:
                meetings_by_day[date] = {}
            meetings_by_day[date][meeting] = attendance

        for day in self.days:
            d = self.tzinfo.localize(
                datetime.datetime(day.year, day.month, day.day)).date()
            meetings = meetings_by_day.get(d)
            if not meetings:
                continue
            title = day.isoformat()
            ws.write(row, 0, title)
            self.print_percentage_cell(
                ws, row, 1, lambda person, person_data: True, meetings)
            self.print_percentage_cell(
                ws, row, 2, is_female, meetings)
            self.print_percentage_cell(
                ws, row, 3, is_male, meetings)
            row += 1
        self.finish('sheet4')

    def add_sheet_3(self, wb):
        self.task_progress.force('sheet3', active=True)
        ws = wb.add_sheet(self.sheet_3_label)
        self.print_headers(ws, self.sheet_3_title)
        self.print_legend_1(ws, row=6, col=5)

        row = 5
        merge = 3
        title = _('Weekly')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])
        headers = [
            'Total % present',
            '% female present',
            '% male present',
        ]
        row += 1
        col = 1
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        row += 1
        meetings_by_week = {}
        for week in self.weeks:
            first = week[0].date
            last = week[-1].date
            for meeting, attendance in self.data.items():
                date = meeting.dtstart.astimezone(self.tzinfo).date()
                if first <= date <= last:
                    if (first, last) not in meetings_by_week:
                        meetings_by_week[(first, last)] = {}
                    meetings_by_week[(first, last)][meeting] = attendance
        for week_no, week in enumerate(self.weeks):
            first = week[0].date
            last = week[-1].date
            title = translate(_('Week ${week} (w/b on ${date})',
                                mapping={'week': week_no + 1,
                                         'date': first.isoformat()}),
                              context=self.request)
            ws.write(row, 0, title)
            meetings = meetings_by_week.get((first, last))
            if meetings:
                self.print_percentage_cell(
                    ws, row, 1, lambda person, person_data: True, meetings)
                self.print_percentage_cell(
                    ws, row, 2, is_female, meetings)
                self.print_percentage_cell(
                    ws, row, 3, is_male, meetings)
                row += 1
        self.finish('sheet3')

    def add_sheet_2(self, wb):
        self.task_progress.force('sheet2', active=True)
        ws = wb.add_sheet(self.sheet_2_label)
        self.print_headers(ws, self.sheet_2_title)
        self.print_legend_1(ws, row=6, col=5)

        row = 5
        merge = 3
        title = _('Monthly')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])
        headers = [
            'Total % present',
            '% female present',
            '% male present',
        ]
        row += 1
        col = 1
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        row += 1
        meetings_by_month = {}
        for month_key, meetings in self.meetings_by_month.items():
            k = month_key[1:]
            if k not in meetings_by_month:
                meetings_by_month[k] = {}
            meetings_by_month[k].update(meetings)

        filter_funcs = [
            lambda person, person_data: True,
            is_female,
            is_male,
        ]

        row_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )

        for year, month in sorted(meetings_by_month.keys()):
            meetings = meetings_by_month[(year, month)]
            if not meetings:
                continue
            ws.write(row, 0, month_names[month], row_header)
            col = 1
            for i, filter_func in enumerate(filter_funcs):
                self.print_percentage_cell(
                    ws, row, col + i, filter_func, meetings)
            row += 1

        self.finish('sheet2')

    def add_sheet_1(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)

        self.print_legend_1(ws, row=6, col=7)
        self.print_legend_2(ws, row=13, col=7)

        row = 5
        merge = 5
        title = _('Year to date - % teachers present')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])
        headers = [
            'Total % present',
            '% full time present',
            '% part time present',
            '% female present',
            '% male present',
        ]
        row += 1
        col = 1
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        column_style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )

        row += 1
        ws.write(row, 0, _('Year to date'), column_style)
        filter_funcs = [
            lambda person, person_data: True,
            is_full_time,
            is_part_time,
            is_female,
            is_male,
        ]
        col = 1
        for i, filter_func in enumerate(filter_funcs):
            self.print_percentage_cell(ws, row, col + i, filter_func, self.data)

        row = 12
        title = _('Year to date - % teachers in different boundaries')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        limit_filter_funcs = [
            lambda value: value >= 95,
            lambda value: 90 <= value < 95,
            lambda value: 80 <= value < 90,
            lambda value: value < 80,
        ]
        
        row += 1
        col = 1
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        rate_row_header = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )

        row += 1
        row_headers = [
            'Teachers above 95%',
            'Teachers less than 95%',
            'Teachers less than 90%',
            'Teachers less than 80%',
        ]
        for i, header in enumerate(row_headers):
            ws.write(row + i, 0, header, rate_row_header)

        for i, limit_func in enumerate(limit_filter_funcs):
            col = 1
            for j, filter_func in enumerate(filter_funcs):
                self.print_rate_cell(ws, row + i, col + j, limit_func,
                                     filter_func, self.data)

        row = 19
        title = _('By term - % teachers present')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        row += 1
        col = 1
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        row += 1
        col = 1
        for term in sorted(self.schoolyear.values(), key=lambda t: t.first):
            meetings = self.meetings_by_term(term)
            if not meetings:
                continue
            ws.write(row, 0, term.title, column_style)
            for i, filter_func in enumerate(filter_funcs):
                self.print_percentage_cell(
                    ws, row, col + i, filter_func, meetings)
            row += 1

        row += 1
        for term in sorted(self.schoolyear.values(), key=lambda t: t.first):
            meetings = self.meetings_by_term(term)
            if not meetings:
                continue
            title = translate(
                _('${term} - % teachers in different boundaries',
                  mapping={'term': term.title}),
                context=self.request),
            ws.write_merge(row, row, 0, merge, title,
                           self.styles['table_title'])
            row += 1
            col = 1
            for i, header in enumerate(headers):
                ws.write(row, col + i, header, self.styles['column_header'])

            row += 1
            col = 0
            for i, header in enumerate(row_headers):
                ws.write(row + i, col, header, rate_row_header)

            for i, limit_func in enumerate(limit_filter_funcs):
                col = 1
                for j, filter_func in enumerate(filter_funcs):
                    self.print_rate_cell(ws, row + i, col + j, limit_func,
                                         filter_func, meetings)
            row += i + 2
        self.finish('sheet1')

    def print_rate_cell(self, ws, row, col, limit, filter_func, meetings):
        borders = self.translate_borders([1, 1, 1, 1])
        background = 'white'
        count = 0
        present = 0
        persons = {}
        for meeting, attendance_data in self.data.items():
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        if person not in persons:
                            persons[person] = {'count': 0, 'present': 0}
                        persons[person]['count'] += 1
                        if person_data['present']:
                            persons[person]['present'] += 1
        for data in persons.values():
            value = self.div(data['present'], data['count']) * 100
            if limit(value):
                count += 1
                present += data['present']
        value = (self.div(count, len(persons)) * 100)
        if value:
            background = self.get_gradient_color(value)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

    def print_percentage_cell(self, ws, row, col, filter_func, meetings):
        value = 0
        background = 'white'
        count = 0
        present = 0
        borders = self.translate_borders([1, 1, 1, 1])
        for meeting, attendance_data in meetings.items():
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        count += 1
                        if person_data['present']:
                            present += 1
        value = (self.div(present, count) * 100)
        if count:
            background = self.get_background_color(value)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)


class OverallTeacherReasonReportRequestView(OverallYearMonthReportRequestView):

    report_builder = 'overall_teacher_reason_report.xls'


class OverallTeacherReasonReportView(TeacherOverallReportBase):

    message_title = _('Overall report: breakdown by attendance code')

    sheet_1_title = _('Teacher attendance - Overall report - reason for absence')
    sheet_1_label = _('Overall - reason')

    def isSkipped(self, score):
        if (score in (UNSCORED, None)
            or score.value is UNSCORED):
            return True
        return False

    def addExporters(self, progress):
        super(OverallTeacherReasonReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Overall - reason sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'overall_teacher_reason_%s' % self.schoolyear.title

    def print_year_rate_cells(self, ws, row, col, filter_func, meetings,
                              last_row=False):
        count = 0
        present = 0
        present_late = 0
        present_commitment = 0
        sick = 0
        authorised = 0
        unauthorised = 0
        for meeting, attendance_data in meetings.items():
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        reason = person_data['reason']
                        if reason in ('present', 'present_late',
                                      'present_commitment',
                                      'sick', 'authorised', 'unauthorised'):
                            count += 1
                        if reason == 'present':
                            present += 1
                        elif reason == 'present_late':
                            present_late += 1
                        elif reason == 'present_commitment':
                            present_commitment += 1
                        elif reason == 'sick':
                            sick += 1
                        elif reason == 'authorised':
                            authorised += 1
                        elif reason == 'unauthorised':
                            unauthorised += 1
        l = [present, present_late, present_commitment,
             sick, authorised, unauthorised]
        borders = self.translate_borders([1, 1, 1, 1])
        for attr in l:
            value = (self.div(attr, count) * 100)
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1

        col += 1
        count = 0
        not_timetabled = 0
        suspended = 0
        for meeting, attendance_data in meetings.items():
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if filter_func(person, person_data):
                        count += 1
                        reason = person_data['reason']
                        if reason == 'not_timetabled':
                            not_timetabled += 1
                        elif reason == 'suspended':
                            suspended += 1
        l = [not_timetabled, suspended]
        for attr in l:
            value = (self.div(attr, count) * 100)
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1

        return col

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_2(ws, row=6, col=11)

        row = 5
        merge = 9
        title = _('Reason for absence - year to date')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        column_headers = [
            '% Present',
            '% present - late or left early',
            '% present - internal commitments / CPD',
            '% absent - sick',
            '% absent - authorised',
            '% absent - unauthorised',
        ]

        not_timetabled_headers = [
            '% not timetabled',
            '% suspended',
        ]

        row_filters_data = [
            (_('Total'), lambda person, person_data: True),
            (_('Full time'), is_full_time),
            (_('Part time'), is_part_time),
            (_('Female'), is_female),
            (_('Male'), is_male),
        ]

        row += 1
        col = 1
        for i, header in enumerate(column_headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        col += len(column_headers) + 1
        for i, header in enumerate(not_timetabled_headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        row += 1
        row_header_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom no_line, left no_line; '
        )
        for i, filter_info in enumerate(row_filters_data):
            col = 0
            header, filter_func = filter_info
            ws.write(row, col, header, row_header_style)
            col += 1
            self.print_year_rate_cells(ws, row, col, filter_func, self.data)
            row += 1

        row = 13
        for term in sorted(self.schoolyear.values(),
                           key=lambda t: t.first):
            meetings = self.meetings_by_term(term)
            
            if not meetings:
                continue
            title = translate(_('${term} - Reason for absence',
                                mapping={'term': term.title}),
                              context=self.request)
            ws.write_merge(row, row, 0, merge, title,
                           self.styles['table_title'])

            row += 1
            col = 1
            for i, header in enumerate(column_headers):
                ws.write(row, col + i, header, self.styles['column_header'])

            row += 1
            for i, filter_info in enumerate(row_filters_data):
                col = 0
                header, filter_func = filter_info
                ws.write(row, col, header, row_header_style)
                col += 1
                self.print_year_rate_cells(ws, row, col, filter_func, meetings)
                row += 1

            row += 1

        self.finish('sheet1')


class DailyReportRequestView(RequestXLSReportDialog):

    fields = field.Fields(IRequestAbsencesByDayForm)

    def updateTaskParams(self, task):
        date = self.form_params.get('date')
        if date is not None:
            day = '%d-%02d-%02d' % (date.year, date.month, date.day)
            task.request_params['day'] = day


class StudentDailyReportRequestView(DailyReportRequestView):

    report_builder = 'student_daily_report.xls'


class TeacherDailyReportRequestView(DailyReportRequestView):

    report_builder = 'teacher_daily_report.xls'


DailyReport_date = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.request.util.today,
    view=DailyReportRequestView,
    field=IRequestAbsencesByDayForm['date']
    )


class TermRate(object):

    def term_rate(self, person, meetings, requirement_factories):
        count = 0
        present = 0
        for meeting, attendance_data in meetings.items():
            requirements = [factory(meeting)
                            for factory in requirement_factories]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for p, person_data in attendance.items():
                    if person is p:
                        score = person_data['score']
                        if (score not in [UNSCORED, None] and
                            score.value not in [UNSCORED, None]):
                            ss = score.scoreSystem
                            if not ss.isSkipped(score):
                                count += 1
                                if person_data['present']:
                                    present += 1
        return self.div(present, count) * 100, count, present


class DailyReportViewBase(TermRate):

    @Lazy
    def term(self):
        for term in self.schoolyear.values():
            if self.day in term:
                return term

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Date selected for data'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.day.strftime('%d/%m/%Y'), style)

    @Lazy
    def day(self):
        day = self.request.get('day', None)
        year, month, day = [int(part) for part in day.split('-')]
        return datetime.date(year, month, day)

    @Lazy
    def schoolyear(self):
        app = ISchoolToolApplication(None)
        schoolyears = ISchoolYearContainer(app)
        for schoolyear in schoolyears.values():
            if schoolyear.first <= self.day <= schoolyear.last:
                return schoolyear

    def days_missed(self, person, last_10_meetings, requirement_factories):
        result = 0
        for meeting, attendance_data in self.data.items():
            if meeting not in last_10_meetings:
                continue
            requirements = [factory(meeting)
                            for factory in requirement_factories]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for p, person_data in attendance.items():
                    if person is p:
                        if not person_data['present']:
                            result += 1
        return result

    def contact_number(self, person):
        result = []
        phones = ['home_phone', 'work_phone', 'mobile_phone']
        contact = IContact(person)
        for phone in phones:
            result.append(getattr(contact, phone))
        return ', '.join(filter(None, result))

    def get_fee_collection_rate(self, person):
        term = self.term
        credits = [fee.amount for fee in IFeeContainer(person).values()
                   if term.first <= fee.date <= term.last and
                   ICredit.providedBy(fee)]
        debits = [fee.amount for fee in IFeeContainer(person).values() 
                  if term.first <= fee.date <= term.last and
                  IDebit.providedBy(fee)]
        fee_collection_rate = 0
        if debits:
            fee_collection_rate = (sum(credits) * 100)/ sum(debits)
        return fee_collection_rate

    def getParents(self, person):
        result = set()
        for contact in IContactable(person).contacts.any(meaning=PARENT):
            result.add(contact)
        return sorted(result, key=lambda c: c.title)


class TeacherDailyReportView(DailyReportViewBase,
                             TeacherOverallReportBase):

    message_title = _('Teacher report: daily absence')

    sheet_1_title = _('Teacher attendance - teacher reports - daily absences')
    sheet_1_label = _('Teacher - daily')

    def addExporters(self, progress):
        super(TeacherDailyReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Teacher - daily sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'teacher_daily_%s' % self.day

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=11, col=7)

        row = 5
        merge = 5
        title = _('Daily attendance report')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        row = 7
        ws.write(row, 1, _('# teachers'), self.styles['column_header'])
        ws.write(row, 2, _('% teachers'), self.styles['column_header'])

        row = 8
        row_header = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Present'), row_header)
        ws.write(row + 1, 0, _('Absent'), row_header)

        present = 0
        absent = 0
        count = 0
        persons = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date != self.day:
                continue
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    persons[person] = person_data
                    count += 1
                    if person_data['present']:
                        present += 1
                    else:
                        absent += 1
        borders = self.translate_borders([1, 1, 1, 1])

        value = present
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(8, 1, value, style)

        value = (self.div(present, count) * 100)
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(8, 2, value, style)

        value = absent
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(9, 1, value, style)

        value = (self.div(absent, count) * 100)
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(9, 2, value, style)

        row = 11
        headers = [
            'Last name',
            'First name',
            'Attendance code',
            'Attendance rate this term',
            'Number days missed in the last 10 days',
            'Teacher contact number',
        ]
        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])


        row = 12
        meetings = self.meetings_by_term(self.term)
        sort_key = lambda data: (data[0].last_name, data[0].first_name)

        last_10_meetings = []
        for i, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date <= self.day:
                last_10_meetings.append(meeting)
        last_10_meetings = last_10_meetings[-10:]

        requirement_factories = [TeacherAMRequirement]

        for person, person_data in sorted(persons.items(), key=sort_key):

            if person_data['present']:
                continue

            ws.write(row, 0, person.last_name)
            ws.write(row, 1, person.first_name)
            ws.write(row, 2, person_data['value'])

            value, count, present = self.term_rate(person,
                                                   meetings,
                                                   requirement_factories)
            if count:
                background = self.get_background_color(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, 3, value, style)

            ws.write(row, 4, self.days_missed(person,
                                              last_10_meetings,
                                              requirement_factories))
            ws.write(row, 5, self.contact_number(person))

            row += 1

        self.finish('sheet1')


class StudentDailyReportView(DailyReportViewBase,
                             OverallReportBase):

    message_title = _('Pupil report: daily absence')

    sheet_1_title = _('Student attendance - Pupil reports - daily absences')
    sheet_1_label = _('Pupil - daily')

    def addExporters(self, progress):
        super(StudentDailyReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Teacher - daily sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'student_daily_%s' % self.day

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=11, col=11)

        row = 5
        merge = 5
        title = _('Daily attendance report')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        row = 7
        ws.write(row, 1, _('# students'), self.styles['column_header'])
        ws.write(row, 2, _('% students'), self.styles['column_header'])

        row = 8
        row_header = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        )
        ws.write(row, 0, _('Present'), row_header)
        ws.write(row + 1, 0, _('Absent'), row_header)

        present = 0
        absent = 0
        persons = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date != self.day:
                continue
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person not in persons:
                        persons[person] = {
                            'count': 0,
                            'present': 0,
                            'absent': 0,
                            'value': set(),
                        }
                    persons[person]['count'] += 1
                    persons[person]['value'].add(person_data['value'])
                    if person_data['present']:
                        persons[person]['present'] += 1
                    else:
                        persons[person]['absent'] += 1
        borders = self.translate_borders([1, 1, 1, 1])
        count = len(persons)
        for data in persons.values():
            person_present = self.div(data['present'], data['count'])
            person_absent = self.div(data['absent'], data['count'])
            present += person_present
            absent += person_absent

        value = present
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(8, 1, value, style)

        value = (self.div(present, count) * 100)
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(8, 2, value, style)

        value = absent
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(9, 1, value, style)

        value = (self.div(absent, count) * 100)
        background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(9, 2, value, style)

        row = 11
        headers = [
            'Class',
            'Last name',
            'First name',
            'Day / Boarding',
            'Attendance code',
            'Attendance rate this term',
            'Number days missed in the last 10 days',
            'Fee collection rate this term',
            'Parent name',
            'Parent contact number',
        ]

        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])

        row = 12
        meetings = self.meetings_by_term(self.term)
        sort_key = lambda data: (self.level(data[0], self.day).title
                                 if self.level(data[0], self.day) else '',
                                 data[0].last_name,
                                 data[0].first_name)

        last_10_meetings = []
        for i, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date <= self.day:
                last_10_meetings.append(meeting)
        last_10_meetings = last_10_meetings[-10:]

        requirement_factories = [StudentAMRequirement, StudentPMRequirement]

        for person, person_data in sorted(persons.items(), key=sort_key):

            if not person_data['absent']:
                continue

            level = self.level(person, self.day).title if self.level(person, self.day) else ''
            ws.write(row, 0, level)
            ws.write(row, 1, person.last_name)
            ws.write(row, 2, person.first_name)
            ws.write(row, 3, get_day_boarding(person))
            ws.write(row, 4, ', '.join(person_data['value']))
            value, count, present = self.term_rate(person,
                                                   meetings,
                                                   requirement_factories)
            if count:
                background = self.get_background_color(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, 5, value, style)

            value = self.days_missed(person,
                                      last_10_meetings,
                                      requirement_factories) / 2.0
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 6, value, style)

            value = self.get_fee_collection_rate(person)
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 7, value, style)

            parents = self.getParents(person)
            phones = lambda x: ', '.join(filter(None, [
                x.home_phone,
                x.work_phone,
                x.mobile_phone]))
            if parents:
                ws.write(row, 8, ', '.join([c.title for c in parents]))
                ws.write(row, 9, ', '.join([phones(c) for c in parents]))

            row += 1

        self.finish('sheet1')


class StudentPersistantReportRequestView(RequestXLSReportDialog):

    report_builder = 'student_persistant_report.xls'


class PersistantReportViewBase(object):

    def print_report_filters(self, ws, row):
        pass

    def print_report_date(self, ws, today, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Date report produced'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, today.strftime('%d/%m/%Y'), style)

    @Lazy
    def day(self):
        if 'day' in self.request:
            day = self.request.get('day', None)
            year, month, day = [int(part) for part in day.split('-')]
            return datetime.date(year, month, day)
        return getUtility(IDateManager).today


class StudentPersistantReportView(PersistantReportViewBase,
                                  DailyReportViewBase,
                                  OverallReportBase):

    message_title = _('Pupil report: persistent absence 2 weeks')

    sheet_1_title = _('Student attendance - Pupil reports - '
                      'persistant absence during the last 2 weeks')
    sheet_1_label = _('Pupil - persistant 2 weeks')

    def addExporters(self, progress):
        super(StudentPersistantReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Pupil - persistant 2 weeks sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'student_persistant_'

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=6, col=10)

        row = 4
        merge = 5
        title = _('Report showing students absent more than 3 days '
                  'in the last 10 day period')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        row = 6
        headers = [
            'Class',
            'Last name',
            'First name',
            'Day / Boarding',
            '# days absent in last 2 weeks',
            'Attendance rate this term',
            'Fee collection rate this term',
            'Parent name',
            'Parent contact number',
        ]

        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])

        row += 1
        meetings = self.meetings_by_term(self.term)
        sort_key = lambda data: (self.level(data[0], self.day).title
                                 if self.level(data[0], self.day) else '',
                                 data[0].last_name,
                                 data[0].first_name)

        last_10_meetings = []
        for i, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date <= self.day:
                last_10_meetings.append(meeting)
        last_10_meetings = last_10_meetings[-10:]

        requirement_factories = [StudentAMRequirement, StudentPMRequirement]

        persons = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > self.day:
                continue
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person not in persons:
                        persons[person] = {
                            'count': 0,
                            'present': 0,
                            'absent': 0,
                            'value': set(),
                        }
                    persons[person]['count'] += 1
                    persons[person]['value'].add(person_data['value'])
                    if person_data['present']:
                        persons[person]['present'] += 1
                    else:
                        persons[person]['absent'] += 1
        borders = self.translate_borders([1, 1, 1, 1])

        for person, person_data in sorted(persons.items(), key=sort_key):

            days_missed = self.days_missed(person,
                                           last_10_meetings,
                                           requirement_factories) / 2.0
            if days_missed < 3:
                continue
            level = self.level(person, self.day).title if self.level(person, self.day) else ''
            ws.write(row, 0, level)
            ws.write(row, 1, person.last_name)
            ws.write(row, 2, person.first_name)
            ws.write(row, 3, get_day_boarding(person))
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 4, days_missed, style)

            value, count, present = self.term_rate(person,
                                                   meetings,
                                                   requirement_factories)
            if count:
                background = self.get_background_color(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, 5, value, style)

            value = self.get_fee_collection_rate(person)
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 6, value, style)

            parents = self.getParents(person)
            phones = lambda x: ', '.join(filter(None, [
                x.home_phone,
                x.work_phone,
                x.mobile_phone]))
            if parents:
                ws.write(row, 7, ', '.join([c.title for c in parents]))
                ws.write(row, 8, ', '.join([phones(c) for c in parents]))

            row += 1

        self.finish('sheet1')


class TeacherPersistantReportRequestView(RequestXLSReportDialog):

    report_builder = 'teacher_persistant_report.xls'


class TeacherPersistantReportView(PersistantReportViewBase,
                                  DailyReportViewBase,
                                  TeacherOverallReportBase):

    message_title = _('Teacher report: persistent absence 2 weeks')

    sheet_1_title = _('Teacher attendance - Teacher reports - persistant '
                      'absence during the last 2 weeks')
    sheet_1_label = _('Teacher - persistant 2 weeks')

    def addExporters(self, progress):
        super(TeacherPersistantReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Teacher - persistant 2 weeks sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'teacher_persistant_'

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=6, col=5)

        row = 4
        merge = 3
        title = _('Teachers absent more than 3 days in the last '
                  '10 day period')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        present = 0
        absent = 0
        count = 0
        persons = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > self.day:
                continue
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    persons[person] = person_data
                    count += 1
                    if person_data['present']:
                        present += 1
                    else:
                        absent += 1
        borders = self.translate_borders([1, 1, 1, 1])

        row = 6
        headers = [
            'Last name',
            'First name',
            '# days absent in last 2 weeks',
            'Average attendance this term',
        ]
        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])


        row += 1
        meetings = self.meetings_by_term(self.term)
        sort_key = lambda data: (data[0].last_name, data[0].first_name)

        last_10_meetings = []
        for i, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date <= self.day:
                last_10_meetings.append(meeting)
        last_10_meetings = last_10_meetings[-10:]

        requirement_factories = [TeacherAMRequirement]

        for person, person_data in sorted(persons.items(), key=sort_key):

            days_missed = self.days_missed(person,
                                           last_10_meetings,
                                           requirement_factories)
            if days_missed < 3:
                continue

            ws.write(row, 0, person.last_name)
            ws.write(row, 1, person.first_name)
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 2, days_missed, style)

            value, count, present = self.term_rate(person,
                                                   meetings,
                                                   requirement_factories)
            if count:
                background = self.get_background_color(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, 3, value, style)

            row += 1

        self.finish('sheet1')


class StudentPersistant75ReportRequestView(RequestXLSReportDialog):

    report_builder = 'student_persistant_75_report.xls'

    fields = field.Fields(ICurrentTerm)

    def updateTaskParams(self, task):
        term = self.form_params.get('term')
        task.request_params['term'] = term_id_getter(ISchoolYear(term), term)


class StudentPersistant75ReportView(PersistantReportViewBase,
                                    DailyReportViewBase,
                                    OverallReportBase):

    message_title = _('Pupil report: persistent absence 75%')

    sheet_1_title = _('Student attendance - Pupil reports - persistant '
                      'absence showing students with less that 75% '
                      'attendance this term')
    sheet_1_label = _('Pupil - persistant 75%')

    def addExporters(self, progress):
        super(StudentPersistant75ReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Pupil - persistant 75% sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'student_persistant_75_'

    def print_report_date(self, ws, today, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Date report produced'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, today.strftime('%d/%m/%Y'), style)

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Term'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.term.title, style)

        row += 1

        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    @Lazy
    def term(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear_id, schoolyear in ISchoolYearContainer(app).items():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return term

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=6, col=10)

        row = 6
        merge = 8
        title = _('Report showing students with less than 75% attendance '
                  'this term')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        row = 8
        headers = [
            'Class',
            'Last name',
            'First name',
            'Day / Boarding',
            'Attendance rate this term',
            '# days missed in  last 10 days',
            'Fee collection rate this term',
            'Parent name',
            'Parent contact number',
        ]

        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])

        row += 1
        meetings = self.meetings_by_term(self.term)
        sort_key = lambda data: (self.level(data[0], self.day).title
                                 if self.level(data[0], self.day) else '',
                                 data[0].last_name,
                                 data[0].first_name)

        last_10_meetings = []
        for i, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date <= self.day:
                last_10_meetings.append(meeting)
        last_10_meetings = last_10_meetings[-10:]

        requirement_factories = [StudentAMRequirement, StudentPMRequirement]

        persons = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > self.day:
                continue
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person not in persons:
                        persons[person] = {
                            'count': 0,
                            'present': 0,
                            'absent': 0,
                            'value': set(),
                        }
                    persons[person]['count'] += 1
                    persons[person]['value'].add(person_data['value'])
                    if person_data['present']:
                        persons[person]['present'] += 1
                    else:
                        persons[person]['absent'] += 1
        borders = self.translate_borders([1, 1, 1, 1])

        for person, person_data in sorted(persons.items(), key=sort_key):

            term_rate, count, present = self.term_rate(person,
                                                       meetings,
                                                       requirement_factories)
            if term_rate >= 75:
                continue

            days_missed = self.days_missed(person,
                                           last_10_meetings,
                                           requirement_factories) / 2.0
            level = self.level(person, self.day).title if self.level(person, self.day) else ''
            ws.write(row, 0, level)
            ws.write(row, 1, person.last_name)
            ws.write(row, 2, person.first_name)
            ws.write(row, 3, get_day_boarding(person))

            if count:
                background = self.get_background_color(term_rate)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, 4, term_rate, style)

            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 5, days_missed, style)

            value = self.get_fee_collection_rate(person)
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 6, value, style)

            parents = self.getParents(person)
            phones = lambda x: ', '.join(filter(None, [
                x.home_phone,
                x.work_phone,
                x.mobile_phone]))
            if parents:
                ws.write(row, 7, ', '.join([c.title for c in parents]))
                ws.write(row, 8, ', '.join([phones(c) for c in parents]))

            row += 1

        self.finish('sheet1')


class TeacherPersistant75ReportRequestView(
        StudentPersistant75ReportRequestView):

    report_builder = 'teacher_persistant_75_report.xls'


class TeacherPersistant75ReportView(PersistantReportViewBase,
                                    DailyReportViewBase,
                                    TeacherOverallReportBase):

    message_title = _('Teacher report: persistent absence 75%')

    sheet_1_title = _('Teacher attendance - Teacher reports - persistant '
                      'absence showing teachers with less that 75% '
                      'attendance this term')
    sheet_1_label = _('Teacher - persistant 75%')

    def addExporters(self, progress):
        super(TeacherPersistant75ReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Teacher - persistant 75% sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'teacher_persistant_75_'

    def print_report_date(self, ws, today, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Date report produced'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, today.strftime('%d/%m/%Y'), style)

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Term'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, self.term.title, style)

        row += 1

        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Year selected'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.schoolyear.title, style)

    @Lazy
    def term(self):
        app = ISchoolToolApplication(None)
        requested = self.request.get('term')
        for schoolyear_id, schoolyear in ISchoolYearContainer(app).items():
            for term in schoolyear.values():
                if term_id_getter(schoolyear, term) == requested:
                    return term

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=8, col=9)

        row = 6
        merge = 7
        title = _('Teachers with less than 75% attendance rate this term')
        ws.write_merge(row, row, 0, merge, title, self.styles['table_title'])

        present = 0
        absent = 0
        count = 0
        persons = {}
        for meeting, attendance_data in self.data.items():
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date > self.day:
                continue
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    persons[person] = person_data
                    count += 1
                    if person_data['present']:
                        present += 1
                    else:
                        absent += 1
        borders = self.translate_borders([1, 1, 1, 1])

        row = 7
        ws.write_merge(row, row, 5, 7,
                       'Subjects that this teacher teaches',
                       self.styles['column_header'])

        row = 8
        headers = [
            'Last name',
            'First name',
            'Part time / full time',
            '# days missed in  last 10 days',
            'Attendance rate this term',
            'Subject 1',
            'Subject 2',
            'Subject 3',
        ]
        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])

        row += 1
        meetings = self.meetings_by_term(self.term)
        sort_key = lambda data: (data[0].last_name, data[0].first_name)

        last_10_meetings = []
        for i, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            if date <= self.day:
                last_10_meetings.append(meeting)
        last_10_meetings = last_10_meetings[-10:]

        requirement_factories = [TeacherAMRequirement]

        for person, person_data in sorted(persons.items(), key=sort_key):

            term_rate, count, present = self.term_rate(person,
                                                       meetings,
                                                       requirement_factories)
            if term_rate >= 75:
                continue

            days_missed = self.days_missed(person,
                                           last_10_meetings,
                                           requirement_factories)

            ws.write(row, 0, person.last_name)
            ws.write(row, 1, person.first_name)
            ws.write(row, 2, get_shift_type(person) or '')
            style = self.easyxf(
                'font: bold off, color black; ',
                num_format_str='0')
            ws.write(row, 3, days_missed, style)

            if count:
                background = self.get_background_color(term_rate)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, 4, term_rate, style)
            ws.write(row, 5, IDemographics(person).get('subject_1'))
            ws.write(row, 6, IDemographics(person).get('subject_2'))
            ws.write(row, 7, IDemographics(person).get('subject_3'))

            row += 1

        self.finish('sheet1')


class TeacherAllReportRequestView(OverallYearMonthReportRequestView):

    report_builder = 'teacher_all_report.xls'


class AllReportViewBase(TermRate):

    def isSkipped(self, score):
        if (score in (UNSCORED, None)
            or score.value is UNSCORED):
            return True
        return False

    def get_column_total(self, table, column):
        result = []
        for row in table:
            for i, col in enumerate(row):
                if i == column:
                    result.append(col)
        return self.div(sum([p for r, c, p in result]) * 100,
                        sum([c for r, c, p in result]))


class TeacherAllReportView(AllReportViewBase, TeacherOverallReportBase):

    message_title = _('Teacher report: all teachers')

    sheet_1_title = _('Teacher attendance - Teacher reports - All '
                      'teachers attendance record')
    sheet_1_label = _('Teacher - all YTD')

    sheet_2_title = _('Teacher attendance - Teacher reports - All teachers '
                      'attendance record by week')
    sheet_2_label = _('Teacher all - weekly view')

    def addExporters(self, progress):
        super(TeacherAllReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Teacher - all YTD sheet'),
            active=False, progress=0.0)
        progress.add(
            'sheet2', title=_('Generating Teacher all - weekly view sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'teacher_all_'

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)

        attendance_headers = [
            '% Present',
            '% present - late or left early',
            '% present - internal commitments / CPD',
            '% absent - sick',
            '% absent - authorised',
            '% absent - unauthorised',
        ]

        not_timetabled_headers = [
            '% not timetabled',
            '% suspended',
        ]

        row = 5
        col = 3
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)
        merge = len(terms)
        ws.write_merge(row, row, col, col + merge, _('Summary'),
                       self.styles['column_header'])
        col += merge + 1
        ws.write_merge(row, row, col, col + 2,
                       _('Subjects that this teacher teaches'),
                       self.styles['column_header'])

        col += 4
        merge = 5
        ws.write_merge(row, row, col, col + merge, _('Year to date breakdown'),
                       self.styles['column_header'])
        col += merge + 2
        for term in terms:
            merge = 5
            ws.write_merge(row, row, col, col + merge,
                           '%s breakdown' % term.title,
                           self.styles['column_header'])
            col += merge + 2

        merge = 1
        ws.write_merge(row, row, col, col + merge,
                       _('Year to date - Not timetabled or suspended'),
                       self.styles['column_header'])
        col += merge + 2
        for term in terms:
            merge = 1
            ws.write_merge(row, row, col, col + merge, term.title,
                           self.styles['column_header'])
            col += merge + 2


        row += 1
        col = 0
        headers = [
            'Last name',
            'First name',
            'F / P time',
        ]
        for i, header in enumerate(headers):
            ws.write(row, col + i, header, self.styles['column_header'])

        col = 3
        for term in terms:
            ws.write(row, col, '%s %% present' % term.title,
                     self.styles['column_header'])
            col += 1

        ws.write(row, col, 'Year to date % present',
                 self.styles['column_header'])

        col += 1
        ws.write(row, col, 'Subject 1',
                 self.styles['column_header'])
        col += 1
        ws.write(row, col, 'Subject 2',
                 self.styles['column_header'])
        col += 1
        ws.write(row, col, 'Subject 3',
                 self.styles['column_header'])

        col += 2
        for header in attendance_headers:
            ws.write(row, col, header, self.styles['column_header'])
            col += 1

        col += 1
        for term in terms:
            for header in attendance_headers:
                ws.write(row, col, header, self.styles['column_header'])
                col += 1
            col += 1

        for header in not_timetabled_headers:
            ws.write(row, col, header, self.styles['column_header'])
            col += 1

        col += 1
        for term in terms:
            for header in not_timetabled_headers:
                ws.write(row, col, header, self.styles['column_header'])
                col += 1
            col += 1

        row += 1
        teachers = IGroupContainer(self.schoolyear)['teachers']
        meetings_by_term = {}
        borders = self.translate_borders([0, 0, 0, 0])
        term_total = {}
        year_breakdown = []
        term_breakdown = {}
        year_breakdown_2 = []
        term_breakdown_2 = {}
        for term in terms:
            meetings_by_term[term] = self.meetings_by_term(term)
            term_total[term] = []
            term_breakdown[term] = []
            term_breakdown_2[term] = []
        for teacher in sorted(teachers.members,
                              key=lambda p: (p.last_name, p.first_name)):
            ws.write(row, 0, teacher.last_name)
            ws.write(row, 1, teacher.first_name)
            ws.write(row, 2, get_shift_type(teacher))
            col = 3
            year_to_date = []
            for term in terms:
                term_rate, count, present = self.term_rate(
                    teacher,
                    meetings_by_term[term],
                    [TeacherAMRequirement])
                if count:
                    background = self.get_background_color(term_rate)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, term_rate, style)
                if count:
                    year_to_date.append((term_rate, count, present))
                    term_total[term].append((term_rate, count, present))
                col += 1
            teacher_year_to_date = [rate for rate, c, p in year_to_date]
            value = self.div(sum(teacher_year_to_date),
                             len(teacher_year_to_date))
            if teacher_year_to_date:
                background = self.get_background_color(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
            ws.write(row, col, IDemographics(teacher).get('subject_1'))
            col += 1
            ws.write(row, col, IDemographics(teacher).get('subject_2'))
            col += 1
            ws.write(row, col, IDemographics(teacher).get('subject_3'))

            col += 2
            col, year_break = self.print_breakdown(ws, row, col, teacher,
                                                   self.data)
            year_breakdown.append(year_break)
            col += 1
            for term in terms:
                meetings = meetings_by_term[term]
                col, term_break = self.print_breakdown(ws, row, col, teacher,
                                                       meetings)
                term_breakdown[term].append(term_break)
                col += 1

            col, year_break = self.print_breakdown_2(ws, row, col, teacher,
                                                     self.data)
            year_breakdown_2.append(year_break)
            col += 1
            for term in terms:
                meetings = meetings_by_term[term]
                col, term_break = self.print_breakdown_2(ws, row, col, teacher,
                                                         meetings)
                term_breakdown_2[term].append(term_break)
                col += 1

            row += 1

        col = 0
        merge = 2
        ws.write_merge(row, row, col, col + merge, 'Total',
                       self.styles['column_header'])
        col = 3
        year_to_date = []
        for term in terms:
            value = self.div(sum([p for r, c, p in term_total[term]]) * 100,
                             sum([c for r, c, p in term_total[term]]))
            if term_total[term]:
                background = self.get_background_color(value)
                year_to_date.append(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        value = self.div(sum(year_to_date), len(year_to_date))
        if year_to_date:
            background = self.get_background_color(value)
        else:
            background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

        col += 5
        codes = 6
        for i in range(codes):
            value = self.get_column_total(year_breakdown, i)
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        col += 1

        for term in terms:
            for i in range(codes):
                value = self.get_column_total(term_breakdown[term], i)
                background = self.get_gradient_color(value)
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
            col += 1

        codes = 2
        for i in range(codes):
            value = self.get_column_total(year_breakdown_2, i)
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        col += 1

        for term in terms:
            for i in range(codes):
                value = self.get_column_total(term_breakdown_2[term], i)
                background = self.get_gradient_color(value)
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
            col += 1

        self.finish('sheet1')

        self.task_progress.force('sheet2', active=True)
        ws = wb.add_sheet(self.sheet_2_label)
        self.print_headers(ws, self.sheet_2_title)

        row = 5
        ws.write(row, 0, 'Name', self.styles['column_header'])
        ws.write(row, 1, 'Total', self.styles['column_header'])

        meetings_by_week = {}
        for week in self.weeks:
            first = week[0].date
            last = week[-1].date
            for meeting, attendance in self.data.items():
                date = meeting.dtstart.astimezone(self.tzinfo).date()
                if first <= date <= last:
                    if (first, last) not in meetings_by_week:
                        meetings_by_week[(first, last)] = {}
                    meetings_by_week[(first, last)][meeting] = attendance
        col = 2
        for week_no, week in enumerate(self.weeks):
            first = week[0].date
            last = week[-1].date
            title = 'Week %d' % (week_no + 1)
            ws.write(row, col, title, self.styles['column_header'])
            col += 1
        borders = self.translate_borders([0, 0, 0, 0])
        row += 1
        for teacher in sorted(teachers.members,
                              key=lambda p: (p.last_name, p.first_name)):
            ws.write(row, 0, '%s, %s' % (teacher.last_name, teacher.first_name))
            meetings = self.data
            value, count, present = self.term_rate(teacher,
                                                   meetings,
                                                   [TeacherAMRequirement])
            if count:
                background = self.get_background_color(value)
            else:
                background = 'white'
            col = 1
            style = self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)

            col += 1
            for week in self.weeks:
                first = week[0].date
                last = week[-1].date
                meetings = meetings_by_week.get((first, last))
                value, count, present = self.term_rate(teacher,
                                                       meetings,
                                                       [TeacherAMRequirement])
                if count:
                    background = self.get_background_color(value)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
            row += 1
        self.finish('sheet2')

    @Lazy
    def terms(self):
        return sorted(self.schoolyear.values(), key=lambda t: t.first)

    def print_breakdown(self, ws, row, col, teacher, meetings):
        count = 0
        present = 0
        present_late = 0
        present_commitment = 0
        sick = 0
        authorised = 0
        unauthorised = 0
        for meeting, attendance_data in meetings.items():
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person is not teacher:
                        continue
                    reason = person_data['reason']
                    if reason in ('present', 'present_late',
                                  'present_commitment',
                                  'sick', 'authorised', 'unauthorised'):
                        count += 1
                    if reason == 'present':
                        present += 1
                    elif reason == 'present_late':
                        present_late += 1
                    elif reason == 'present_commitment':
                        present_commitment += 1
                    elif reason == 'sick':
                        sick += 1
                    elif reason == 'authorised':
                        authorised += 1
                    elif reason == 'unauthorised':
                        unauthorised += 1
        l = [present, present_late, present_commitment,
             sick, authorised, unauthorised]
        borders = self.translate_borders([1, 1, 1, 1])
        result = []
        for attr in l:
            value = (self.div(attr, count) * 100)
            result.append((value, count, attr))
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        return col, result

    def print_breakdown_2(self, ws, row, col, teacher, meetings):
        count = 0
        not_timetabled = 0
        suspended = 0
        for meeting, attendance_data in meetings.items():
            requirements = [TeacherAMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person is not teacher:
                        continue
                    count += 1
                    reason = person_data['reason']
                    if reason == 'not_timetabled':
                        not_timetabled += 1
                    elif reason == 'suspended':
                        suspended += 1
        l = [not_timetabled, suspended]
        borders = self.translate_borders([1, 1, 1, 1])
        result = []
        for attr in l:
            value = (self.div(attr, count) * 100)
            result.append((value, count, attr))
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        return col, result


class StudentAllReportRequestView(OverallYearMonthReportRequestView):

    report_builder = 'student_all_report.xls'


class StudentAllReportView(AllReportViewBase, OverallReportBase):

    @Lazy
    def data(self):
        result = {}
        app = ISchoolToolApplication(None)
        persons = app['persons']
        students = IGroupContainer(self.schoolyear).get('students')
        meetings_count = len(self.meetings)
        persons_count = len(app['persons'])
        for meeting_no, meeting in enumerate(self.meetings):
            date = meeting.dtstart.astimezone(self.tzinfo).date()
            am_requirement = StudentAMRequirement(meeting)
            am_evaluations = IEvaluateRequirement(am_requirement)
            pm_requirement = StudentPMRequirement(meeting)
            pm_evaluations = IEvaluateRequirement(pm_requirement)
            students_on_date = list(students.members.on(date))
            am_attendance = {}
            pm_attendance = {}
            for person_no, person in enumerate(persons.values()):
                self.progress(
                    'data', normalized_progress(
                        meeting_no, meetings_count, person_no, persons_count)
                )
                level = self.level(person, date)
                # XXX: level should be required for some reports only
                if (person in students_on_date or
                    self.is_learner(person, date)):
                    am_score = am_evaluations.getEvaluation(
                        person, am_requirement, default=UNSCORED)
                    if not self.isSkipped(am_score):
                        if person not in am_attendance:
                            am_attendance[person] = {}
                        is_present = self.isPresent(am_score)
                        am_attendance[person]['present'] = is_present
                        am_attendance[person]['level'] = level
                        am_attendance[person]['reason'] = self.reason(am_score)
                        am_attendance[person]['value'] = am_score.value
                        am_attendance[person]['score'] = am_score
                    pm_score = pm_evaluations.getEvaluation(
                        person, pm_requirement, default=UNSCORED)
                    if not self.isSkipped(pm_score):
                        if person not in pm_attendance:
                            pm_attendance[person] = {}
                        is_present = self.isPresent(pm_score)
                        pm_attendance[person]['present'] = is_present
                        pm_attendance[person]['level'] = level
                        pm_attendance[person]['reason'] = self.reason(pm_score)
                        pm_attendance[person]['value'] = pm_score.value
                        pm_attendance[person]['score'] = pm_score
            result[meeting] = {
                am_requirement: am_attendance,
                pm_requirement: pm_attendance,
            }
        return result

    message_title = _('Pupil report: all students')

    sheet_1_title = _('Student attendance - Pupil reports - All pupils '
                      'attendance record')
    sheet_1_label = _('Pupil - all students')

    def addExporters(self, progress):
        super(StudentAllReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Pupil - all students sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'student_all_'

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)

        attendance_headers = [
            '% present',
            '% present - late',
            '% absent - sick',
            '% absent - fees',
            '% absent other',
        ]

        not_timetabled_headers = [
            '% suspended',
        ]

        row = 5
        col = 5
        terms = sorted(self.schoolyear.values(), key=lambda t: t.first)
        merge = len(terms)
        ws.write_merge(row, row, col, col + merge, _('Summary'),
                       self.styles['column_header'])
        col += merge + 2
        merge = 4
        ws.write_merge(row, row, col, col + merge, _('Year to date breakdown'),
                       self.styles['column_header'])
        col += merge + 2
        for term in terms:
            merge = 4
            ws.write_merge(row, row, col, col + merge,
                           '%s breakdown' % term.title,
                           self.styles['column_header'])
            col += merge + 2

        merge = len(terms)
        ws.write_merge(row, row, col, col + merge,
                       _('Year to date - suspended'),
                       self.styles['column_header'])

        row += 1
        col = 0
        headers = [
            'Class',
            'Last name',
            'First name',
            'Gender',
            'Day / Boarding',
        ]
        for i, header in enumerate(headers):
            ws.write(row, col, header, self.styles['column_header'])
            col += 1

        for term in terms:
            ws.write(row, col, '%s %% present' % term.title,
                     self.styles['column_header'])
            col += 1

        ws.write(row, col, 'Year to date % present',
                 self.styles['column_header'])

        col += 2
        for header in attendance_headers:
            ws.write(row, col, header, self.styles['column_header'])
            col += 1

        col += 1
        for term in terms:
            for header in attendance_headers:
                ws.write(row, col, header, self.styles['column_header'])
                col += 1
            col += 1

        ws.write(row, col, 'Year to date', self.styles['column_header'])
        col += 1
        for term in terms:
            ws.write(row, col, term.title, self.styles['column_header'])
            col += 1

        row += 1
        students = IGroupContainer(self.schoolyear)['students']
        meetings_by_term = {}
        borders = self.translate_borders([0, 0, 0, 0])
        term_total = {}
        year_breakdown = []
        term_breakdown = {}
        year_breakdown_2 = []
        term_breakdown_2 = {}
        for term in terms:
            meetings_by_term[term] = self.meetings_by_term(term)
            term_total[term] = []
            term_breakdown[term] = []
            term_breakdown_2[term] = []
        today = getUtility(IDateManager).today
        sort_key = lambda p: (self.level(p, today).title
                              if self.level(p, today) else '',
                              p.last_name, p.first_name)
        for person in sorted(students.members, key=sort_key):
            level = self.level(person, today).title if self.level(person, today) else ''
            ws.write(row, 0, level)
            ws.write(row, 1, person.last_name)
            ws.write(row, 2, person.first_name)
            ws.write(row, 3, person.gender.title())
            ws.write(row, 4, get_day_boarding(person))

            col = 5
            year_to_date = []

            for term in terms:
                term_rate, count, present = self.term_rate(
                    person,
                    meetings_by_term[term],
                    [StudentAMRequirement, StudentPMRequirement])
                if count:
                    background = self.get_background_color(term_rate)
                else:
                    background = 'white'
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, term_rate, style)
                if count:
                    year_to_date.append((term_rate, count, present))
                    term_total[term].append((term_rate, count, present))
                col += 1
            student_year_to_date = [rate for rate, c, p in year_to_date]
            value = self.div(sum(student_year_to_date),
                             len(student_year_to_date))
            if student_year_to_date:
                background = self.get_background_color(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)

            col += 2
            col, year_break = self.print_breakdown(ws, row, col, person,
                                                   self.data)
            year_breakdown.append(year_break)
            col += 1
            for term in terms:
                meetings = meetings_by_term[term]
                col, term_break = self.print_breakdown(ws, row, col, person,
                                                       meetings)
                term_breakdown[term].append(term_break)
                col += 1

            col, year_break = self.print_breakdown_2(ws, row, col, person,
                                                     self.data)
            year_breakdown_2.append(year_break)
            for term in terms:
                meetings = meetings_by_term[term]
                col, term_break = self.print_breakdown_2(ws, row, col, person,
                                                         meetings)
                term_breakdown_2[term].append(term_break)

            row += 1

        col = 0
        merge = 4
        ws.write_merge(row, row, col, col + merge, 'Total',
                       self.styles['column_header'])
        col = 5
        year_to_date = []
        for term in terms:
            value = self.div(sum([p for r, c, p in term_total[term]]) * 100,
                             sum([c for r, c, p in term_total[term]]))
            if term_total[term]:
                background = self.get_background_color(value)
                year_to_date.append(value)
            else:
                background = 'white'
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        value = self.div(sum(year_to_date), len(year_to_date))
        if year_to_date:
            background = self.get_background_color(value)
        else:
            background = 'white'
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color %s; '
            'alignment: horizontal center; '
            'borders: %s' % (background, borders),
            num_format_str='0')
        ws.write(row, col, value, style)

        col += 2
        codes = 5
        for i in range(codes):
            value = self.get_column_total(year_breakdown, i)
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        col += 1

        for term in terms:
            for i in range(codes):
                value = self.get_column_total(term_breakdown[term], i)
                background = self.get_gradient_color(value)
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1
            col += 1

        codes = 1
        for i in range(codes):
            value = self.get_column_total(year_breakdown_2, i)
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1

        for term in terms:
            for i in range(codes):
                value = self.get_column_total(term_breakdown_2[term], i)
                background = self.get_gradient_color(value)
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'alignment: horizontal center; '
                    'borders: %s' % (background, borders),
                    num_format_str='0')
                ws.write(row, col, value, style)
                col += 1

        self.finish('sheet1')

    def print_breakdown(self, ws, row, col, teacher, meetings):
        count = 0
        present = 0
        present_late = 0
        sick = 0
        fees = 0
        other = 0
        for meeting, attendance_data in meetings.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person is not teacher:
                        continue
                    reason = person_data['reason']
                    if reason in ('present', 'present_late',
                                  'sick', 'fees', 'other'):
                        count += 1
                    if reason == 'present':
                        present += 1
                    elif reason == 'present_late':
                        present_late += 1
                    elif reason == 'sick':
                        sick += 1
                    elif reason == 'fees':
                        fees += 1
                    elif reason == 'other':
                        other += 1
        l = [present, present_late,
             sick, fees, other]
        borders = self.translate_borders([1, 1, 1, 1])
        result = []
        for attr in l:
            value = (self.div(attr, count) * 100)
            result.append((value, count, attr))
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        return col, result

    def print_breakdown_2(self, ws, row, col, teacher, meetings):
        count = 0
        suspended = 0
        for meeting, attendance_data in meetings.items():
            requirements = [StudentAMRequirement(meeting),
                            StudentPMRequirement(meeting)]
            for requirement in requirements:
                attendance = attendance_data[requirement]
                for person, person_data in attendance.items():
                    if person is not teacher:
                        continue
                    count += 1
                    reason = person_data['reason']
                    if reason == 'suspended':
                        suspended += 1
        l = [suspended]
        borders = self.translate_borders([1, 1, 1, 1])
        result = []
        for attr in l:
            value = (self.div(attr, count) * 100)
            result.append((value, count, attr))
            background = self.get_gradient_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, value, style)
            col += 1
        return col, result

    @Lazy
    def SCORE_TO_REASON(self):
        result = OrderedDict([
            ('P', 'present'),
            ('Pl', 'present_late'),
            ('As', 'sick'),
            ('Af', 'fees'),
            ('Ao', 'other'),
            ('Ss', 'suspended'),
        ])
        return result


class EnrolmentDemographicsReportRequestView(RequestXLSReportDialog):

    report_builder = 'enrolment_demographics_report.xls'


class EnrolmentDemographicsReportView(OverallReportBase):

    message_title = _('Enrolment and demo')

    sheet_1_title = _('Enrolment and demographic report - current status')

    sheet_1_label = _('Enrolment and demo')

    @Lazy
    def custom_colors(self):
        result = super(EnrolmentDemographicsReportView, self).custom_colors
        result['total_bg'] = (30, 217, 217, 217)
        result['girls_bg'] = (31, 230, 185, 184)
        result['boys_bg'] = (32, 142, 180, 227)
        result['orange'] = (33, 255, 192, 0)
        result['red'] = (34, 255, 0, 0)
        result['blue_header'] = (35, 0, 112, 192)
        return result

    def get_background_color(self, value):
        if 60 <= value <= 70:
            return 'orange'
        elif value > 70:
            return 'red'
        else:
            return 'total_bg'

    @Lazy
    def schoolyear(self):
        return self.context

    def print_headers(self, ws, title, merge):
        row = 0
        # XXX: overwritten only for merge
        self.print_report_title(ws, title, row=row, merge=merge)

        row = 2
        today = getUtility(IDateManager).today
        self.print_report_date(ws, today, row=row)

        row += 1
        self.print_report_filters(ws, row=row)

    def addExporters(self, progress):
        super(EnrolmentDemographicsReportView, self).addExporters(progress)
        progress.add(
            'sheet1', title=_('Generating Enrolment and demo sheet'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'enrolment_and_demo_'

    @Lazy
    def data(self):
        result = {}
        return result

    @Lazy
    def day(self):
        day = self.request.get('day', None)
        if day is not None:
            year, month, day = [int(part) for part in day.split('-')]
            return datetime.date(year, month, day)
        today = getUtility(IDateManager).today
        if self.schoolyear.first <= today <= self.schoolyear.last:
            return today
        else:
            return self.schoolyear.last

    @Lazy
    def streams(self):
        result = []
        groups = IGroupContainer(self.schoolyear)
        for group in groups.values():
            if group.__name__ not in defaultGroups:
                result.append(group)
        return sorted(result, key=lambda g: g.title)

    def get_age(self, birth_date):
        today = getUtility(IDateManager).today
        return (
            today.year -
            birth_date.year -
            ((today.month, today.day) < (birth_date.month, birth_date.day)))

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        merge = (len(self.levels) * 3) + 3
        self.print_headers(ws, self.sheet_1_title, merge)
        row = 5
        title = _('Enrolment by full school and class')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color table_title_bg; '
            'borders: top medium, right medium, bottom thin, left medium; '
        )
        ws.write_merge(row, row, 0, merge, title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, '', style)
        ws.write_merge(row, row, 1, 3,
                       'Full school', self.styles['column_header'])
        col = 4
        for i, level in enumerate(self.levels.values()):
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
            )
            if (i == len(self.levels) - 1):
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top thin, right medium, bottom thin, left thin; '
                    'alignment: horizontal center; '
                )
            ws.write_merge(row, row, col, col + 2, level.title, style)
            col += 3
        row += 1
        style = self.easyxf(
            'font: bold on, color blue_header; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, 'Number of students', style)
        headers = [
            ('Total', 'total_bg'),
            ('Girls', 'girls_bg'),
            ('Boys', 'boys_bg'),
        ]
        col = 1
        for header, bg in headers:
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
                % bg)
            ws.write(row, col, header, style)
            col += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left thin; '
            'alignment: horizontal center; '
        )
        for i, level in enumerate(self.levels.values()):
            for header in ['Total', 'Girls', 'Boys']:
                if ((i == len(self.levels) - 1) and
                    header == 'Boys'):
                    style = self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'borders: top thin, right medium, bottom thin, left thin; '
                        'alignment: horizontal center; '
                    )
                ws.write(row, col, header, style)
                col += 1

        # # students
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '# students', style)
        girls = []
        boys = []
        boarding = []
        day = []
        boarding_girls = []
        boarding_boys = []
        day_girls = []
        day_boys = []
        disability = []
        disability_girls = []
        disability_boys = []
        orphan = []
        orphan_girls = []
        orphan_boys = []
        use = []
        use_girls = []
        use_boys = []
        non_use = []
        non_use_girls = []
        non_use_boys = []
        out_of_school = []
        out_of_school_girls = []
        out_of_school_boys = []
        average_age = []
        average_age_girls = []
        average_age_boys = []
        average_distance = []
        average_distance_girls = []
        average_distance_boys = []
        average_ple = []
        average_ple_girls = []
        average_ple_boys = []
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            level_girls = set()
            level_boys = set()
            level_boarding = set()
            level_day = set()
            level_boarding_girls = set()
            level_boarding_boys = set()
            level_day_girls = set()
            level_day_boys = set()
            level_disability = set()
            level_disability_girls = set()
            level_disability_boys = set()
            level_orphan = set()
            level_orphan_girls = set()
            level_orphan_boys = set()
            level_use = set()
            level_use_girls = set()
            level_use_boys = set()
            level_non_use = set()
            level_non_use_girls = set()
            level_non_use_boys = set()
            level_out_of_school = set()
            level_out_of_school_girls = set()
            level_out_of_school_boys = set()
            level_age = []
            level_age_girls = []
            level_age_boys = []
            level_distance = []
            level_distance_girls = []
            level_distance_boys = []
            level_ple = []
            level_ple_girls = []
            level_ple_boys = []
            rel = StudentLevels.bind(level=level).on(self.day).relationships
            for link_info in rel:
                person = link_info.target
                demographics = IDemographics(person)
                if is_female(person, None):
                    level_girls.add(person)
                elif is_male(person, None):
                    level_boys.add(person)
                if person.birth_date is not None:
                    age = self.get_age(person.birth_date)
                    level_age.append(age)
                    if is_female(person, None):
                        level_age_girls.append(age)
                    elif is_male(person, None):
                        level_age_boys.append(age)
                if is_boarding(person, None):
                    level_boarding.add(person)
                elif is_day(person, None):
                    level_day.add(person)
                if is_female_boarding(person, None):
                    level_boarding_girls.add(person)
                elif is_male_boarding(person, None):
                    level_boarding_boys.add(person)
                if is_female_day(person, None):
                    level_day_girls.add(person)
                elif is_male_day(person, None):
                    level_day_boys.add(person)
                if 'yes' in (demographics.get('disability') or '').lower():
                    level_disability.add(person)
                    if is_female(person, None):
                        level_disability_girls.add(person)
                    elif is_male(person, None):
                        level_disability_boys.add(person)
                if 'both dead' in (demographics.get(
                        'parental_status') or '').lower():
                    level_orphan.add(person)
                    if is_female(person, None):
                        level_orphan_girls.add(person)
                    elif is_male(person, None):
                        level_orphan_boys.add(person)
                distance = demographics.get('distance')
                if distance is not None:
                    level_distance.append(distance)
                    if is_female(person, None):
                        level_distance_girls.append(distance)
                    elif is_male(person, None):
                        level_distance_boys.append(distance)
                ple = demographics.get('ple')
                if ple is not None:
                    level_ple.append(ple)
                    if is_female(person, None):
                        level_ple_girls.append(ple)
                    elif is_male(person, None):
                        level_ple_boys.append(ple)
                person_use = demographics.get('use')
                if person_use is not None:
                    if 'non' in person_use.lower():
                        level_non_use.add(person)
                        if is_female(person, None):
                            level_non_use_girls.add(person)
                        elif is_male(person, None):
                            level_non_use_boys.add(person)
                    else:
                        level_use.add(person)
                        if is_female(person, None):
                            level_use_girls.add(person)
                        elif is_male(person, None):
                            level_use_boys.add(person)
                if demographics.get('out_of_school'):
                    level_out_of_school.add(person)
                    if is_female(person, None):
                        level_out_of_school_girls.add(person)
                    elif is_male(person, None):
                        level_out_of_school_boys.add(person)

            girls.append(len(level_girls))
            boys.append(len(level_boys))
            boarding.append(len(level_boarding))
            day.append(len(level_day))
            boarding_girls.append(len(level_boarding_girls))
            boarding_boys.append(len(level_boarding_boys))
            day_girls.append(len(level_day_girls))
            day_boys.append(len(level_day_boys))
            disability.append(len(level_disability))
            disability_girls.append(len(level_disability_girls))
            disability_boys.append(len(level_disability_boys))
            orphan.append(len(level_orphan))
            orphan_girls.append(len(level_orphan_girls))
            orphan_boys.append(len(level_orphan_boys))
            use.append(len(level_use))
            use_girls.append(len(level_use_girls))
            use_boys.append(len(level_use_boys))
            non_use.append(len(level_non_use))
            non_use_girls.append(len(level_non_use_girls))
            non_use_boys.append(len(level_non_use_boys))
            out_of_school.append(len(level_out_of_school))
            out_of_school_girls.append(len(level_out_of_school_girls))
            out_of_school_boys.append(len(level_out_of_school_boys))
            
            level_total = len(level_girls) + len(level_boys)
            if level_total:
                average_age.append(sum(level_age)/float(level_total))
                average_distance.append(sum(level_distance)/float(level_total))
                average_ple.append(sum(level_ple)/float(level_total))


            if level_girls:
                average_age_girls.append(sum(level_age_girls)/float(len(level_girls)))
                average_distance_girls.append(sum(level_distance_girls)/float(len(level_girls)))
                average_ple_girls.append(sum(level_ple_girls)/float(len(level_girls)))
            if level_boys:
                average_age_boys.append(sum(level_age_boys)/float(len(level_boys)))
                average_distance_boys.append(sum(level_distance_boys)/float(len(level_boys)))
                average_ple_boys.append(sum(level_ple_boys)/float(len(level_boys)))

            ws.write(row, col, level_total or '', style)
            col += 1
            ws.write(row, col, len(level_girls) or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, len(level_boys) or '', last_style)
            else:
                ws.write(row, col, len(level_boys) or '', style)
            col += 1

        total_girls = sum(girls)
        total_boys = sum(boys)
        total = total_girls + total_boys
        ws.write(row, 1, total or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
        ))
        ws.write(row, 2, total_girls or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
        ))
        ws.write(row, 3, total_boys or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
        ))

        # % students Girls vs. Boys
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% students Girls vs. Boys', style)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; '
        )
        ws.write(row, 1, '', style)
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, '', style)
            col += 1
            level_total = girls[i] + boys[i]
            if not level_total:
                value = ''
            else:
                value = girls[i] / float(level_total)
            ws.write(row, col, value, style)
            col += 1
            if not level_total:
                value = ''
            else:
                value = boys[i] / float(level_total)
            if (i == len(self.levels) - 1):
                ws.write(row, col, value, last_style)
            else:
                ws.write(row, col, value, style)
            col += 1

        if not total:
            value = 0
        else:
            value = total_girls / float(total)
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        if not total:
            value = 0
        else:
            value = total_boys / float(total)
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))

        # # boarders
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '# borders', style)
        ws.write(row, 1, sum(boarding) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(boarding_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(boarding_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, boarding[i] or '', style)
            col += 1
            ws.write(row, col, boarding_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, boarding_boys[i] or '', last_style)
            else:
                ws.write(row, col, boarding_boys[i] or '', style)
            col += 1

        # # day student
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '# day student', style)
        ws.write(row, 1, sum(day) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(day_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(day_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, day[i] or '', style)
            col += 1
            ws.write(row, col, day_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, day_boys[i] or '', last_style)
            else:
                ws.write(row, col, day_boys[i] or '', style)
            col += 1

        # % boarders (of total # students)
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '% boarders (of total # students)', style)
        value = sum(boarding) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0%'
        ))
        value = sum(boarding_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0%'
        ))
        value = sum(boarding_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0%'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = boarding[i] / float(level_total) if level_total else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = boarding_girls[i] / float(girls[i]) if girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = boarding_boys[i] / float(boys[i]) if boys[i] else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # % day students (of total # students)
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% day students (of total # students)', style)
        value = sum(day) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(day_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(day_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = day[i] / float(level_total) if level_total else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = day_girls[i] / float(girls[i]) if girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = day_boys[i] / float(boys[i]) if boys[i] else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # # students dropped out / unenrolled - total over the year
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '# students dropped out / unenrolled - total over the year', style)
        value = 0
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        value = 0
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        value = 0
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom thin, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            value = 0
            ws.write(row, col, value or '', style)
            col += 1
            value = 0
            ws.write(row, col, value or '', style)
            col += 1
            value = 0
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # blank row
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        first_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; ')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; ')
        for i in range(merge + 1):
            if i == 0:
                ws.write(row, i, '', first_style)
            elif (i == merge):
                ws.write(row, i, '', last_style)
            else:
                ws.write(row, i, '', style)

        # school capacity

        row += 1
        style = self.easyxf(
            'font: bold on, color blue_header; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, 'School capacity used', style)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        for i in range(merge):
            if (i == merge - 1):
                ws.write(row, i + 1, '', last_style)
            else:
                ws.write(row, i + 1, '', style)

        # Available capacity - total students
        app = ISchoolToolApplication(None)
        preferences = IApplicationPreferences(app)
        student_capacity = preferences.student_capacity
        boarding_capacity = preferences.boarding_capacity

        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, 'Available capacity - total students', style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
        )
        ws.write(row, 1, student_capacity or '', style)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        for i in range(merge - 1):
            if (i == merge - 2):
                ws.write(row, i + 2, '', last_style)
            else:
                ws.write(row, i + 2, '', style)
            

        # % capacity used
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% capacity used', style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        )
        value = total / float(student_capacity) if student_capacity else ''
        ws.write(row, 1, value or '', style)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        for i in range(merge - 1):
            if (i == merge - 2):
                ws.write(row, i + 2, '', last_style)
            else:
                ws.write(row, i + 2, '', style)

        # Available capacity - boarders
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, 'Available capacity - boarders', style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
        )
        ws.write(row, 1, boarding_capacity or '', style)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        for i in range(merge - 1):
            if (i == merge - 2):
                ws.write(row, i + 2, '', last_style)
            else:
                ws.write(row, i + 2, '', style)

        # % capacity used
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom medium, left medium; '
        )
        ws.write(row, 0, '% capacity used', style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom medium, left thin; ',
            num_format_str='0%'
        )
        value = sum(boarding) / float(boarding_capacity) if boarding_capacity else ''
        ws.write(row, 1, value or '', style)
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left no_line; ')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; ')
        for i in range(merge - 1):
            if (i == merge - 2):
                ws.write(row, i + 2, '', last_style)
            else:
                ws.write(row, i + 2, '', style)

        # Enrolment by stream

        row += 2
        title = _('Enrolment by stream')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color table_title_bg; '
            'borders: top medium, right medium, bottom thin, left medium; '
        )
        ws.write_merge(row, row, 0, 3, title, style)
        
        row += 1
        title = _('Orange = 60-70 students in a class')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color orange; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, title, style)

        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; ')
        for i in range(1, 4):
            if i == 3:
                ws.write(row, i, '', last_style)
            else:
                ws.write(row, i, '', style)

        row += 1
        title = _('Red = 70+ students in a class')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color red; '
            'borders: top no_line, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, title, style)

        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; ')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom no_line, left no_line; ')
        for i in range(1, 4):
            if i == 3:
                ws.write(row, i, '', last_style)
            else:
                ws.write(row, i, '', style)

        row += 1
        style = self.easyxf(
            'font: bold on, color blue_header; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom thin, left medium; '
        )
        ws.write(row, 0, 'Streams', style)

        headers = [
            ('Total', 'total_bg'),
            ('Girls', 'girls_bg'),
            ('Boys', 'boys_bg'),
        ]
        col = 1
        for header, bg in headers:
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
                % bg)
            if header == 'Boys':
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'borders: top thin, right medium, bottom thin, left thin; '
                    'alignment: horizontal center; '
                    % bg)
            ws.write(row, col, header, style)
            col += 1

        row += 1
        for stream in self.streams:
            ws.write(row, 0, stream.title, self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            members = stream.members.on(self.day)
            female = [1 for p in members if is_female(p, None)]
            male = [1 for p in members if is_male(p, None)]
            value = len(members)
            bg = self.get_background_color(value)
            ws.write(row, 1, value or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; ' % bg, num_format_str='0'
            ))
            ws.write(row, 2, len(female) or '', self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color girls_bg; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            ws.write(row, 3, len(male) or '', self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color boys_bg; '
                'borders: top no_line, right medium, bottom no_line, left no_line; '
            ))
            row += 1

        title = 'Note: red and orange indicators are intended to prompt questions on  capacity - not used for scoring'
        ws.write(row, 0, title, self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left medium; '
        ))
        ws.write(row, 1, '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        ws.write(row, 2, '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        ws.write(row, 3, '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left no_line; '
        ))

        # Other demographic information
        row += 2
        title = _('Other demographic information')
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color table_title_bg; '
            'borders: top medium, right medium, bottom thin, left medium; '
        )
        ws.write_merge(row, row, 0, merge, title, style)
        row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, '', style)
        ws.write_merge(row, row, 1, 3,
                       'Full school', self.styles['column_header'])
        col = 4
        for i, level in enumerate(self.levels.values()):
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
            )
            if (i == len(self.levels) - 1):
                style = self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top thin, right medium, bottom thin, left thin; '
                    'alignment: horizontal center; '
                )
            ws.write_merge(row, row, col, col + 2, level.title, style)
            col += 3
        row += 1
        style = self.easyxf(
            'font: bold on, color blue_header; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, 'Demographic details', style)
        headers = [
            ('Total', 'total_bg'),
            ('Girls', 'girls_bg'),
            ('Boys', 'boys_bg'),
        ]
        col = 1
        for header, bg in headers:
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
                % bg)
            ws.write(row, col, header, style)
            col += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left thin; '
            'alignment: horizontal center; '
        )
        for i, level in enumerate(self.levels.values()):
            for header in ['Total', 'Girls', 'Boys']:
                if ((i == len(self.levels) - 1) and
                    header == 'Boys'):
                    style = self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'borders: top thin, right medium, bottom thin, left thin; '
                        'alignment: horizontal center; '
                    )
                ws.write(row, col, header, style)
                col += 1

        # # students with disabilities
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '# students with disabilities', style)
        ws.write(row, 1, sum(disability) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(disability_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(disability_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, disability[i] or '', style)
            col += 1
            ws.write(row, col, disability_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, disability_boys[i] or '', last_style)
            else:
                ws.write(row, col, disability_boys[i] or '', style)
            col += 1

        # % students with disabilities
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% students with disabilities', style)
        value = sum(disability) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(disability_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(disability_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = disability[i] / float(level_total) if level_total else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = disability_girls[i] / float(girls[i]) if girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = disability_boys[i] / float(boys[i]) if boys[i] else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # # orphans
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '# orphans', style)
        ws.write(row, 1, sum(orphan) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(orphan_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(orphan_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, orphan[i] or '', style)
            col += 1
            ws.write(row, col, orphan_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, orphan_boys[i] or '', last_style)
            else:
                ws.write(row, col, orphan_boys[i] or '', style)
            col += 1

        # % students orphans
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% students orphans', style)
        value = sum(orphan) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(orphan_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(orphan_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = orphan[i] / float(level_total) if level_total else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = orphan_girls[i] / float(girls[i]) if girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = orphan_boys[i] / float(boys[i]) if boys[i] else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # # USE students
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '# USE students', style)
        ws.write(row, 1, sum(use) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(use_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(use_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, use[i] or '', style)
            col += 1
            ws.write(row, col, use_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, use_boys[i] or '', last_style)
            else:
                ws.write(row, col, use_boys[i] or '', style)
            col += 1

        # % USE students
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% USE students', style)
        value = sum(use) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(use_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(use_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = use[i] / float(level_total) if level_total else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = use_girls[i] / float(girls[i]) if girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = use_boys[i] / float(boys[i]) if boys[i] else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # # non-USE students
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, '# non-USE students', style)
        ws.write(row, 1, sum(non_use) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(non_use_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(non_use_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, non_use[i] or '', style)
            col += 1
            ws.write(row, col, non_use_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, non_use_boys[i] or '', last_style)
            else:
                ws.write(row, col, non_use_boys[i] or '', style)
            col += 1

        # % non-USE students
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '% non-USE students', style)
        value = sum(non_use) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(non_use_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        value = sum(non_use_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0%')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0%')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = non_use[i] / float(level_total) if level_total else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = non_use_girls[i] / float(girls[i]) if girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = non_use_boys[i] / float(boys[i]) if boys[i] else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # # Out of School students enrolled this year
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, '# Out of School students enrolled this year', style)
        ws.write(row, 1, sum(out_of_school) or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 2, sum(out_of_school_girls) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        ws.write(row, 3, sum(out_of_school_boys) or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom thin, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            ws.write(row, col, out_of_school[i] or '', style)
            col += 1
            ws.write(row, col, out_of_school_girls[i] or '', style)
            col += 1
            if (i == len(self.levels) - 1):
                ws.write(row, col, out_of_school_boys[i] or '', last_style)
            else:
                ws.write(row, col, out_of_school_boys[i] or '', style)
            col += 1

        # Average age of students
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left medium; '
        )
        ws.write(row, 0, 'Average age of students', style)
        value = 'n/a' if average_age else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
            'alignment: horizontal left; ',
            num_format_str='0'
        ))
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
            'alignment: horizontal left; ',
            num_format_str='0'
        ))
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom no_line, left thin; '
            'alignment: horizontal left; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom no_line, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom no_line, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = average_age[i]  if average_age[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = average_age_girls[i] if average_age_girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = average_age_boys[i] if average_age_boys else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # Average distance to school (mins)
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, 'Average distance to school (mins)', style)
        value = sum(average_distance) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        value = sum(average_distance_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        value = sum(average_distance_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right thin, bottom thin, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom thin, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = average_distance[i]  if average_distance[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = average_distance_girls[i] if average_distance_girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = average_distance_boys[i] if average_distance_boys else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        # Average PLE result
        row += 1
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left medium; '
        )
        ws.write(row, 0, 'Average PLE result', style)
        value = sum(average_ple) / float(total) if total else ''
        ws.write(row, 1, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color total_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        value = sum(average_ple_girls) / float(total_girls) if total_girls else ''
        ws.write(row, 2, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color girls_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        value = sum(average_ple_boys) / float(total_boys) if total_boys else ''
        ws.write(row, 3, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color boys_bg; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0'
        ))
        col = 4
        style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right thin, bottom thin, left thin; ',
            num_format_str='0')
        last_style = self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top thin, right medium, bottom thin, left thin; ',
            num_format_str='0')
        for i, level in enumerate(self.levels.values()):
            level_total = girls[i] + boys[i]
            value = average_ple[i]  if average_ple[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = average_ple_girls[i] if average_ple_girls[i] else ''
            ws.write(row, col, value or '', style)
            col += 1
            value = average_ple_boys[i] if average_ple_boys else ''
            if (i == len(self.levels) - 1):
                ws.write(row, col, value or '', last_style)
            else:
                ws.write(row, col, value or '', style)
            col += 1

        self.finish('sheet1')
